/* include/config.h.  Generated by configure.  */
/*
 *  Configuration header file for compilation of the ALSA driver
 */

#include "config1.h"

/* ALSA section */
#define CONFIG_SND_ISA 1
#define CONFIG_SND_SEQUENCER_MODULE 1
#define CONFIG_SND_OSSEMUL 1
#define CONFIG_SND_MIXER_OSS_MODULE 1
#define CONFIG_SND_PCM_OSS_MODULE 1
#define CONFIG_SND_SEQUENCER_OSS 1
#define CONFIG_SND_VERBOSE_PRINTK 1
/* #undef CONFIG_SND_DEBUG */
/* #undef CONFIG_SND_DEBUG_MEMORY */
/* #undef CONFIG_SND_DEBUG_FULL */
/* #undef CONFIG_SND_DEBUG_DETECT */
/* #undef CONFIG_SND_BIT32_EMUL_MODULE */
/* #undef CONFIG_SND_RTCTIMER_MODULE */
/* #undef CONFIG_SND_HPET_MODULE */
#define CONFIG_SND_SEQ_DUMMY_MODULE 1

/* card-dependent section */
#define CONFIG_SND_GENERIC_PM 1

/* build section */
#define CONFIG_SND_KERNELDIR "/lib/modules/2.4.21-273-default/build"
#define CONFIG_ISAPNP_KERNEL 1
/* #undef CONFIG_PNP_KERNEL */
#define CONFIG_SND_ISAPNP 1
#define CONFIG_SND_PNP 1

/* 2.4 kernels */
#define CONFIG_HAVE_OLD_REQUEST_MODULE 1
/* #undef CONFIG_HAVE_STRLCPY */
#define CONFIG_HAVE_SNPRINTF 1
#define CONFIG_HAVE_VSNPRINTF 1
#define CONFIG_HAVE_VMALLOC_TO_PAGE 1
/* #undef CONFIG_HAVE_PDE */
/* #undef CONFIG_HAVE_TTY_COUNT_ATOMIC */
/* #undef CONFIG_HAVE_VIDEO_GET_DRVDATA */
/* #undef CONFIG_HAVE_DUMP_STACK */

/* 2.2 kernels */
/* #undef CONFIG_OLD_KILL_FASYNC */
/* #undef CONFIG_HAVE_DMA_ADDR_T */
/* #undef CONFIG_HAVE_MUTEX_MACROS */
/* #undef CONFIG_HAVE_SCNPRINTF */
#define CONFIG_HAVE_SSCANF 1

/* 2.4/2.6 kernels */
#define CONFIG_HAVE_PCI_CONSISTENT_DMA_MASK 1
/* #undef CONFIG_HAVE_KCALLOC */
/* #undef CONFIG_HAVE_IO_REMAP_PFN_RANGE */
#define CONFIG_OLD_IO_REMAP_PAGE_RANGE 1
/* #undef CONFIG_HAVE_PCI_SAVED_CONFIG */
/* #undef CONFIG_HAVE_NEW_PCI_SAVE_STATE */
/* #undef CONFIG_HAVE_MSLEEP */
/* #undef CONFIG_HAVE_MSLEEP_INTERRUPTIBLE */
/* #undef CONFIG_HAVE_PCI_DEV_PRESENT */
/* #undef CONFIG_SND_HAVE_NEW_IOCTL */
/* #undef CONFIG_SND_HAVE_CLASS_SIMPLE */
