/* include/config1.h.  Generated by configure.  */
/* Soundcard configuration for ALSA driver */
/* Copyright (c) by Jaroslav Kysela <perex@suse.cz>, */
/*                  Anders Semb Hermansen <ahermans@vf.telia.no> */

/* #undef CONFIG_SOUND_MODULE */
#define CONFIG_SND_MODULE 1
#define CONFIG_SND_TIMER_MODULE 1
#define CONFIG_SND_PCM_MODULE 1
#define CONFIG_SND_HWDEP_MODULE 1
#define CONFIG_SND_RAWMIDI_MODULE 1
#define CONFIG_SND_SEQUENCER_MODULE 1
#define CONFIG_SND_SEQ_DUMMY_MODULE 1
#define CONFIG_SND_OSSEMUL 1
#define CONFIG_SND_MIXER_OSS_MODULE 1
#define CONFIG_SND_PCM_OSS_MODULE 1
#define CONFIG_SND_SEQUENCER_OSS 1
/* #undef CONFIG_SND_RTCTIMER_MODULE */
/* #undef CONFIG_RTC_MODULE */
#define CONFIG_SND_VERBOSE_PRINTK 1
/* #undef CONFIG_SND_DEBUG */
/* #undef CONFIG_SND_DEBUG_MEMORY */
/* #undef CONFIG_SND_DEBUG_DETECT */
#define CONFIG_SND_GENERIC_PM 1
/* #undef CONFIG_SND_BIT32_EMUL_MODULE */
/* #undef CONFIG_SND_HPET_MODULE */
/* #undef CONFIG_HPET_MODULE */
#define CONFIG_SND_MPU401_UART_MODULE 1
#define CONFIG_SND_OPL3_LIB_MODULE 1
#define CONFIG_SND_OPL4_LIB_MODULE 1
#define CONFIG_SND_VX_LIB_MODULE 1
#define CONFIG_SND_DUMMY_MODULE 1
#define CONFIG_SND_VIRMIDI_MODULE 1
#define CONFIG_SND_MTPAV_MODULE 1
#define CONFIG_SND_SERIAL_U16550_MODULE 1
#define CONFIG_SND_MPU401_MODULE 1
#define CONFIG_SND_SERIALMIDI_MODULE 1
#define CONFIG_SND_LOOPBACK_MODULE 1
#define CONFIG_SND_PORTMAN2X4_MODULE 1
#define CONFIG_PARPORT_MODULE 1
/* #undef CONFIG_SND_PCSP_MODULE */
/* #undef CONFIG_X86_MODULE */
/* #undef CONFIG_HAVE_PCSP_HOOK_MODULE */
/* #undef CONFIG_ISA_MODULE */
#define CONFIG_SND_AD1848_LIB_MODULE 1
#define CONFIG_SND_CS4231_LIB_MODULE 1
#define CONFIG_SND_AD1816A_MODULE 1
/* #undef CONFIG_ISAPNP_MODULE */
#define CONFIG_SND_AD1848_MODULE 1
#define CONFIG_SND_CS4231_MODULE 1
#define CONFIG_SND_CS4232_MODULE 1
#define CONFIG_SND_CS4236_MODULE 1
#define CONFIG_SND_ES968_MODULE 1
#define CONFIG_SND_ES1688_MODULE 1
#define CONFIG_SND_ES18XX_MODULE 1
#define CONFIG_SND_GUS_SYNTH_MODULE 1
#define CONFIG_SND_GUSCLASSIC_MODULE 1
#define CONFIG_SND_GUSEXTREME_MODULE 1
#define CONFIG_SND_GUSMAX_MODULE 1
#define CONFIG_SND_INTERWAVE_MODULE 1
#define CONFIG_SND_INTERWAVE_STB_MODULE 1
#define CONFIG_SND_OPTI92X_AD1848_MODULE 1
#define CONFIG_SND_OPTI92X_CS4231_MODULE 1
#define CONFIG_SND_OPTI93X_MODULE 1
#define CONFIG_SND_SB8_MODULE 1
#define CONFIG_SND_SB16_MODULE 1
#define CONFIG_SND_SBAWE_MODULE 1
#define CONFIG_SND_SB16_CSP 1
/* #undef CONFIG_BROKEN_MODULE */
/* #undef CONFIG_PPC_MODULE */
#define CONFIG_SND_WAVEFRONT_MODULE 1
#define CONFIG_SND_ALS100_MODULE 1
#define CONFIG_SND_AZT2320_MODULE 1
#define CONFIG_SND_CMI8330_MODULE 1
#define CONFIG_SND_DT019X_MODULE 1
#define CONFIG_SND_OPL3SA2_MODULE 1
#define CONFIG_SND_SGALAXY_MODULE 1
#define CONFIG_SND_SSCAPE_MODULE 1
/* #undef CONFIG_SND_PC98_CS4232_MODULE */
/* #undef CONFIG_X86_PC9800_MODULE */
#define CONFIG_SND_MIRO_MODULE 1
/* #undef CONFIG_SND_MSND_PINNACLE_MODULE */
/* #undef CONFIG_PCI_MODULE */
#define CONFIG_SND_AC97_CODEC_MODULE 1
#define CONFIG_SND_ALI5451_MODULE 1
#define CONFIG_SND_ATIIXP_MODULE 1
#define CONFIG_SND_ATIIXP_MODEM_MODULE 1
#define CONFIG_SND_AU8810_MODULE 1
#define CONFIG_SND_AU8820_MODULE 1
#define CONFIG_SND_AU8830_MODULE 1
#define CONFIG_SND_AZT3328_MODULE 1
/* #undef CONFIG_EXPERIMENTAL_MODULE */
#define CONFIG_SND_BT87X_MODULE 1
#define CONFIG_SND_BT87X_OVERCLOCK 1
#define CONFIG_SND_CS46XX_MODULE 1
#define CONFIG_SND_CS46XX_NEW_DSP 1
#define CONFIG_SND_CS4281_MODULE 1
#define CONFIG_SND_EMU10K1_MODULE 1
#define CONFIG_SND_EMU10K1X_MODULE 1
#define CONFIG_SND_CA0106_MODULE 1
#define CONFIG_SND_KORG1212_MODULE 1
#define CONFIG_SND_MIXART_MODULE 1
#define CONFIG_SND_NM256_MODULE 1
#define CONFIG_SND_RME32_MODULE 1
#define CONFIG_SND_RME96_MODULE 1
#define CONFIG_SND_RME9652_MODULE 1
#define CONFIG_SND_HDSP_MODULE 1
#define CONFIG_SND_TRIDENT_MODULE 1
#define CONFIG_SND_YMFPCI_MODULE 1
#define CONFIG_SND_ALS4000_MODULE 1
#define CONFIG_SND_CMIPCI_MODULE 1
#define CONFIG_SND_ENS1370_MODULE 1
#define CONFIG_SND_ENS1371_MODULE 1
#define CONFIG_SND_ES1938_MODULE 1
#define CONFIG_SND_ES1968_MODULE 1
#define CONFIG_SND_MAESTRO3_MODULE 1
#define CONFIG_SND_FM801_MODULE 1
#define CONFIG_SND_FM801_TEA575X_MODULE 1
/* #undef CONFIG_VIDEO_DEV_MODULE */
#define CONFIG_SND_ICE1712_MODULE 1
#define CONFIG_SND_ICE1724_MODULE 1
#define CONFIG_SND_INTEL8X0_MODULE 1
#define CONFIG_SND_INTEL8X0M_MODULE 1
#define CONFIG_SND_SONICVIBES_MODULE 1
#define CONFIG_SND_VIA82XX_MODULE 1
#define CONFIG_SND_VIA82XX_MODEM_MODULE 1
#define CONFIG_SND_VX222_MODULE 1
#define CONFIG_SND_HDA_INTEL_MODULE 1
#define CONFIG_SND_PDPLUS_MODULE 1
#define CONFIG_SND_HDSPM_MODULE 1
/* #undef CONFIG_SND_ASIHPI_MODULE */
/* #undef CONFIG_ASIHPK_MODULE */
#define CONFIG_SND_PCXHR_MODULE 1
#define CONFIG_SND_DARLA20_MODULE 1
/* #undef CONFIG_FW_LOADER_MODULE */
#define CONFIG_SND_GINA20_MODULE 1
#define CONFIG_SND_LAYLA20_MODULE 1
#define CONFIG_SND_DARLA24_MODULE 1
#define CONFIG_SND_GINA24_MODULE 1
#define CONFIG_SND_LAYLA24_MODULE 1
#define CONFIG_SND_MONA_MODULE 1
#define CONFIG_SND_MIA_MODULE 1
#define CONFIG_SND_ECHO3G_MODULE 1
#define CONFIG_SND_INDIGO_MODULE 1
#define CONFIG_SND_INDIGOIO_MODULE 1
#define CONFIG_SND_INDIGODJ_MODULE 1
/* #undef CONFIG_SND_POWERMAC_MODULE */
/* #undef CONFIG_I2C_MODULE */
/* #undef CONFIG_INPUT_MODULE */
/* #undef CONFIG_PPC_PMAC_MODULE */
/* #undef CONFIG_ARM_MODULE */
/* #undef CONFIG_SND_SA11XX_UDA1341_MODULE */
/* #undef CONFIG_ARCH_SA1100_MODULE */
/* #undef CONFIG_L3_MODULE */
/* #undef CONFIG_SND_ARMAACI_MODULE */
/* #undef CONFIG_ARM_AMBA_MODULE */
/* #undef CONFIG_SND_PXA2XX_AC97_MODULE */
/* #undef CONFIG_ARCH_PXA_MODULE */
/* #undef CONFIG_SND_PXA2XX_PCM_MODULE */
/* #undef CONFIG_SND_S3C2410_MODULE */
/* #undef CONFIG_ARCH_S3C2410_MODULE */
/* #undef CONFIG_I2C_SENSOR_MODULE */
/* #undef CONFIG_MIPS_MODULE */
/* #undef CONFIG_SND_AU1X00_MODULE */
/* #undef CONFIG_SOC_AU1000_MODULE */
/* #undef CONFIG_SOC_AU1100_MODULE */
/* #undef CONFIG_SOC_AU1500_MODULE */
/* #undef CONFIG_USB_MODULE */
#define CONFIG_SND_USB_AUDIO_MODULE 1
/* #undef CONFIG_SND_USB_USX2Y_MODULE */
/* #undef CONFIG_ALPHA_MODULE */
/* #undef CONFIG_PCMCIA_MODULE */
#define CONFIG_SND_VXPOCKET_MODULE 1
#define CONFIG_SND_VXP440_MODULE 1
#define CONFIG_SND_PDAUDIOCF_MODULE 1
/* #undef CONFIG_SPARC32_MODULE */
/* #undef CONFIG_SPARC64_MODULE */
/* #undef CONFIG_SND_SUN_AMD7930_MODULE */
/* #undef CONFIG_SBUS_MODULE */
/* #undef CONFIG_SND_SUN_CS4231_MODULE */
/* #undef CONFIG_SND_SUN_DBRI_MODULE */
/* #undef CONFIG_GSC_MODULE */
/* #undef CONFIG_SND_HARMONY_MODULE */
/* #undef CONFIG_SOUND_PRIME_MODULE */
