/* include/autoconf-extra.h.  Generated by configure.  */
/*
 * external ALSA section by alsa-driver
 */
#define CONFIG_SND_MODULE 1
/* #undef CONFIG_SND_BIT32_EMUL_MODULE */
/* #undef CONFIG_SND_RTCTIMER_MODULE */
/* #undef CONFIG_SND_HPET_MODULE */
#define CONFIG_SND_SEQUENCER_MODULE 1
#define CONFIG_SND_SEQ_DUMMY_MODULE 1
#define CONFIG_SND_OSSEMUL 1
#define CONFIG_SND_MIXER_OSS_MODULE 1
#define CONFIG_SND_PCM_OSS_MODULE 1
#define CONFIG_SND_SEQUENCER_OSS 1
/* #undef CONFIG_SND_DEBUG */
/* #undef CONFIG_SND_DEBUG_MEMORY */
/* #undef CONFIG_SND_DEBUG_FULL */
/* #undef CONFIG_SND_DEBUG_DETECT */
#define CONFIG_SND_VERBOSE_PRINTK 1

