#ifndef __SOUND_JULI_H
#define __SOUND_JULI_H

#define JULI_DEVICE_DESC		"{ESI,Juli@},"

#define VT1724_SUBDEVICE_JULI		0x31305345	/* Juli@ */

extern struct snd_ice1712_card_info  snd_vt1724_juli_cards[];

#endif	/* __SOUND_JULI_H */
