#include <sound/driver.h>
#include <linux/pci.h>

#ifndef PCI_VENDOR_ID_AUREAL
#define PCI_VENDOR_ID_AUREAL 0x12eb
#endif
#ifndef PCI_DEVICE_ID_AUREAL_VORTEX_1
#define PCI_DEVICE_ID_AUREAL_VORTEX_1 0x0001
#endif
#ifndef PCI_DEVICE_ID_AUREAL_VORTEX_2
#define PCI_DEVICE_ID_AUREAL_VORTEX_2 0x0002
#endif
#ifndef PCI_DEVICE_ID_AUREAL_ADVANTAGE
#define PCI_DEVICE_ID_AUREAL_ADVANTAGE 0x0003
#endif
#ifndef PCI_DEVICE_ID_VIA_8365_1
#define PCI_DEVICE_ID_VIA_8365_1 0x8305
#endif
#ifndef PCI_DEVICE_ID_AMD_FE_GATE_7007
#define PCI_DEVICE_ID_AMD_FE_GATE_7007 0x7007
#endif
