/* -*- Mode: C -*- */

#ifndef PDPLUS_PGA_ADAT_H
#define PDPLUS_PGA_ADAT_H

static unsigned char const pga_adat[] = {
	0xff, 0x04, 0xd4, 0xbe, 0xf9, 0xea, 0xff, 0x7c, 
	0xfd, 0xad, 0xd5, 0x5f, 0x7f, 0xfd, 0xcd, 0xd4, 
	0xdf, 0x4b, 0x73, 0x9d, 0x69, 0xbe, 0xfe, 0xe6, 
	0xea, 0xaf, 0xb7, 0xfe, 0xee, 0xea, 0x6f, 0xbd, 
	0x3f, 0xf8, 0xff, 0xf6, 0xfb, 0xef, 0xbd, 0xff, 
	0xfe, 0xfb, 0x6f, 0xbb, 0xff, 0xbe, 0x5b, 0xdf, 
	0x76, 0xfb, 0xfd, 0xb7, 0xde, 0x7f, 0xdf, 0xfd, 
	0xb7, 0xde, 0x7f, 0xff, 0xff, 0xe2, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xfe, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0x9b, 0xff, 0xbf, 0xff, 0xfe, 0xcb, 0xef, 
	0xbf, 0xff, 0xfe, 0xfb, 0xef, 0x8f, 0xfe, 0xf6, 
	0xf7, 0xdf, 0x7f, 0xff, 0xed, 0xf7, 0xdf, 0x7f, 
	0x7f, 0xec, 0xf7, 0x5f, 0xff, 0x3f, 0xfe, 0xff, 
	0xd6, 0x5f, 0x6f, 0xfc, 0xf5, 0xd7, 0x5f, 0x7f, 
	0xfd, 0xf5, 0xc7, 0x1b, 0xff, 0x78, 0xeb, 0xaf, 
	0x37, 0xfe, 0xfa, 0xeb, 0xaf, 0x3f, 0xfe, 0x7a, 
	0xeb, 0xff, 0xfb, 0xff, 0xd5, 0x5f, 0x5f, 0xe5, 
	0xf5, 0xd7, 0x5f, 0x7f, 0xfd, 0xf5, 0x57, 0x56, 
	0xf9, 0x4a, 0xea, 0xaf, 0xaf, 0xf2, 0xfa, 0xeb, 
	0xaf, 0xb7, 0xf2, 0xfa, 0xea, 0xff, 0xc9, 0xff, 
	0x3b, 0xff, 0xbc, 0xf3, 0xcf, 0x3f, 0xff, 0xfc, 
	0xf3, 0xcf, 0x3b, 0xef, 0xfc, 0xe7, 0x9f, 0x7f, 
	0xde, 0xf9, 0xe7, 0x9f, 0x7f, 0xfe, 0xf9, 0xe7, 
	0x9d, 0xff, 0xb3, 0xff, 0xff, 0xfe, 0xfb, 0xef, 
	0xbe, 0xff, 0xfe, 0xfb, 0xef, 0xbf, 0xff, 0xde, 
	0xbb, 0xdf, 0x7d, 0xff, 0x7d, 0xf7, 0xde, 0x7f, 
	0xff, 0xfd, 0x77, 0xdf, 0x7f, 0xff, 0x2f, 0xfe, 
	0xff, 0xfb, 0xaf, 0xb7, 0xff, 0xfa, 0xeb, 0xaf, 
	0xbd, 0xfb, 0xfa, 0xfb, 0xef, 0x7f, 0xf7, 0xf5, 
	0xd7, 0xdf, 0x7f, 0xfd, 0xfd, 0xd7, 0xdf, 0x7f, 
	0xbd, 0xf5, 0xf7, 0xf9, 0xff, 0xed, 0xbf, 0xdf, 
	0xf6, 0xfb, 0xef, 0xbf, 0xff, 0xfe, 0xfb, 0xed, 
	0xb7, 0xff, 0xcd, 0xf7, 0xdf, 0x7f, 0xfb, 0xfd, 
	0xf7, 0xdf, 0x7f, 0xff, 0xfd, 0xf6, 0xff, 0xc7, 
	0xff, 0xdf, 0x7f, 0xff, 0xfd, 0xf7, 0xdf, 0x77, 
	0xff, 0xfd, 0xf7, 0xdf, 0x3f, 0xff, 0xfb, 0xef, 
	0xbf, 0xef, 0xfe, 0xfb, 0xef, 0xbb, 0xff, 0xfe, 
	0xfb, 0xef, 0xff, 0x8f, 0xff, 0x77, 0xfd, 0xf7, 
	0xdf, 0x7f, 0xff, 0xfd, 0xf7, 0xdf, 0x7e, 0xff, 
	0xfd, 0xf7, 0xbe, 0xbf, 0xfe, 0xfb, 0xeb, 0xbf, 
	0xfb, 0xfe, 0xfb, 0xaf, 0xbf, 0xff, 0xfe, 0x3f, 
	0xfe, 0xfd, 0xed, 0xb7, 0xdf, 0x7e, 0xfb, 0xed, 
	0xb7, 0xde, 0x7e, 0xfb, 0xcd, 0x37, 0xaf, 0xdd, 
	0xf6, 0xdb, 0x6f, 0xaf, 0xbd, 0xf6, 0xdb, 0x6f, 
	0xbb, 0xfd, 0xf6, 0x3f, 0xf9, 0xff, 0xf7, 0xdf, 
	0x7f, 0xfd, 0xfd, 0xf7, 0xdf, 0x7f, 0xdf, 0xfd, 
	0xf7, 0xdf, 0xbf, 0xfc, 0xfb, 0xef, 0xbf, 0xff, 
	0xfe, 0xfb, 0xaf, 0xbf, 0xff, 0xfe, 0xfb, 0xff, 
	0xee, 0xff, 0xdf, 0x7f, 0xff, 0x7d, 0xf7, 0xdf, 
	0x7f, 0xef, 0xfd, 0xf7, 0xfd, 0x7f, 0xdf, 0xff, 
	0xcf, 0xbb, 0xff, 0x7e, 0xfb, 0xeb, 0xbf, 0xff, 
	0xfe, 0xfb, 0xef, 0xff, 0xb3, 0xfd, 0xc7, 0x3e, 
	0xbf, 0xfc, 0xf3, 0xcf, 0x1f, 0xff, 0xdc, 0xf3, 
	0xcb, 0x3e, 0xff, 0xfd, 0xf1, 0x9f, 0x7f, 0xfe, 
	0xfd, 0xf7, 0x9f, 0x7f, 0xfe, 0x79, 0xe5, 0xff, 
	0x3f, 0x7e, 0xff, 0xfd, 0xf7, 0xff, 0x7f, 0xff, 
	0xfd, 0xf7, 0xff, 0x7f, 0xff, 0xfd, 0xef, 0x7f, 
	0xff, 0xff, 0xbb, 0xff, 0x3b, 0xff, 0xfe, 0xfb, 
	0xef, 0xbf, 0xff, 0x7e, 0x3f, 0x5b, 0x75, 0xfc, 
	0xf3, 0xcf, 0x3f, 0xff, 0xfc, 0xf3, 0xcf, 0x3f, 
	0xff, 0xf4, 0xf3, 0x9f, 0x7f, 0xee, 0xf9, 0xe7, 
	0x9f, 0x5f, 0xfe, 0xf9, 0xe7, 0x9f, 0x7f, 0xfe, 
	0xd5, 0xe1, 0xff, 0xfd, 0xbd, 0xff, 0x9b, 0xfb, 
	0xef, 0xbf, 0xff, 0xfb, 0xfb, 0xbf, 0xfb, 0xe6, 
	0x77, 0xff, 0xde, 0xff, 0xcd, 0xfd, 0xf7, 0xdf, 
	0xff, 0xdd, 0xfd, 0xff, 0xf6, 0x97, 0xff, 0xef, 
	0xbf, 0xff, 0xf6, 0xfa, 0xef, 0xbf, 0xef, 0xf6, 
	0xfb, 0x6f, 0xbf, 0xfe, 0xfd, 0xf7, 0xdd, 0x7f, 
	0xfb, 0xfd, 0xf7, 0xdf, 0x7f, 0xfd, 0xfd, 0xf3, 
	0xff, 0x9f, 0x7a, 0xbf, 0xfd, 0xf6, 0xdb, 0x6f, 
	0xbf, 0xf5, 0xf6, 0x9b, 0x6f, 0x9f, 0xff, 0xd2, 
	0xf7, 0xdf, 0x3e, 0xfb, 0xad, 0xb7, 0xde, 0x7e, 
	0xfb, 0xe5, 0xb7, 0xde, 0xfe, 0xef, 0xfa, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xdf, 0xff, 
	0xff, 0xff, 0x7f, 0xfd, 0x7f, 0xff, 0xff, 0xff, 
	0xff, 0xfb, 0xff, 0xff, 0xff, 0xff, 0xfb, 0xff, 
	0xff, 0xff, 0xea, 0xff, 0xff, 0xff, 0xdf, 0xff, 
	0xff, 0x7f, 0xff, 0xff, 0xff, 0xdf, 0xff, 0xff, 
	0xbd, 0xcf, 0xff, 0xff, 0xfe, 0xfb, 0xff, 0xff, 
	0xff, 0xff, 0xfb, 0xff, 0xff, 0xff, 0xb3, 0xff, 
	0xbf, 0xff, 0xff, 0xd7, 0xfd, 0xff, 0xdf, 0x7f, 
	0xf7, 0xfd, 0xfd, 0xde, 0x7f, 0x3f, 0xff, 0xff, 
	0xaf, 0xdb, 0xff, 0xff, 0xfc, 0xbf, 0xfd, 0xff, 
	0xde, 0xff, 0x5f, 0xfe, 0xfb, 0xff, 0xff, 0xff, 
	0xfb, 0xdf, 0xff, 0xfb, 0xff, 0xff, 0xff, 0x3b, 
	0xef, 0xfe, 0xe7, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x2f, 0xf9, 
	0xff, 0xf3, 0x9f, 0x3f, 0xff, 0xfd, 0xf7, 0xdf, 
	0x3f, 0xfd, 0xfd, 0xf7, 0xce, 0xff, 0xfe, 0xf9, 
	0xef, 0x9f, 0xff, 0xfe, 0xfb, 0xef, 0x9f, 0xff, 
	0xde, 0xf9, 0xff, 0xe2, 0xff, 0xff, 0xff, 0xbf, 
	0xff, 0xff, 0xff, 0xfd, 0x3f, 0xf9, 0x7f, 0xd3, 
	0x2f, 0x3f, 0xf9, 0xbf, 0xff, 0xdf, 0xff, 0x7f, 
	0xff, 0xff, 0x57, 0xfe, 0x7f, 0xff, 0xff, 0x97, 
	0xff, 0xff, 0xff, 0xff, 0xfe, 0xff, 0xff, 0xf7, 
	0xff, 0xfe, 0xfd, 0xef, 0x6f, 0xff, 0xfb, 0x7f, 
	0xff, 0x5f, 0x7f, 0xff, 0xbd, 0xff, 0xbd, 0xdf, 
	0xdf, 0x7f, 0xff, 0x6f, 0xec, 0xff, 0xf7, 0xff, 
	0xff, 0xff, 0xff, 0x7f, 0xff, 0xff, 0xff, 0xff, 
	0xef, 0xef, 0xff, 0xf7, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xfb, 0xff, 0xfb, 0xdf, 0xbf, 
	0xfa, 0xff, 0xff, 0xff, 0xff, 0xef, 0xbf, 0xff, 
	0xff, 0xff, 0xfe, 0xff, 0xff, 0xff, 0xfb, 0xdf, 
	0xff, 0xff, 0x7f, 0xff, 0xbf, 0xbf, 0xff, 0x7f, 
	0xff, 0xff, 0xf7, 0xf7, 0xea, 0xdf, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xbd, 0xff, 0xff, 0xf7, 0xff, 
	0xff, 0x7f, 0xdf, 0xbf, 0x7e, 0xff, 0xfb, 0xff, 
	0xff, 0xff, 0xfd, 0xf7, 0xff, 0xff, 0xff, 0xdf, 
	0x93, 0xff, 0x8f, 0xff, 0xdf, 0xfa, 0xff, 0xff, 
	0xf3, 0xf7, 0xfa, 0xbf, 0xfd, 0xbd, 0xfc, 0xf7, 
	0xe7, 0xfd, 0x7f, 0xfb, 0xef, 0xff, 0xff, 0xff, 
	0x7f, 0xff, 0xc7, 0xff, 0xcf, 0xfe, 0xff, 0xdf, 
	0xff, 0xda, 0xf7, 0xff, 0xff, 0xff, 0xef, 0xff, 
	0xff, 0xbf, 0x97, 0xef, 0xfd, 0xff, 0xef, 0xf7, 
	0xff, 0xfd, 0xff, 0xbf, 0xff, 0xff, 0xdf, 0xef, 
	0xff, 0xfa, 0xff, 0xfa, 0xe7, 0xff, 0xff, 0xde, 
	0xff, 0xfd, 0xef, 0xff, 0xfe, 0x77, 0x6b, 0x6f, 
	0xff, 0xec, 0xf3, 0xfb, 0xff, 0xff, 0x7f, 0xff, 
	0xff, 0xff, 0xdf, 0xff, 0xf7, 0xed, 0xff, 0x7f, 
	0xaf, 0xff, 0xff, 0xdf, 0x7b, 0x7f, 0xff, 0xf7, 
	0xac, 0xff, 0xef, 0x37, 0x7f, 0xd5, 0xff, 0xff, 
	0xfe, 0xff, 0xef, 0xff, 0xff, 0xfe, 0x7f, 0xef, 
	0xff, 0x83, 0xff, 0x7d, 0xff, 0xff, 0x7f, 0xdf, 
	0xbf, 0xff, 0xff, 0xff, 0xef, 0xff, 0xff, 0xfc, 
	0xe6, 0xfe, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xfd, 0xff, 0xef, 0xfe, 0xfe, 
	0xf3, 0xbf, 0x7f, 0xcd, 0x3f, 0xfc, 0xff, 0xff, 
	0xff, 0xfd, 0x6f, 0x6c, 0xa5, 0xff, 0xf5, 0xff, 
	0xef, 0xff, 0xfb, 0xff, 0xff, 0xff, 0xff, 0xfb, 
	0xff, 0xff, 0xf8, 0xff, 0xff, 0xff, 0xff, 0xdf, 
	0x6e, 0xfb, 0xff, 0xff, 0xfd, 0xff, 0xbf, 0xf3, 
	0x9e, 0xbf, 0xf7, 0xbe, 0xff, 0xef, 0xff, 0xef, 
	0xbf, 0xff, 0xff, 0xff, 0xff, 0xff, 0xe8, 0xff, 
	0xcf, 0xaf, 0xf7, 0xff, 0xb9, 0xdf, 0xff, 0xff, 
	0xff, 0xef, 0xde, 0xb7, 0xfc, 0xbd, 0x7f, 0x57, 
	0xff, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xfd, 0xff, 
	0xff, 0xf7, 0x8b, 0xff, 0xff, 0xff, 0xfa, 0xf7, 
	0xe6, 0xff, 0xff, 0x7f, 0xff, 0xfd, 0xfe, 0x7f, 
	0xeb, 0xf3, 0xff, 0x7e, 0xdf, 0xff, 0xff, 0xff, 
	0xff, 0xef, 0xfb, 0xff, 0xff, 0xff, 0x0f, 0xfe, 
	0xbf, 0xff, 0xfd, 0xff, 0xbe, 0xfb, 0xf7, 0xff, 
	0xff, 0xff, 0xff, 0xfb, 0xff, 0xdf, 0xff, 0xf7, 
	0xfe, 0x9f, 0xff, 0xff, 0xf7, 0xff, 0xff, 0xff, 
	0xff, 0xef, 0xff, 0xf2, 0xff, 0xff, 0x7f, 0xff, 
	0x97, 0xc9, 0x3f, 0xff, 0x7c, 0xfe, 0x5f, 0xe7, 
	0x4b, 0x5f, 0xf8, 0xff, 0xfd, 0x77, 0xf9, 0x7f, 
	0xef, 0xb7, 0xd7, 0x5e, 0x7f, 0xed, 0x3f, 0x60, 
	0xff, 0xff, 0xff, 0xff, 0x9f, 0x3e, 0xff, 0xfc, 
	0xe7, 0xff, 0x3e, 0xbf, 0xef, 0x7f, 0xeb, 0xbd, 
	0xf7, 0xff, 0xcd, 0xff, 0xfd, 0xf7, 0xff, 0xff, 
	0xff, 0xff, 0x7f, 0xbf, 0xff, 0xe9, 0xf3, 0xff, 
	0x7f, 0xaa, 0xe9, 0xff, 0xff, 0x2f, 0xef, 0xff, 
	0xe7, 0xce, 0xf7, 0xf4, 0xff, 0x2b, 0xfd, 0xff, 
	0xde, 0xdb, 0xef, 0xbf, 0xff, 0xfe, 0xff, 0x1f, 
	0xd2, 0xff, 0xff, 0xff, 0xff, 0xc9, 0xbf, 0xff, 
	0xff, 0xff, 0xad, 0xff, 0x9f, 0xfe, 0xf7, 0xdf, 
	0xef, 0x3f, 0xfd, 0xff, 0x6b, 0xaf, 0xbd, 0xf4, 
	0xd2, 0x4b, 0xff, 0xf5, 0x98, 0x7f, 0xde, 0x7d, 
	0xff, 0xbf, 0x5d, 0xda, 0x7f, 0xff, 0xaf, 0x7f, 
	0xff, 0xcb, 0x5f, 0xfe, 0x7d, 0xfd, 0xf7, 0xea, 
	0x7f, 0xff, 0x37, 0xd3, 0x4c, 0x2b, 0xcd, 0xee, 
	0xe1, 0xfe, 0x6e, 0xf3, 0xfe, 0xff, 0xea, 0xab, 
	0xff, 0xff, 0xff, 0xee, 0xfc, 0xef, 0x7f, 0xff, 
	0x55, 0x73, 0xff, 0xab, 0xf7, 0xfd, 0xf7, 0xfe, 
	0xfb, 0xef, 0xbe, 0x3f, 0xa7, 0xfb, 0xff, 0xed, 
	0xcf, 0xff, 0x6b, 0xe5, 0xd6, 0xff, 0xdf, 0xeb, 
	0xf1, 0xbf, 0x5e, 0xf5, 0xd7, 0x5f, 0x7f, 0xfd, 
	0xff, 0xdf, 0xdf, 0xff, 0xff, 0xbf, 0xfd, 0xff, 
	0xc6, 0xee, 0xff, 0xfe, 0x4e, 0xfd, 0x2f, 0x97, 
	0xd7, 0x6f, 0xfd, 0xaf, 0xff, 0xff, 0xf2, 0x97, 
	0x5f, 0x7f, 0xbd, 0xbc, 0xff, 0x7f, 0xff, 0xfd, 
	0xf5, 0xd7, 0x5f, 0xff, 0xbb, 0x72, 0xcf, 0xbf, 
	0xdf, 0xff, 0xff, 0xff, 0xf7, 0xde, 0xfc, 0xff, 
	0xef, 0xff, 0xff, 0xec, 0xf7, 0xff, 0xff, 0xff, 
	0xf9, 0xf7, 0xdf, 0x7d, 0xf7, 0xdd, 0x77, 0xdf, 
	0xf7, 0xed, 0x7d, 0xff, 0x71, 0xfe, 0xff, 0xff, 
	0x9f, 0x7f, 0xf6, 0xff, 0x7f, 0xfc, 0xff, 0xe7, 
	0xbf, 0xfe, 0x7c, 0xf3, 0xef, 0x9d, 0x5f, 0xfe, 
	0xf9, 0xe7, 0x9f, 0x7f, 0xde, 0x37, 0xfb, 0xb6, 
	0xdf, 0xfc, 0xfb, 0x4d, 0x37, 0xfe, 0xbc, 0xfb, 
	0xed, 0xbf, 0xff, 0x74, 0xe6, 0xd8, 0x7f, 0xff, 
	0xdd, 0xb7, 0xd3, 0x4f, 0x2f, 0xbd, 0xf4, 0xd7, 
	0xff, 0x86, 0xee, 0xcf, 0xef, 0xff, 0xff, 0xf6, 
	0xcf, 0xff, 0xff, 0xff, 0xfb, 0xff, 0xff, 0xbf, 
	0xff, 0xef, 0xff, 0xff, 0xfe, 0xff, 0x4d, 0x36, 
	0xf9, 0xe5, 0x97, 0x4f, 0xfe, 0xb3, 0xb9, 0xef, 
	0xff, 0xbf, 0xff, 0xf7, 0xef, 0xff, 0xbf, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xf7, 0xff, 0xdf, 
	0xfe, 0xff, 0xef, 0xbb, 0xef, 0xbf, 0xff, 0xf6, 
	0xfb, 0xef, 0xed, 0xfe, 0xff, 0x37, 0xfd, 0xff, 
	0xe2, 0x0f, 0x27, 0xfe, 0xff, 0xfa, 0xff, 0x2f, 
	0x7e, 0xf9, 0xf5, 0xdf, 0x5f, 0x6b, 0xfd, 0xff, 
	0xff, 0x7f, 0xff, 0xfd, 0xf7, 0xbf, 0x8b, 0xf7, 
	0xff, 0xfe, 0xf9, 0xff, 0xf9, 0x27, 0xbe, 0xf8, 
	0xff, 0xcb, 0xff, 0xbf, 0xfe, 0xf5, 0x97, 0x7f, 
	0x7c, 0xf5, 0xe5, 0x7f, 0xfd, 0x7d, 0xf7, 0xdd, 
	0x77, 0x7f, 0xc3, 0xfe, 0xbf, 0xff, 0xfb, 0xff, 
	0xaf, 0xfe, 0xf4, 0xbf, 0xff, 0xbf, 0xff, 0xff, 
	0xf7, 0xd7, 0x6f, 0x7f, 0xfd, 0x77, 0xff, 0x5f, 
	0x7f, 0xfd, 0xfd, 0xf7, 0xdf, 0xff, 0x3b, 0xf8, 
	0xff, 0xfa, 0xfc, 0xf3, 0x8f, 0x3f, 0xda, 0xdc, 
	0xf3, 0x8f, 0x3f, 0xff, 0xfc, 0x73, 0x1f, 0x7f, 
	0x74, 0xf1, 0xe7, 0x1f, 0x7d, 0xf4, 0xd9, 0x67, 
	0x9f, 0x7d, 0xee, 0xee, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xfb, 0xff, 0x2f, 0xfe, 0xff, 0xe3, 0xff, 
	0xaf, 0x3f, 0xff, 0xf7, 0x1f, 0xff, 0xff, 0xff, 
	0xf5, 0xd7, 0xdf, 0x7f, 0xff, 0xfd, 0xff, 0xa7, 
	0xff, 0xff, 0xfd, 0xf7, 0xff, 0xff, 0xbf, 0xbf, 
	0xfe, 0xff, 0xef, 0xff, 0xff, 0xfe, 0xfd, 0xdf, 
	0x7f, 0xd7, 0xf7, 0xff, 0xd7, 0x5e, 0xfb, 0xed, 
	0xb7, 0xdf, 0xfe, 0x3f, 0xfe, 0xff, 0xfe, 0xff, 
	0xff, 0xbf, 0xff, 0xff, 0xff, 0xff, 0xbf, 0xff, 
	0xff, 0xfa, 0xf7, 0x7f, 0xff, 0xef, 0xf7, 0xff, 
	0xff, 0xff, 0xff, 0xf7, 0xdf, 0x7f, 0xff, 0x7f, 
	0xfa, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xdb, 
	0x7f, 0xff, 0xff, 0xfc, 0xf7, 0xef, 0xef, 0xff, 
	0xfd, 0x7f, 0xf7, 0xff, 0xfe, 0xff, 0xff, 0xf7, 
	0xdf, 0x7f, 0xff, 0xff, 0xeb, 0xff, 0xff, 0x6f, 
	0xfe, 0xff, 0xff, 0x93, 0xff, 0xff, 0xdf, 0xf6, 
	0xff, 0x5f, 0xbd, 0xf2, 0xed, 0xb7, 0xdf, 0xfe, 
	0xff, 0xed, 0xbf, 0xdf, 0x7f, 0xfa, 0xef, 0xff, 
	0x8f, 0xff, 0xff, 0x78, 0xeb, 0xaf, 0xbf, 0xee, 
	0xe2, 0xcb, 0xaf, 0xba, 0xfc, 0xfa, 0x8b, 0x5f, 
	0x6e, 0x75, 0x61, 0x97, 0x5f, 0x7f, 0xfd, 0x75, 
	0xd7, 0x4b, 0x7f, 0xf5, 0x3f, 0xf6, 0xff, 0xfb, 
	0xf6, 0xff, 0xff, 0xff, 0xfb, 0xff, 0xff, 0xed, 
	0xff, 0x9c, 0xff, 0xff, 0xdf, 0xf7, 0xff, 0xe4, 
	0xff, 0x8f, 0x7d, 0xdf, 0x7f, 0xfb, 0xdf, 0xff, 
	0xfb, 0xf3, 0xff, 0xff, 0xf6, 0xff, 0xfb, 0xcf, 
	0xff, 0xfd, 0xff, 0xf7, 0x7b, 0xdb, 0xff, 0xfb, 
	0xf7, 0xdf, 0xff, 0xf7, 0xfb, 0xe7, 0xff, 0xfe, 
	0xfe, 0xff, 0xff, 0xff, 0x7f, 0xea, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xdf, 0xff, 0xfe, 0xfd, 0xff, 
	0xef, 0xdb, 0x7f, 0xff, 0xff, 0xfe, 0xff, 0x7f, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0x7f, 0x23, 0xff, 0xcf, 0xfb, 0xdf, 0xef, 0xeb, 
	0xbf, 0xff, 0xff, 0xff, 0xcd, 0xfb, 0xbf, 0x7c, 
	0xf5, 0xff, 0x5f, 0xef, 0xff, 0xf7, 0xd7, 0x9f, 
	0x6f, 0xee, 0xf9, 0xe7, 0xfb, 0x4f, 0xfe, 0xfb, 
	0x7f, 0xff, 0xfe, 0x7f, 0xbe, 0xff, 0xff, 0xff, 
	0xef, 0xff, 0xff, 0xfb, 0xb3, 0xff, 0xff, 0xff, 
	0xfe, 0xff, 0xdf, 0xfa, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0x3f, 0xfb, 0xff, 0xff, 0xff, 0x9f, 0xb3, 
	0xdf, 0xfb, 0x7f, 0x9f, 0xf5, 0xff, 0xfb, 0x83, 
	0x6f, 0xfe, 0xef, 0xb7, 0xbf, 0xcf, 0xff, 0xeb, 
	0xff, 0xff, 0xff, 0xf7, 0xff, 0xff, 0xe9, 0xff, 
	0xfd, 0x2f, 0xba, 0xfc, 0x7f, 0xf3, 0xff, 0xbf, 
	0xff, 0xfe, 0x7b, 0xef, 0x2f, 0xf6, 0xff, 0xc7, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xfe, 0xbf, 0xfb, 
	0xff, 0xff, 0xb7, 0xfe, 0xff, 0xff, 0x77, 0xdf, 
	0x7f, 0xef, 0xf7, 0xff, 0xfe, 0x6f, 0x6f, 0xbf, 
	0x3b, 0xbd, 0xfd, 0xff, 0xbf, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xbf, 0xff, 0xff, 0x8e, 0xfe, 
	0x3f, 0x7e, 0xff, 0xfb, 0xf6, 0xdf, 0xfd, 0xfe, 
	0xff, 0xff, 0x3f, 0xfe, 0xff, 0xff, 0x1f, 0xbf, 
	0xf7, 0xff, 0x67, 0xff, 0xfe, 0xff, 0xff, 0xbc, 
	0xff, 0xff, 0x7f, 0xf9, 0xff, 0xf7, 0xff, 0xef, 
	0xff, 0xff, 0xff, 0xf9, 0xff, 0xff, 0xff, 0xf7, 
	0xfe, 0xfe, 0xf5, 0xfd, 0xff, 0xff, 0x7f, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xe0, 
	0xff, 0xff, 0xbf, 0xff, 0xfc, 0xdf, 0xff, 0x7e, 
	0xdb, 0x76, 0xff, 0xff, 0xfd, 0x7f, 0xfd, 0x67, 
	0xdb, 0xff, 0xbf, 0xf6, 0xff, 0xff, 0xff, 0xff, 
	0xf7, 0xff, 0xfd, 0x0b, 0xff, 0x37, 0x7e, 0xff, 
	0xfb, 0xff, 0xff, 0xff, 0xff, 0xfd, 0x4f, 0x3f, 
	0xfb, 0xfe, 0xbf, 0xff, 0x5e, 0xf7, 0xfb, 0xff, 
	0xff, 0xff, 0xbe, 0xfb, 0xef, 0xbf, 0xfb, 0x97, 
	0xfe, 0xff, 0xfa, 0xef, 0xeb, 0xff, 0xff, 0xff, 
	0xef, 0xff, 0xbf, 0x7e, 0xf7, 0x7b, 0xff, 0xdf, 
	0x7f, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x7f, 
	0xff, 0xfd, 0xff, 0x7f, 0xe2, 0xcf, 0xef, 0xff, 
	0xfc, 0xf2, 0xf7, 0x53, 0xb9, 0xa5, 0xff, 0xff, 
	0xee, 0xfb, 0xf9, 0x7f, 0xf7, 0xbf, 0xf7, 0xf9, 
	0x78, 0xed, 0xb5, 0xd7, 0x5e, 0x7f, 0xed, 0x3f, 
	0x69, 0x7f, 0xfd, 0xef, 0xd7, 0xff, 0xff, 0xfb, 
	0xf9, 0xbf, 0xff, 0xfe, 0xdb, 0xff, 0x73, 0xff, 
	0xed, 0xe7, 0xff, 0xaf, 0xfb, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0x7f, 0xa7, 0xff, 0xa5, 0xe6, 
	0x5f, 0xeb, 0xff, 0xa5, 0xa7, 0xf2, 0xf2, 0xfb, 
	0xe5, 0xbf, 0xff, 0x67, 0xde, 0xd3, 0x7e, 0xfd, 
	0xff, 0xfe, 0xfb, 0xef, 0xbf, 0xff, 0xfe, 0xff, 
	0x8f, 0xd6, 0xbf, 0xfe, 0xff, 0xfb, 0xff, 0x3f, 
	0xce, 0xfe, 0xd3, 0xac, 0xb7, 0xff, 0x7f, 0x97, 
	0xdb, 0x7f, 0xfd, 0xfd, 0xff, 0x4b, 0x2f, 0xbd, 
	0xf4, 0xd2, 0x4b, 0xff, 0x75, 0xb9, 0xff, 0xdf, 
	0x6b, 0xa6, 0xfd, 0xf7, 0xb2, 0x79, 0x37, 0xbf, 
	0x1e, 0xdf, 0x6b, 0x71, 0x13, 0xef, 0x9d, 0xf2, 
	0xea, 0x7f, 0xcd, 0x34, 0xd3, 0x4c, 0x2b, 0xcd, 
	0xee, 0x65, 0xfe, 0xff, 0xbf, 0x95, 0x5a, 0xff, 
	0xbb, 0xff, 0x8e, 0x7b, 0xea, 0x7d, 0xef, 0x55, 
	0xad, 0xbc, 0xd3, 0xea, 0xab, 0xff, 0xbf, 0xff, 
	0xfe, 0xfb, 0xef, 0xbe, 0x3f, 0x07, 0xfb, 0xaf, 
	0xf6, 0xff, 0xfb, 0xbf, 0xaf, 0x97, 0xbe, 0xfa, 
	0xea, 0xab, 0xbd, 0xde, 0xf7, 0xdf, 0xd7, 0x6f, 
	0xbd, 0xff, 0xff, 0xff, 0xff, 0xff, 0xbf, 0xfd, 
	0xff, 0xc6, 0xee, 0xbf, 0x16, 0xff, 0x3b, 0xfc, 
	0xbf, 0xfd, 0x0e, 0xcb, 0xaf, 0xbf, 0xf6, 0x1e, 
	0xd6, 0xdf, 0x7b, 0x1d, 0x3d, 0xfc, 0x5f, 0x7f, 
	0xfd, 0xf5, 0xd7, 0x5f, 0xff, 0x7b, 0x7a, 0xff, 
	0xb7, 0xfd, 0x7f, 0xdb, 0xfd, 0xff, 0xdf, 0xfe, 
	0xf3, 0xff, 0x3f, 0xff, 0xec, 0xb3, 0xff, 0x3f, 
	0xfe, 0xf9, 0x7f, 0xdf, 0x7d, 0xf7, 0xdd, 0x77, 
	0xdf, 0xf7, 0xea, 0xfd, 0xbf, 0xff, 0xe7, 0x99, 
	0xe7, 0xff, 0x7d, 0xc6, 0x9f, 0x7f, 0xfe, 0xff, 
	0xf7, 0xbf, 0xff, 0xff, 0xfb, 0x6f, 0xff, 0x7d, 
	0xf6, 0xd9, 0x67, 0x9f, 0x7d, 0xde, 0x3b, 0xfb, 
	0xbf, 0xdf, 0x7e, 0xf3, 0x8f, 0x3f, 0xf9, 0x7c, 
	0xd3, 0xef, 0xbf, 0xdf, 0x4e, 0xf7, 0xd8, 0x6f, 
	0xa7, 0xdd, 0xf7, 0xd3, 0x5f, 0x7f, 0xfd, 0xf5, 
	0xd7, 0xff, 0xc6, 0xe6, 0xdf, 0x3f, 0xff, 0xfc, 
	0xff, 0xdf, 0xff, 0xff, 0xfc, 0xf7, 0xff, 0xef, 
	0xff, 0xfb, 0xee, 0x9f, 0xef, 0xfe, 0xff, 0x47, 
	0x7e, 0xf9, 0xe4, 0x93, 0x4f, 0xfe, 0xb3, 0xbb, 
	0xef, 0xbf, 0xff, 0xfb, 0xfe, 0xdf, 0xff, 0xbf, 
	0xfb, 0xef, 0xff, 0xff, 0xff, 0xff, 0xef, 0xdf, 
	0xff, 0xfd, 0xff, 0xdf, 0xfb, 0xef, 0xbd, 0xf7, 
	0xf6, 0xfb, 0xef, 0xa6, 0xfe, 0x0f, 0xef, 0xbf, 
	0xd0, 0xe7, 0x8b, 0x2f, 0xbc, 0x7c, 0xfa, 0x6b, 
	0xbf, 0x7d, 0xfd, 0xb7, 0xd7, 0x7f, 0x7b, 0xfd, 
	0xf7, 0xdf, 0x7f, 0xff, 0xfd, 0xf7, 0xbf, 0xb2, 
	0xf7, 0x77, 0xf2, 0xfb, 0xe6, 0xbf, 0xaf, 0xbf, 
	0x78, 0xfe, 0xcb, 0xaf, 0xfe, 0xda, 0xf5, 0x57, 
	0x5f, 0x7f, 0xf7, 0xe5, 0x77, 0xdf, 0x7d, 0xf7, 
	0xdd, 0x77, 0x7f, 0xff, 0xfe, 0xff, 0xeb, 0x7f, 
	0xcf, 0xff, 0xf7, 0xfb, 0xcb, 0xef, 0xbf, 0xff, 
	0xfe, 0xfb, 0xff, 0x7f, 0xff, 0xfd, 0x75, 0xd7, 
	0xdf, 0x7f, 0xff, 0xfd, 0xf7, 0xdf, 0xff, 0xff, 
	0xf8, 0xff, 0xff, 0xec, 0xf3, 0xcf, 0x3d, 0xff, 
	0x6c, 0xf3, 0x8f, 0x3e, 0xfb, 0xf8, 0xf3, 0x1f, 
	0x7d, 0xf4, 0xd1, 0x47, 0x9f, 0x7d, 0xf6, 0xd9, 
	0x67, 0x9f, 0x7d, 0xfe, 0xee, 0xff, 0xdf, 0xff, 
	0xbf, 0xf8, 0xf7, 0x8f, 0x7f, 0xbf, 0xfe, 0xfb, 
	0xfb, 0xbf, 0xff, 0xfd, 0xff, 0xff, 0x7f, 0xff, 
	0xff, 0xfd, 0xf7, 0xdf, 0x7f, 0xff, 0xfd, 0xfb, 
	0x83, 0xff, 0x7f, 0xfd, 0xf6, 0xde, 0xbf, 0xef, 
	0xff, 0xf9, 0xfe, 0xef, 0xef, 0xff, 0xfe, 0xf4, 
	0xdf, 0xff, 0x7f, 0xff, 0xff, 0xdf, 0x7e, 0xfb, 
	0xed, 0xb7, 0xdf, 0xfe, 0x9f, 0xfe, 0xff, 0xff, 
	0xff, 0xef, 0x7f, 0xff, 0xfe, 0xfb, 0xef, 0xbf, 
	0xff, 0xff, 0xfb, 0xdf, 0xff, 0xff, 0xfd, 0xff, 
	0xff, 0x7f, 0xff, 0xfd, 0xf7, 0xdf, 0x7f, 0xff, 
	0x7f, 0xfb, 0xff, 0xff, 0xef, 0xff, 0xff, 0xff, 
	0xf6, 0xef, 0xff, 0xbf, 0xfe, 0xdf, 0x7b, 0x5f, 
	0xff, 0xed, 0xf5, 0xff, 0xfe, 0x7e, 0xff, 0xfd, 
	0xf7, 0xdf, 0x7f, 0xbf, 0xff, 0xe3, 0xff, 0xff, 
	0xef, 0xff, 0xff, 0x7f, 0xdb, 0xef, 0xfe, 0xfc, 
	0xf6, 0xdb, 0x6f, 0xbf, 0xfb, 0xed, 0x2f, 0xdb, 
	0xfe, 0xef, 0xfd, 0xf5, 0xdf, 0x7f, 0xff, 0xfd, 
	0xff, 0x8f, 0xfe, 0xff, 0x72, 0xeb, 0xaf, 0xbf, 
	0xdc, 0xfa, 0xab, 0xaf, 0xb5, 0xea, 0xfa, 0x8b, 
	0x5f, 0x77, 0xfd, 0x71, 0xd7, 0x5f, 0x77, 0xfd, 
	0x75, 0xd7, 0x5d, 0x77, 0xfd, 0x7b, 0xfe, 0xff, 
	0xff, 0xff, 0xff, 0x3c, 0xbe, 0xff, 0xfd, 0xcf, 
	0xea, 0xaf, 0xff, 0xec, 0xf7, 0xdf, 0xf7, 0xf7, 
	0xfd, 0x7f, 0xef, 0xff, 0xde, 0x7f, 0xff, 0xff, 
	0xff, 0xff, 0xf2, 0xbf, 0x6f, 0x7d, 0xff, 0xbf, 
	0x7f, 0xff, 0xfe, 0xdf, 0xef, 0xf7, 0x7f, 0xff, 
	0xf7, 0xff, 0xff, 0xff, 0xfd, 0xff, 0xff, 0xff, 
	0xfe, 0xfb, 0xef, 0xbf, 0xff, 0x7f, 0xab, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xfd, 0x7f, 0xf7, 0xfc, 
	0xfe, 0xff, 0xff, 0xff, 0xff, 0xdf, 0xdf, 0xfb, 
	0xeb, 0xff, 0xff, 0xbf, 0xff, 0xfd, 0xfb, 0xff, 
	0x7f, 0xff, 0x97, 0xff, 0x2f, 0xbf, 0xef, 0xff, 
	0xff, 0xad, 0xbf, 0xfe, 0xff, 0xef, 0x2f, 0xff, 
	0xfd, 0x77, 0xf7, 0x5f, 0x7f, 0xff, 0xf5, 0xe7, 
	0x97, 0x6f, 0xfe, 0xf9, 0xe7, 0xfb, 0xaf, 0xfe, 
	0xff, 0xed, 0xfb, 0xfe, 0xfb, 0xff, 0xfe, 0xeb, 
	0xff, 0xeb, 0xff, 0xfe, 0xff, 0xf3, 0xfb, 0xfe, 
	0x7b, 0xef, 0xff, 0xff, 0xff, 0xef, 0xff, 0xff, 
	0xff, 0xdb, 0x3f, 0xfb, 0xff, 0xfe, 0xff, 0x9d, 
	0xdb, 0xee, 0x78, 0xeb, 0xfd, 0xff, 0xf6, 0xf8, 
	0xbf, 0x7f, 0xf7, 0xa9, 0xe5, 0x8f, 0xff, 0xff, 
	0xaf, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xe5, 
	0xff, 0xeb, 0x7f, 0xff, 0xe9, 0xff, 0xdb, 0x87, 
	0xff, 0xbd, 0xff, 0xff, 0x7f, 0x7f, 0xfe, 0xf5, 
	0xf7, 0xdf, 0xf7, 0x7f, 0xff, 0xff, 0xff, 0xfb, 
	0xff, 0xff, 0xff, 0x8f, 0xff, 0xff, 0xfe, 0xff, 
	0xff, 0xdf, 0xbf, 0xff, 0xdf, 0xff, 0x5f, 0xfd, 
	0xff, 0x7b, 0xfc, 0xbf, 0xdf, 0x7e, 0x77, 0xbf, 
	0xfe, 0xff, 0xbf, 0xff, 0xbf, 0xff, 0xff, 0x4f, 
	0xfe, 0xff, 0xff, 0xfc, 0xff, 0xff, 0x3f, 0xfb, 
	0xef, 0xff, 0xdf, 0xff, 0xff, 0xd8, 0xff, 0x1d, 
	0xfb, 0xff, 0xff, 0xfe, 0xff, 0x7f, 0xfc, 0xff, 
	0xff, 0xbf, 0xff, 0xff, 0xfb, 0xff, 0xef, 0xe9, 
	0xff, 0xfc, 0xff, 0xfa, 0xff, 0xff, 0x7f, 0xff, 
	0xdf, 0xef, 0xff, 0xff, 0xdd, 0xff, 0xff, 0xff, 
	0xfd, 0xff, 0xf7, 0xff, 0xff, 0xff, 0xfe, 0xff, 
	0xe6, 0xff, 0xff, 0xbf, 0xff, 0xff, 0xb7, 0xcd, 
	0x3b, 0xff, 0x7f, 0xf7, 0xcb, 0xff, 0x7f, 0xfd, 
	0xee, 0xd9, 0x5f, 0xbf, 0xff, 0xef, 0xbf, 0xff, 
	0xff, 0xff, 0xef, 0xff, 0x33, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0x7f, 0x9f, 0x77, 0xf4, 0xff, 0xff, 
	0x3f, 0xdf, 0xff, 0xbf, 0xbf, 0x1e, 0xff, 0xf9, 
	0xff, 0xbf, 0xff, 0xbe, 0xfb, 0xee, 0xbf, 0xff, 
	0x97, 0xfe, 0xff, 0xfb, 0xfe, 0xff, 0xff, 0xff, 
	0xfe, 0xfb, 0xff, 0xff, 0x7f, 0xff, 0xff, 0xfd, 
	0xfe, 0xff, 0xbf, 0xf7, 0xff, 0xfb, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0x7f, 0xf3, 0xff, 0x7e, 
	0xfe, 0xf9, 0xd5, 0xfb, 0x77, 0xbe, 0xff, 0xf2, 
	0xff, 0x7b, 0xbe, 0x75, 0xc8, 0xff, 0xf7, 0xf4, 
	0xff, 0xaf, 0x97, 0xdf, 0x7a, 0xff, 0xe7, 0xf7, 
	0x3f, 0x6c, 0xff, 0xff, 0xff, 0xff, 0xcf, 0xff, 
	0x7f, 0xfc, 0xd6, 0xcb, 0xde, 0xfb, 0xff, 0x73, 
	0xa7, 0xfc, 0xf7, 0xf3, 0xff, 0xfe, 0xdf, 0xfa, 
	0xeb, 0xfd, 0x9f, 0xef, 0x7f, 0x87, 0xff, 0xe9, 
	0xd7, 0xff, 0x7e, 0x9d, 0xa5, 0xa7, 0xff, 0xff, 
	0xff, 0x2c, 0xff, 0xfb, 0xfd, 0xff, 0x59, 0x7e, 
	0xfb, 0xff, 0xff, 0xff, 0xfd, 0xff, 0xff, 0xff, 
	0xff, 0xcf, 0xd2, 0xff, 0xff, 0xff, 0xef, 0xe9, 
	0xa7, 0xfe, 0xfb, 0xff, 0xef, 0xa7, 0xfe, 0x7f, 
	0xf6, 0xdf, 0x6f, 0xbd, 0xe7, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xb5, 0x9b, 0x5f, 
	0xde, 0x7d, 0xbf, 0xbd, 0xf2, 0xf3, 0x6f, 0xf7, 
	0xfd, 0xf6, 0xce, 0x7d, 0x7f, 0x3f, 0x6f, 0x9c, 
	0xd1, 0xed, 0xfb, 0xef, 0xbf, 0xff, 0xfe, 0xff, 
	0xef, 0xef, 0xe2, 0xfe, 0x7f, 0xf3, 0xbe, 0xfb, 
	0xda, 0xf9, 0xe6, 0xce, 0xff, 0xea, 0x7b, 0xf3, 
	0x7f, 0x57, 0xd4, 0xf1, 0xc6, 0xaf, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0x3f, 0x87, 0xfb, 
	0xff, 0xad, 0xcf, 0x6a, 0xeb, 0xb1, 0xb1, 0xbe, 
	0xff, 0xaf, 0xed, 0xe7, 0xdf, 0xf5, 0xf7, 0xc3, 
	0x4f, 0xfd, 0xff, 0x7f, 0xfb, 0xed, 0xe7, 0xff, 
	0x7f, 0xeb, 0xa6, 0xe6, 0xff, 0xfe, 0x4e, 0x69, 
	0xb8, 0xbd, 0xcd, 0x4a, 0x3d, 0xa5, 0xa1, 0xbf, 
	0x4f, 0xd4, 0x5f, 0x61, 0x3f, 0xf5, 0x7f, 0xeb, 
	0xcb, 0x2f, 0xbf, 0xff, 0xff, 0xab, 0x3b, 0x50, 
	0xff, 0xbf, 0xdf, 0x7f, 0xdb, 0x67, 0xff, 0xdf, 
	0x7e, 0xfb, 0xcd, 0x9f, 0xdf, 0xef, 0xe7, 0xdb, 
	0x3f, 0xfc, 0xfe, 0xfe, 0x7b, 0xef, 0xbd, 0xff, 
	0xff, 0xfb, 0xf7, 0xe5, 0xfd, 0xff, 0x71, 0xfe, 
	0xd9, 0x7f, 0xfe, 0x7d, 0xc6, 0x59, 0x67, 0xfe, 
	0x71, 0xbe, 0xff, 0xcf, 0xf8, 0xfb, 0xed, 0xf3, 
	0xcf, 0x3f, 0xff, 0xfc, 0xff, 0xcf, 0xdf, 0x2f, 
	0xfb, 0xb7, 0xff, 0x74, 0xc3, 0xec, 0x3f, 0xdb, 
	0xfc, 0xb3, 0x8d, 0xbf, 0xff, 0x7c, 0xf7, 0x9b, 
	0x6f, 0xb7, 0xdd, 0xe7, 0x1f, 0x7f, 0xfc, 0xf1, 
	0xf7, 0x1f, 0xff, 0xb6, 0xee, 0xcf, 0x7f, 0xff, 
	0xfc, 0xfb, 0xfd, 0x3f, 0xff, 0xff, 0xf3, 0xdf, 
	0xf7, 0xff, 0xfc, 0xf7, 0x9f, 0x6f, 0xbe, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xf3, 
	0xba, 0xef, 0xbf, 0xbf, 0xfb, 0xff, 0xff, 0xff, 
	0xbe, 0xff, 0xee, 0xbf, 0xff, 0xbf, 0xfe, 0xff, 
	0xd7, 0x7f, 0xff, 0xff, 0xfd, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xef, 0xeb, 0x7e, 0xff, 0x3f, 
	0x9c, 0xd4, 0xde, 0x8b, 0x3f, 0xbc, 0x70, 0xc2, 
	0xeb, 0x7f, 0xfc, 0xfd, 0x87, 0xd6, 0x5f, 0xff, 
	0xfb, 0x8f, 0x3f, 0xfe, 0xf8, 0xbf, 0x8f, 0xbf, 
	0xa6, 0xf7, 0xff, 0xde, 0xf9, 0xe2, 0xaf, 0xaf, 
	0xb7, 0xf9, 0xe2, 0x8b, 0xaf, 0xfe, 0xfb, 0xe5, 
	0x33, 0x5f, 0x6f, 0xdd, 0xdf, 0x7f, 0xff, 0xfd, 
	0xf7, 0xff, 0x7f, 0x7f, 0x8f, 0xfe, 0xbf, 0xff, 
	0x3b, 0x6f, 0xaf, 0xbb, 0x5d, 0xdb, 0xff, 0xfc, 
	0xf5, 0xfe, 0xcf, 0xdd, 0x7f, 0xfb, 0xfd, 0xf7, 
	0xff, 0xf9, 0xfd, 0xf7, 0xdf, 0xff, 0xff, 0xff, 
	0xff, 0xfb, 0xff, 0xfa, 0xdc, 0xe3, 0x8d, 0x3e, 
	0xfe, 0x4c, 0x73, 0xcf, 0x35, 0xfa, 0x7c, 0x53, 
	0x9f, 0x69, 0xf4, 0xd1, 0xe7, 0x9f, 0x7f, 0xfe, 
	0xf9, 0xe7, 0x9f, 0x7b, 0xfe, 0xe0, 0xff, 0xff, 
	0x3f, 0xff, 0xff, 0xfb, 0xef, 0x3f, 0xbe, 0xf8, 
	0xf3, 0xfb, 0x7f, 0x3f, 0xfd, 0xe7, 0xff, 0xff, 
	0xff, 0xf3, 0xff, 0xbf, 0xfe, 0xfb, 0xff, 0xcf, 
	0xff, 0xa7, 0xff, 0xff, 0xfd, 0xf5, 0xff, 0xbb, 
	0xbf, 0xbf, 0xff, 0xfa, 0xdb, 0xef, 0xff, 0xfd, 
	0xf5, 0xb7, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xfe, 0xff, 0xff, 0xff, 0xff, 0x23, 0xfe, 0xdf, 
	0xfe, 0xfb, 0xef, 0xff, 0xff, 0xff, 0xeb, 0xbf, 
	0xbf, 0xfe, 0xff, 0xef, 0xff, 0x7f, 0xfd, 0xfd, 
	0xf7, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xfb, 0xff, 0xfb, 0xff, 0xff, 0xff, 0xef, 0xfd, 
	0xff, 0xff, 0xef, 0xff, 0xbf, 0xff, 0xee, 0xff, 
	0xef, 0xfe, 0xfd, 0xf5, 0xd7, 0xff, 0xfb, 0xf7, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xe4, 0x7f, 
	0xd3, 0x4f, 0xff, 0xff, 0xd6, 0xff, 0xff, 0xff, 
	0xff, 0xf4, 0x4b, 0xff, 0x7d, 0xff, 0xe5, 0x37, 
	0x9f, 0xfe, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xbf, 0xff, 0xdf, 0xba, 0xab, 0xaf, 
	0xbf, 0xfa, 0xfa, 0xeb, 0xaf, 0xb7, 0xde, 0xfa, 
	0xeb, 0x5b, 0x6f, 0xdd, 0xf5, 0x16, 0x5f, 0x7f, 
	0xfd, 0xf1, 0xc3, 0x5f, 0x7f, 0xfd, 0x2f, 0xfe, 
	0xff, 0xff, 0xbf, 0xcf, 0xce, 0xf6, 0xff, 0x73, 
	0xdf, 0xfe, 0xfb, 0xbf, 0xef, 0x9f, 0xfe, 0x75, 
	0x7f, 0xff, 0x9b, 0x6f, 0xff, 0xf7, 0xd7, 0xff, 
	0xff, 0xff, 0x77, 0xf2, 0xbf, 0xdf, 0xfd, 0x97, 
	0xff, 0xff, 0xff, 0xfd, 0xdd, 0xef, 0x7f, 0xff, 
	0xfe, 0x5f, 0xfe, 0xbf, 0xff, 0xfb, 0x3f, 0xff, 
	0xbf, 0xff, 0xfe, 0xff, 0xff, 0xff, 0x7f, 0xed, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xfd, 0xaf, 
	0xdf, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xfb, 0xb7, 0xff, 0x7e, 0xff, 
	0xff, 0xff, 0xdf, 0xbf, 0xff, 0xbf, 0xff, 0xff, 
	0x7a, 0xef, 0xbf, 0xdf, 0xff, 0xff, 0xcf, 0x2e, 
	0xfe, 0xff, 0xf7, 0xbd, 0x1b, 0x7f, 0xbd, 0xff, 
	0xff, 0xfb, 0xff, 0xff, 0xff, 0xff, 0xff, 0x8f, 
	0xfc, 0xbf, 0xff, 0xff, 0xe6, 0xff, 0xbf, 0xff, 
	0xff, 0xff, 0xdb, 0xfe, 0xfe, 0xff, 0xf3, 0xff, 
	0xdd, 0xfe, 0xf3, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xbf, 0x7b, 0xff, 0xbb, 0xff, 
	0xff, 0xad, 0xff, 0xbf, 0xfd, 0xfe, 0xfb, 0xff, 
	0xbe, 0x7f, 0xfb, 0xf7, 0xdf, 0xf7, 0x9e, 0xfe, 
	0xfb, 0xdd, 0x7f, 0xff, 0xff, 0xfe, 0x7f, 0xff, 
	0xc3, 0xff, 0xff, 0xff, 0xff, 0xff, 0xdf, 0xbf, 
	0xfb, 0xdf, 0xfd, 0xf7, 0xdf, 0xff, 0xff, 0xdb, 
	0xcf, 0xe7, 0xbe, 0xff, 0xff, 0xfb, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0x7f, 0x8b, 0xbd, 0x7f, 0xef, 
	0xff, 0xff, 0xbf, 0xff, 0xdf, 0xff, 0xdf, 0xff, 
	0xf6, 0xfe, 0xff, 0xff, 0xff, 0xcf, 0xff, 0xff, 
	0x9f, 0xff, 0xff, 0xff, 0xff, 0xff, 0x7f, 0xff, 
	0xfe, 0xfe, 0xff, 0x5f, 0xff, 0xf3, 0xcf, 0x3f, 
	0xff, 0xfc, 0xfb, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xfa, 0xff, 0xff, 0xf0, 0xfb, 0xff, 0xff, 0xfe, 
	0xfb, 0xe7, 0xff, 0xff, 0xff, 0xf9, 0xff, 0xff, 
	0xff, 0x6f, 0xbd, 0xfe, 0xf6, 0xdf, 0xef, 0xfe, 
	0xff, 0xff, 0x7f, 0xff, 0xff, 0xef, 0xff, 0xdf, 
	0xff, 0xff, 0xff, 0xf7, 0xdd, 0x77, 0xff, 0xff, 
	0xff, 0x60, 0xff, 0xdf, 0xff, 0xff, 0x7d, 0xfb, 
	0xdd, 0xdf, 0xfe, 0x5f, 0xf7, 0xef, 0xf2, 0xfb, 
	0xdf, 0x75, 0xd6, 0x7f, 0xbf, 0xff, 0xf5, 0xdf, 
	0xfb, 0xde, 0xfa, 0xff, 0xff, 0x1b, 0xfd, 0xff, 
	0xff, 0xfb, 0xff, 0x6f, 0xff, 0xfd, 0xff, 0xff, 
	0xbf, 0xb7, 0xff, 0xff, 0xff, 0xfb, 0xef, 0xf9, 
	0xfc, 0xef, 0xdc, 0x1e, 0xfb, 0xf7, 0xff, 0xff, 
	0xff, 0x57, 0xfe, 0xff, 0xfb, 0xef, 0xfd, 0xcf, 
	0xff, 0xff, 0xff, 0xbf, 0xfe, 0xbf, 0xfe, 0xff, 
	0x7f, 0xff, 0xff, 0xff, 0xfb, 0xff, 0x7f, 0xfb, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xe3, 0xff, 
	0x5b, 0xb9, 0xff, 0xd7, 0xfb, 0x7f, 0xbd, 0xf9, 
	0xe4, 0xfb, 0xff, 0x7f, 0xfc, 0xcf, 0x87, 0xf7, 
	0x5f, 0xfe, 0x7d, 0xef, 0x5f, 0x5e, 0xff, 0xad, 
	0xf3, 0x3f, 0x63, 0xff, 0xfb, 0xfc, 0xff, 0xcf, 
	0xff, 0xff, 0x79, 0xff, 0x9f, 0x7e, 0xfd, 0xff, 
	0xf3, 0xff, 0xdd, 0xf2, 0xdf, 0xfe, 0xf7, 0xfd, 
	0xbf, 0xd3, 0xce, 0x9f, 0xff, 0x7f, 0xab, 0xff, 
	0x39, 0xbf, 0xf8, 0xff, 0x9f, 0xff, 0xff, 0xff, 
	0x6f, 0xb9, 0xbf, 0xff, 0xcf, 0x75, 0xde, 0x7e, 
	0xff, 0xfd, 0xff, 0xf2, 0xff, 0xff, 0xed, 0xdf, 
	0x7f, 0xff, 0x7f, 0xd6, 0xa7, 0x9b, 0xfe, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xa9, 0xb3, 0xff, 
	0x3f, 0xd7, 0x5f, 0x7f, 0xff, 0xfd, 0xff, 0xdb, 
	0xff, 0xff, 0xfd, 0xff, 0xff, 0xff, 0x35, 0xb8, 
	0xff, 0xbb, 0x99, 0xf7, 0xfd, 0xf1, 0xdf, 0x7f, 
	0xff, 0xfd, 0x77, 0xde, 0x7d, 0xff, 0xfb, 0xef, 
	0xbd, 0xf3, 0xed, 0xdb, 0xef, 0xbf, 0xdf, 0xee, 
	0xfb, 0xff, 0xef, 0x6e, 0xfe, 0xfb, 0xf7, 0xce, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xee, 0xac, 
	0xf3, 0xff, 0xf7, 0x5d, 0x73, 0xc3, 0xff, 0xff, 
	0xbf, 0xff, 0xdf, 0x9d, 0xff, 0xff, 0x3f, 0x2b, 
	0xfb, 0xaf, 0x87, 0x8e, 0x7f, 0xfd, 0xf5, 0xd7, 
	0xfb, 0xcf, 0xbb, 0xab, 0xc6, 0xff, 0x95, 0x5e, 
	0xdb, 0x7d, 0xff, 0xdf, 0xd7, 0xfb, 0xe5, 0x35, 
	0xdf, 0xff, 0xff, 0xe6, 0xee, 0xff, 0xc7, 0x36, 
	0xbf, 0x7d, 0xfc, 0xc3, 0xdf, 0x7f, 0xac, 0xb1, 
	0x56, 0xef, 0xd7, 0x52, 0x4b, 0x8f, 0xbd, 0xff, 
	0x58, 0xc3, 0xaf, 0x3d, 0xfc, 0xfe, 0xff, 0x7b, 
	0x79, 0xff, 0x1f, 0xff, 0x7c, 0xef, 0xf7, 0xf7, 
	0xdf, 0x7f, 0xff, 0xed, 0xb7, 0xdf, 0xef, 0xde, 
	0x7b, 0x6f, 0xfe, 0xed, 0xfe, 0xdf, 0xef, 0xbd, 
	0xff, 0xfe, 0xff, 0xf7, 0xea, 0xfd, 0xbf, 0x7f, 
	0xe6, 0xf9, 0x77, 0x9f, 0x7f, 0xfe, 0xf9, 0x67, 
	0x9c, 0x71, 0xbe, 0xf3, 0xcf, 0x3f, 0xf3, 0xfd, 
	0xf3, 0xff, 0x3b, 0xff, 0xfc, 0xf3, 0xff, 0xdf, 
	0x3e, 0xfb, 0xbf, 0xff, 0xf8, 0xf3, 0xcd, 0x3b, 
	0xff, 0xf8, 0x63, 0x8d, 0x3f, 0xfe, 0xfc, 0x87, 
	0x1f, 0x6f, 0xfc, 0xfd, 0xc7, 0xdf, 0x7f, 0xfc, 
	0xf1, 0xc7, 0xdf, 0xff, 0x56, 0xe6, 0xff, 0x77, 
	0xdf, 0xff, 0xfb, 0xff, 0xff, 0xff, 0xff, 0xf7, 
	0xff, 0xff, 0xff, 0xff, 0xef, 0xb7, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0x33, 0xbb, 0xff, 0x7f, 0xff, 0xff, 0xfe, 0xff, 
	0xef, 0xff, 0xff, 0xef, 0xff, 0xff, 0xbf, 0xff, 
	0xf7, 0xdf, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xef, 0xa9, 0xfe, 0x7b, 
	0xbf, 0xf5, 0xf9, 0xeb, 0x5f, 0x7f, 0xfd, 0xf1, 
	0xd3, 0x0f, 0x7f, 0xff, 0xe3, 0x85, 0x97, 0x5a, 
	0xfb, 0xe3, 0xb5, 0x1f, 0xfa, 0xe8, 0xa3, 0xff, 
	0xbf, 0x9a, 0xf7, 0xbf, 0xfa, 0xfb, 0xef, 0xbf, 
	0xff, 0xf6, 0xfb, 0xef, 0x99, 0x37, 0xfe, 0xfb, 
	0xdf, 0x33, 0xdf, 0x7c, 0xd7, 0xdf, 0x57, 0x6f, 
	0xfc, 0xf7, 0xdf, 0xff, 0x7f, 0x5f, 0xfe, 0xf7, 
	0xfe, 0xff, 0xbf, 0x7f, 0xff, 0xff, 0x9f, 0x3f, 
	0xaf, 0xbd, 0xf6, 0xff, 0xff, 0x5e, 0xb9, 0xed, 
	0xf7, 0x7f, 0x7e, 0xff, 0xdd, 0xdf, 0xbf, 0xff, 
	0xff, 0xbf, 0xf8, 0xff, 0xfe, 0xfc, 0xf3, 0xcf, 
	0x3f, 0xff, 0x7c, 0xf3, 0x8d, 0x36, 0xd2, 0xfc, 
	0xf3, 0x1a, 0x6f, 0xbe, 0xf9, 0xe7, 0x9b, 0x7d, 
	0xe6, 0xf9, 0xe7, 0x9f, 0x7f, 0xfe, 0xee, 0xff, 
	0xeb, 0xff, 0xff, 0xfd, 0xff, 0xdf, 0xff, 0xff, 
	0xfd, 0xe3, 0xcf, 0x7f, 0xbe, 0xfb, 0xf7, 0xd7, 
	0x5f, 0xff, 0xfb, 0xff, 0x1f, 0xff, 0xff, 0xfb, 
	0xef, 0xff, 0xb3, 0xff, 0xaf, 0xfe, 0xfb, 0xf7, 
	0xbf, 0x7f, 0xff, 0xff, 0xf7, 0xfb, 0xbf, 0xff, 
	0xff, 0xff, 0xff, 0xdf, 0x7f, 0xef, 0xef, 0xff, 
	0x7e, 0xff, 0xff, 0xff, 0xff, 0xff, 0xcf, 0xfe, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xfb, 0xef, 
	0xbf, 0xff, 0xfe, 0xf9, 0xff, 0xbf, 0x7f, 0xfd, 
	0xf5, 0xf7, 0xff, 0x7f, 0xff, 0xfd, 0xff, 0xbf, 
	0xfe, 0xfe, 0x7f, 0xf9, 0xff, 0xff, 0xff, 0xff, 
	0xfd, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0x7f, 0xff, 0xfb, 0xff, 0xff, 0xff, 0xff, 0xfb, 
	0xfd, 0xd7, 0x7f, 0xff, 0xff, 0xfb, 0xff, 0xe3, 
	0xff, 0xfb, 0x7f, 0xf7, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xf4, 0xff, 0xff, 0xbf, 0xff, 0xe9, 
	0x97, 0x7f, 0xfe, 0xff, 0xed, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xa3, 0xfe, 0xef, 0xf8, 0xe3, 
	0x8f, 0x3f, 0xf6, 0xea, 0xeb, 0xaf, 0x3a, 0xfe, 
	0xea, 0xcb, 0x4e, 0x6d, 0xb5, 0xc5, 0xd7, 0x5f, 
	0x7f, 0xed, 0xf5, 0xc7, 0x5f, 0x7f, 0xfd, 0x2b, 
	0xfe, 0xff, 0xfb, 0xf6, 0xbf, 0xf7, 0xff, 0xfd, 
	0xf3, 0xff, 0x3e, 0xfe, 0xeb, 0xb3, 0xff, 0xe1, 
	0xfe, 0xf9, 0xff, 0x1e, 0x9b, 0xfd, 0xff, 0xff, 
	0xff, 0x7f, 0xff, 0xbf, 0xf0, 0x7f, 0xbf, 0xf9, 
	0xfb, 0xff, 0xff, 0xfd, 0xe7, 0x7b, 0x6f, 0x7f, 
	0xff, 0xf5, 0xfb, 0x7f, 0xff, 0xf4, 0xff, 0xcd, 
	0xef, 0xff, 0x7f, 0xff, 0xff, 0xff, 0xff, 0x7f, 
	0xa7, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xdd, 0xff, 0xdf, 0xff, 0xff, 
	0xbf, 0xbf, 0xff, 0xff, 0xfd, 0xff, 0xff, 0xff, 
	0x9f, 0xff, 0xfb, 0xff, 0x8f, 0xff, 0xad, 0xbb, 
	0xfe, 0x7f, 0xfb, 0xff, 0xdf, 0xff, 0x7f, 0xfb, 
	0xbf, 0xe5, 0xff, 0xfe, 0xc7, 0xfb, 0xff, 0xff, 
	0xdf, 0xd7, 0xf9, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0x2f, 0xfe, 0xef, 0xfe, 0xba, 0xff, 0xbf, 0xfb, 
	0xef, 0xaf, 0xff, 0xbb, 0xfb, 0xff, 0xbf, 0xff, 
	0xbf, 0xdf, 0x7f, 0xff, 0xff, 0x7e, 0xff, 0xff, 
	0xfe, 0xfd, 0xff, 0xff, 0x3f, 0xf8, 0xff, 0xfe, 
	0xff, 0xbf, 0xfb, 0xfe, 0xfb, 0xff, 0xff, 0xfd, 
	0xff, 0xfe, 0xff, 0xef, 0xff, 0xdf, 0x3f, 0xff, 
	0x39, 0x7f, 0xdd, 0xf7, 0xff, 0xfd, 0xff, 0xff, 
	0xff, 0xe9, 0xff, 0xff, 0xff, 0xbf, 0xfd, 0xf8, 
	0xe3, 0xfe, 0xff, 0xff, 0xdf, 0xde, 0xff, 0xff, 
	0xfb, 0xff, 0xf7, 0xff, 0x7b, 0xf7, 0xf9, 0xf7, 
	0xff, 0xfe, 0xff, 0xff, 0xff, 0x97, 0xff, 0xfb, 
	0xdf, 0xff, 0x73, 0xff, 0xff, 0xf7, 0xff, 0xff, 
	0x7f, 0xff, 0xff, 0xfb, 0xff, 0xff, 0xcf, 0xfd, 
	0xff, 0xb7, 0xde, 0xdb, 0xfe, 0xff, 0xdf, 0xff, 
	0xff, 0x0f, 0xfe, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0x3f, 0xfe, 0xfc, 0xff, 0xe7, 0x1f, 0x7f, 0xb8, 
	0xfd, 0xef, 0x7f, 0xfc, 0xb9, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xf8, 0xdf, 
	0xff, 0xff, 0xff, 0xff, 0xdf, 0xfb, 0xdb, 0xff, 
	0xbf, 0xdf, 0xfb, 0xdf, 0xff, 0xff, 0xef, 0xef, 
	0xb7, 0xfd, 0xff, 0xff, 0xff, 0xff, 0xfd, 0xff, 
	0xff, 0xff, 0xec, 0xff, 0xdf, 0xac, 0xbf, 0x7e, 
	0xfb, 0xdd, 0x7d, 0xfb, 0xed, 0xbf, 0xf7, 0xf7, 
	0xfb, 0x7f, 0xf7, 0xdf, 0xff, 0xbf, 0xfd, 0xe5, 
	0xfb, 0xff, 0xff, 0xff, 0xff, 0xbf, 0x1b, 0xff, 
	0xff, 0xff, 0xfe, 0xfb, 0xef, 0xff, 0xff, 0x7d, 
	0xff, 0xff, 0xff, 0xfe, 0xfd, 0xff, 0x8f, 0x3f, 
	0xbe, 0xff, 0xf3, 0xde, 0x7f, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xc6, 0xfe, 0xff, 0xff, 0xab, 0xf7, 
	0xcf, 0xff, 0xfb, 0xaf, 0xbf, 0xff, 0xff, 0xff, 
	0xff, 0x7f, 0x7f, 0xff, 0xfd, 0xff, 0xdf, 0x6f, 
	0xff, 0xb6, 0xff, 0xff, 0xff, 0xff, 0x3f, 0xe3, 
	0xff, 0x5a, 0x9e, 0xbf, 0xbf, 0xff, 0x6f, 0xbd, 
	0xff, 0xff, 0xf3, 0xfb, 0xbf, 0x75, 0xcb, 0xe7, 
	0xff, 0x77, 0xdf, 0x4f, 0xfd, 0xb7, 0x5f, 0xfe, 
	0xfe, 0xff, 0x3f, 0x69, 0xff, 0xff, 0xfc, 0xff, 
	0xff, 0xff, 0xdb, 0xfc, 0xff, 0x5f, 0xf7, 0xfb, 
	0x7f, 0xe7, 0xff, 0xff, 0xff, 0xcb, 0x7f, 0x3f, 
	0xfd, 0xf7, 0xff, 0xfd, 0xff, 0xff, 0x7f, 0xa7, 
	0xff, 0xaf, 0xff, 0x5b, 0x7f, 0xfd, 0xff, 0xff, 
	0xfb, 0xff, 0xff, 0xbf, 0xff, 0xff, 0xf5, 0x7f, 
	0xdf, 0xaf, 0xe5, 0xf7, 0x94, 0x59, 0x7e, 0xfd, 
	0xff, 0xff, 0xff, 0x1f, 0xd6, 0xff, 0xff, 0x3f, 
	0xff, 0xfc, 0xbf, 0xff, 0xff, 0xff, 0xff, 0xbf, 
	0xff, 0xff, 0xf7, 0xff, 0xe7, 0x3d, 0xfd, 0xf6, 
	0xd3, 0x7e, 0x95, 0xfd, 0xff, 0xff, 0xff, 0xf5, 
	0xb9, 0xff, 0x9a, 0xfd, 0x2f, 0xbf, 0x6c, 0x9e, 
	0x7d, 0xff, 0xfd, 0xf7, 0x5b, 0x7d, 0x7f, 0xb3, 
	0xef, 0xb7, 0xf3, 0xd6, 0xbf, 0xaf, 0xfc, 0xdb, 
	0xed, 0xff, 0xff, 0xef, 0x61, 0xfe, 0x7f, 0xf5, 
	0xbf, 0x7d, 0xb6, 0xa9, 0xf7, 0xff, 0xff, 0xff, 
	0xbb, 0xf6, 0xff, 0xf7, 0xfe, 0x77, 0xc7, 0x35, 
	0x37, 0xbe, 0xf2, 0xe7, 0xff, 0xff, 0xff, 0x3f, 
	0x1b, 0xfb, 0xaf, 0xd7, 0xff, 0xff, 0xfe, 0xe5, 
	0xe6, 0xff, 0xef, 0xbf, 0xbf, 0xfc, 0xff, 0xf5, 
	0x7c, 0x5b, 0xcd, 0xa5, 0xf5, 0xd7, 0xdf, 0x5f, 
	0xfd, 0xff, 0xfa, 0xff, 0x56, 0xee, 0xff, 0xc7, 
	0xff, 0xdf, 0x7f, 0xbf, 0x86, 0xcf, 0xff, 0xfc, 
	0xbd, 0xff, 0xcf, 0x77, 0xf0, 0x43, 0x0d, 0xbd, 
	0xd4, 0x5f, 0x7f, 0xff, 0xbd, 0xff, 0xf9, 0xff, 
	0x3b, 0x79, 0xff, 0xdf, 0xff, 0x7f, 0xdb, 0xcd, 
	0xf7, 0xdf, 0x7f, 0xff, 0xfd, 0xf7, 0xdf, 0xef, 
	0xe6, 0xdb, 0xef, 0xbf, 0xf7, 0xff, 0xcf, 0x3f, 
	0xff, 0xfd, 0xff, 0xff, 0xf7, 0xe3, 0xfd, 0xff, 
	0xff, 0xff, 0xdf, 0x7f, 0x9f, 0x7f, 0xfe, 0xf9, 
	0xe7, 0x9f, 0x7f, 0xbe, 0xf3, 0xcf, 0x38, 0xff, 
	0xfc, 0x3f, 0xff, 0x78, 0xeb, 0xbd, 0xff, 0xff, 
	0xdf, 0x2f, 0xfb, 0xb7, 0xd7, 0x7e, 0xb3, 0xcd, 
	0x33, 0xfc, 0xfc, 0xe3, 0x8f, 0x3f, 0xfe, 0xfc, 
	0xa7, 0x1f, 0x6f, 0xf8, 0xf1, 0xf6, 0xdf, 0x6f, 
	0xdf, 0xfd, 0xf7, 0xdf, 0xff, 0xe4, 0xe6, 0xcf, 
	0x77, 0xff, 0xfc, 0xf3, 0xcf, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x9f, 0xff, 
	0x7e, 0xfb, 0xef, 0xf3, 0xef, 0xfe, 0xff, 0xff, 
	0xff, 0x73, 0xb9, 0xbf, 0xff, 0xfe, 0xfd, 0xf7, 
	0xef, 0xff, 0xbf, 0xff, 0xff, 0xff, 0xff, 0xbf, 
	0xff, 0xfd, 0xdf, 0xff, 0xfe, 0xfb, 0xef, 0xff, 
	0xff, 0xfe, 0xff, 0xff, 0xff, 0xef, 0xae, 0xfe, 
	0x7b, 0xbf, 0xf7, 0xfc, 0xf3, 0x4f, 0x7f, 0xfd, 
	0xd9, 0xd7, 0x9f, 0x7f, 0xfd, 0xe1, 0x87, 0x17, 
	0x7e, 0x78, 0xfd, 0xf5, 0xd7, 0x5f, 0xfb, 0xff, 
	0xff, 0xbf, 0x8e, 0xf7, 0xbd, 0xde, 0xfc, 0xfe, 
	0xfb, 0x27, 0xf6, 0xfb, 0xef, 0xbf, 0xff, 0xf6, 
	0x7b, 0xcd, 0x1f, 0x7f, 0x7c, 0xf3, 0xe4, 0x77, 
	0x5f, 0x65, 0xd5, 0xff, 0xff, 0x7f, 0xaf, 0xfe, 
	0xb7, 0xef, 0xbb, 0xff, 0xbf, 0xf7, 0xd6, 0xff, 
	0xbf, 0xff, 0xfd, 0xfd, 0xdf, 0xdf, 0x5e, 0x79, 
	0xe9, 0xa5, 0xd7, 0x5b, 0xff, 0xfd, 0xf7, 0xff, 
	0xff, 0xff, 0x3f, 0xf9, 0xff, 0xfb, 0xec, 0xf3, 
	0xcf, 0x3f, 0xda, 0xfc, 0xf3, 0xcf, 0x37, 0xff, 
	0x7c, 0xd3, 0x1a, 0x6d, 0xb6, 0xd1, 0x66, 0x1f, 
	0x7f, 0xf4, 0xf1, 0xe7, 0x9f, 0x7f, 0xfe, 0xe9, 
	0xff, 0xfb, 0xbf, 0xff, 0xfc, 0xe7, 0x9f, 0x7f, 
	0xff, 0xfd, 0xff, 0xdf, 0xff, 0x3f, 0xf9, 0xe7, 
	0x9f, 0x7f, 0xfc, 0xfd, 0xb5, 0xf7, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0x97, 0xff, 0xbf, 0xbe, 0xff, 
	0xff, 0xff, 0xbf, 0xff, 0xfd, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xf5, 0xdf, 0xff, 0x7f, 0xff, 0xfd, 
	0xf7, 0xdf, 0xff, 0xef, 0xff, 0xff, 0xff, 0xcf, 
	0xfe, 0x9f, 0xff, 0xfb, 0xcf, 0xbf, 0xff, 0xff, 
	0xef, 0xdf, 0xff, 0xfe, 0xfd, 0xff, 0x3f, 0xff, 
	0xfc, 0xf1, 0xdf, 0xdf, 0x7f, 0xff, 0xff, 0xf7, 
	0xff, 0xff, 0xff, 0x7f, 0xfb, 0xff, 0xfb, 0xfb, 
	0x7f, 0xff, 0xfd, 0xdb, 0xff, 0xff, 0xfd, 0xff, 
	0xdf, 0x7f, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xdd, 0xf7, 0xff, 0xff, 0xff, 0xff, 
	0xe4, 0xff, 0xdb, 0xff, 0xbf, 0xbd, 0x7e, 0x93, 
	0xff, 0xfe, 0xff, 0xef, 0xff, 0xff, 0x77, 0xf1, 
	0xff, 0xff, 0xdf, 0x7f, 0xfb, 0xe9, 0xa7, 0xfa, 
	0xff, 0xff, 0xff, 0xff, 0x97, 0xfe, 0xef, 0xfa, 
	0xeb, 0xae, 0xb7, 0xee, 0xda, 0xeb, 0xaf, 0xbe, 
	0xfa, 0xda, 0x8b, 0x5d, 0x7d, 0xfd, 0xf5, 0xd7, 
	0x5f, 0x6f, 0xdd, 0xf5, 0xd7, 0x5f, 0x7f, 0xf5, 
	0xab, 0xfe, 0xbf, 0xff, 0xff, 0x4b, 0xfd, 0xfe, 
	0xec, 0xf7, 0x8f, 0xbc, 0xff, 0xed, 0xfb, 0x9f, 
	0xff, 0xfc, 0xf9, 0xfb, 0xe6, 0xff, 0xff, 0xf9, 
	0xe7, 0xfb, 0xff, 0xff, 0xff, 0xf0, 0x7f, 0xff, 
	0xfc, 0xef, 0xff, 0xdf, 0xf9, 0xfa, 0xef, 0xff, 
	0x7f, 0xfe, 0xff, 0x5d, 0x3f, 0xf3, 0xdd, 0xfb, 
	0xef, 0x9f, 0x6f, 0xfc, 0xf3, 0xff, 0xff, 0xff, 
	0x6f, 0xa0, 0xff, 0xff, 0xff, 0xff, 0xfe, 0xbf, 
	0xfd, 0xff, 0x7d, 0xff, 0xcd, 0xff, 0xdf, 0xff, 
	0xff, 0x5f, 0x5f, 0xff, 0xfa, 0xff, 0xff, 0xdf, 
	0x7f, 0xff, 0xff, 0xff, 0x7f, 0xa7, 0xff, 0xfd, 
	0xbf, 0xff, 0xfc, 0xeb, 0x8f, 0xd7, 0xff, 0xff, 
	0xff, 0xff, 0xf7, 0xef, 0xf4, 0xdf, 0x9b, 0x6f, 
	0xfc, 0xb5, 0xdf, 0x7a, 0x7f, 0xfc, 0xff, 0xff, 
	0xff, 0xdf, 0xfe, 0xef, 0xff, 0xf7, 0xff, 0x9f, 
	0xff, 0xff, 0xef, 0xff, 0xff, 0xfb, 0xbf, 0xff, 
	0xce, 0xff, 0xff, 0x7f, 0xfd, 0x37, 0xbf, 0xff, 
	0xfe, 0xfb, 0xfd, 0xff, 0xff, 0xef, 0xfb, 0xff, 
	0xff, 0xe7, 0x7d, 0xef, 0xef, 0xfe, 0x5f, 0xff, 
	0xef, 0xff, 0xdf, 0xbf, 0xfe, 0x5f, 0x79, 0xdf, 
	0xdf, 0xfe, 0xfe, 0xdd, 0xff, 0x7f, 0xff, 0xff, 
	0xff, 0xff, 0xeb, 0xdf, 0xdf, 0x6f, 0xff, 0xff, 
	0xff, 0xbf, 0xff, 0xff, 0xfd, 0xff, 0xff, 0xff, 
	0xbf, 0xe7, 0xf9, 0xff, 0xdf, 0xf7, 0xff, 0xf9, 
	0xff, 0xff, 0xef, 0xff, 0xff, 0xff, 0x93, 0xff, 
	0xff, 0xff, 0xfe, 0x6f, 0xff, 0xff, 0xff, 0xfb, 
	0xef, 0xff, 0xfd, 0xff, 0xf3, 0xff, 0x5f, 0xff, 
	0x77, 0xeb, 0x9f, 0xf3, 0xff, 0xbf, 0xbf, 0xff, 
	0xff, 0xff, 0xce, 0xf6, 0xff, 0xff, 0xdf, 0xff, 
	0xff, 0xbf, 0xef, 0xff, 0xf3, 0xcf, 0xdf, 0xff, 
	0xfc, 0xe6, 0xef, 0xbf, 0xff, 0x5e, 0xfb, 0xff, 
	0xef, 0xff, 0xf9, 0xff, 0xff, 0xff, 0xbf, 0xda, 
	0xff, 0xdf, 0xfd, 0xff, 0xfe, 0x7b, 0xfe, 0xfd, 
	0xcf, 0xf7, 0xfd, 0xff, 0xeb, 0x5b, 0xff, 0xbf, 
	0xff, 0xfe, 0xff, 0xef, 0xff, 0xff, 0xdf, 0xff, 
	0xff, 0xff, 0xff, 0xe1, 0x7f, 0xff, 0xbf, 0xfb, 
	0x6f, 0xff, 0xff, 0xbf, 0xdf, 0xff, 0x7d, 0xdf, 
	0xf1, 0xfb, 0xef, 0xd7, 0xff, 0x6f, 0x7f, 0xfd, 
	0xf7, 0xbb, 0xb9, 0xf7, 0xff, 0xff, 0xbf, 0x0f, 
	0xff, 0x7f, 0x7e, 0xfe, 0xff, 0xff, 0xbf, 0xff, 
	0xff, 0xff, 0xfd, 0xff, 0xff, 0xdf, 0x7f, 0xff, 
	0x7f, 0xfe, 0xec, 0xf7, 0xd7, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0x86, 0xfe, 0xff, 0xff, 0xfb, 
	0xff, 0xff, 0xff, 0xfe, 0xfe, 0xff, 0xff, 0xff, 
	0xfb, 0xff, 0x7f, 0xdf, 0xff, 0xfe, 0xd7, 0xdf, 
	0xdf, 0xff, 0xf7, 0xff, 0xff, 0xff, 0xff, 0xbf, 
	0xe2, 0x5f, 0xff, 0xfb, 0xe5, 0xf2, 0x5b, 0xef, 
	0xef, 0xf9, 0xe7, 0x9a, 0x7f, 0xbd, 0xf5, 0xf9, 
	0x97, 0x97, 0xf4, 0xf2, 0xcd, 0x2f, 0xdd, 0xff, 
	0xfe, 0x7f, 0xfd, 0x3f, 0x6e, 0x7f, 0xfe, 0xef, 
	0xb3, 0x5f, 0x2f, 0xdf, 0xef, 0xf3, 0xcf, 0x36, 
	0xff, 0xfc, 0xd7, 0x3f, 0x5f, 0x76, 0xff, 0xff, 
	0xf7, 0xfd, 0xff, 0xff, 0xe7, 0xff, 0xff, 0x7f, 
	0x87, 0xff, 0x35, 0xd7, 0xff, 0xe6, 0xdf, 0x7f, 
	0xd7, 0xfc, 0xff, 0xef, 0xff, 0xff, 0xfb, 0xdf, 
	0xdc, 0xd3, 0x4f, 0x9b, 0xbd, 0x9c, 0x5f, 0xfd, 
	0xff, 0xff, 0xd7, 0xff, 0x2f, 0xd6, 0x37, 0xcd, 
	0xff, 0xfb, 0x7f, 0xff, 0xfd, 0xee, 0xff, 0xef, 
	0xff, 0xff, 0xff, 0xff, 0xd9, 0x7f, 0xff, 0xe7, 
	0xf7, 0x5f, 0x66, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xf5, 0xbb, 0xbf, 0x5f, 0xcd, 0xb6, 0xfc, 0xf5, 
	0xd7, 0xef, 0xf6, 0xfd, 0x66, 0xdf, 0x7f, 0xff, 
	0x7b, 0xef, 0x36, 0xd3, 0x6b, 0xfb, 0xdd, 0xbc, 
	0xff, 0xfe, 0x7b, 0xcd, 0xee, 0x6d, 0xfe, 0x9f, 
	0x6f, 0xff, 0xf6, 0xff, 0xff, 0xef, 0xdf, 0xff, 
	0x6e, 0xfd, 0xff, 0xff, 0xff, 0xd7, 0x71, 0xdf, 
	0xb7, 0xf5, 0xf5, 0xfb, 0xea, 0xff, 0xff, 0xbf, 
	0x3a, 0x13, 0xfb, 0xbf, 0xf3, 0xff, 0xfb, 0x7f, 
	0xff, 0xbf, 0xb6, 0x7f, 0x6f, 0xfd, 0xd6, 0xff, 
	0xff, 0xd7, 0xd7, 0x7d, 0xfd, 0xd5, 0xda, 0x57, 
	0xff, 0xb7, 0xfe, 0xd6, 0xff, 0x16, 0xee, 0xbf, 
	0x5b, 0xf7, 0x6f, 0x7d, 0xd4, 0x51, 0x5f, 0x3d, 
	0xe9, 0xc4, 0xc7, 0x4f, 0x7d, 0x5e, 0x29, 0xaf, 
	0xf4, 0xd7, 0x4a, 0x7f, 0xef, 0xbf, 0x7a, 0xf2, 
	0xff, 0xbb, 0x7b, 0xff, 0xff, 0xdf, 0xfe, 0xed, 
	0x67, 0x9f, 0x7d, 0x7f, 0xef, 0xfd, 0xf7, 0xdf, 
	0xef, 0xfe, 0xdb, 0xef, 0xff, 0xfe, 0xfe, 0xcf, 
	0x7f, 0xbb, 0xff, 0xde, 0xff, 0xf7, 0xec, 0xfd, 
	0x9f, 0xff, 0xd7, 0xff, 0x77, 0xde, 0xf9, 0xfe, 
	0xf9, 0xe7, 0x9f, 0x7f, 0xbe, 0xf3, 0xcf, 0x38, 
	0xff, 0xed, 0xf3, 0xdf, 0xfa, 0xfb, 0xfc, 0xf3, 
	0xff, 0xdf, 0x03, 0xfb, 0xb7, 0xde, 0xec, 0xfb, 
	0xcf, 0x35, 0xff, 0xfe, 0xf3, 0x4f, 0x3f, 0xfe, 
	0xfc, 0xe7, 0x9f, 0x6f, 0xbc, 0x7d, 0xc6, 0xdb, 
	0x67, 0xf7, 0xf1, 0xe7, 0xdf, 0xff, 0x06, 0xe6, 
	0xcf, 0x3f, 0xff, 0x7f, 0xf7, 0xff, 0x7f, 0xdf, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x9f, 
	0x7f, 0xbe, 0xf9, 0xfd, 0xb3, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0x33, 0xb9, 0xbf, 0x7f, 0xff, 0xff, 
	0xfb, 0xff, 0xbf, 0xff, 0xff, 0xfe, 0xfb, 0xff, 
	0xbf, 0xff, 0xfd, 0xd7, 0x7f, 0xff, 0xf7, 0xff, 
	0xbf, 0xff, 0xff, 0xff, 0xfd, 0xff, 0xef, 0xa7, 
	0xfe, 0xbd, 0x3f, 0xf7, 0xf7, 0xeb, 0xaf, 0xff, 
	0xf5, 0xf1, 0xc7, 0x1f, 0x7f, 0xf4, 0xe3, 0x85, 
	0x13, 0x5e, 0xbf, 0xe1, 0xf7, 0xd7, 0xfe, 0xf8, 
	0xe3, 0xff, 0xbf, 0xbe, 0xf7, 0x7f, 0xbe, 0xff, 
	0xef, 0xbf, 0xff, 0xb6, 0xea, 0x6f, 0xbf, 0xfd, 
	0xfe, 0xfb, 0xdf, 0x17, 0x7f, 0x6c, 0xd5, 0xc7, 
	0xdf, 0x5d, 0xfd, 0xf7, 0xff, 0x7f, 0x7f, 0x9f, 
	0xfe, 0xd7, 0xe2, 0xfb, 0xff, 0xff, 0xfa, 0xeb, 
	0xff, 0x5f, 0xfd, 0xfc, 0xef, 0xef, 0xbf, 0x7e, 
	0x79, 0xed, 0xf7, 0x5e, 0x7e, 0xff, 0xfd, 0xef, 
	0xbf, 0xfd, 0xf7, 0x7f, 0xfb, 0xff, 0xdf, 0xfc, 
	0xf3, 0xcf, 0x3f, 0xff, 0xe8, 0xf3, 0xcd, 0x37, 
	0xf7, 0xfc, 0xf3, 0x9b, 0x6d, 0xa6, 0xd1, 0x67, 
	0x9b, 0x7d, 0xfc, 0xf9, 0xe7, 0x9f, 0x7f, 0xfe, 
	0xed, 0xff, 0xff, 0xff, 0xbf, 0xfe, 0xff, 0xff, 
	0xbf, 0xff, 0xff, 0xff, 0xdf, 0x7f, 0xff, 0xfb, 
	0xff, 0x1f, 0xff, 0xff, 0xfd, 0xf5, 0xff, 0xff, 
	0xfe, 0xff, 0xf5, 0xff, 0x8b, 0xff, 0xff, 0xbf, 
	0xff, 0xea, 0xbf, 0xff, 0xfe, 0xfa, 0xdf, 0xff, 
	0xff, 0xff, 0xbd, 0xfe, 0xff, 0xff, 0xff, 0xff, 
	0xfd, 0xd7, 0xff, 0xff, 0xff, 0xff, 0xd7, 0xff, 
	0x0f, 0xfe, 0xdf, 0xfb, 0xff, 0xeb, 0xff, 0xff, 
	0xff, 0xfb, 0xff, 0xff, 0xfe, 0xf9, 0xef, 0x3f, 
	0x7f, 0xfd, 0xf5, 0xf7, 0x5f, 0x7f, 0xff, 0xfd, 
	0xaf, 0xff, 0x5f, 0xff, 0x3f, 0xfb, 0xff, 0xff, 
	0xff, 0xaf, 0xff, 0xfd, 0xf7, 0xef, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xfe, 0xff, 0xfd, 0xff, 0x57, 
	0xdf, 0x7f, 0xff, 0xbd, 0xff, 0xff, 0x7f, 0xed, 
	0xff, 0xef, 0xff, 0xff, 0xef, 0xbf, 0xfd, 0xff, 
	0xbf, 0xef, 0xfe, 0xff, 0xef, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0x9f, 0xdb, 0x7a, 0xfa, 0xdd, 0xb7, 
	0xff, 0xff, 0xff, 0xed, 0xff, 0xb3, 0xfe, 0xff, 
	0xfa, 0x8b, 0x2f, 0xbf, 0xfa, 0xba, 0xe3, 0xaf, 
	0xbf, 0xfe, 0xfa, 0xab, 0x5f, 0x7f, 0xf5, 0x75, 
	0x57, 0x5b, 0x77, 0xdc, 0xd5, 0x97, 0x5f, 0x7d, 
	0xfd, 0x0b, 0xfe, 0xff, 0xff, 0xdd, 0xb3, 0xb7, 
	0xff, 0xef, 0xeb, 0xff, 0xfe, 0xdf, 0x68, 0x5f, 
	0xff, 0x6f, 0xfc, 0x71, 0xfd, 0x9f, 0xfb, 0x7f, 
	0xdf, 0xe7, 0x7f, 0x9e, 0xfd, 0x3f, 0xf3, 0x7f, 
	0xff, 0xfd, 0xff, 0xff, 0xfe, 0xbf, 0xfd, 0xff, 
	0xff, 0xff, 0xfe, 0xff, 0xff, 0x67, 0xf5, 0xfc, 
	0xf8, 0xcf, 0xbb, 0xff, 0xff, 0xdf, 0xfd, 0xbf, 
	0xbf, 0x7f, 0xa3, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xf7, 0xff, 0xdd, 0xff, 0xff, 0xff, 0xf7, 0xff, 
	0xbf, 0xff, 0xbf, 0xdf, 0xff, 0xff, 0xf9, 0xbf, 
	0xff, 0xdf, 0xff, 0xff, 0xff, 0xff, 0x9b, 0xff, 
	0xad, 0xbf, 0xff, 0xfb, 0xeb, 0xef, 0xbf, 0xfe, 
	0xff, 0xfe, 0xff, 0xf7, 0xff, 0xbf, 0xd7, 0x7f, 
	0x6f, 0xf9, 0x77, 0x57, 0x1b, 0xff, 0xbf, 0xfe, 
	0xdf, 0xff, 0x8f, 0xfe, 0x6f, 0xfe, 0xfb, 0xff, 
	0xbf, 0xff, 0xfd, 0xfb, 0xff, 0xff, 0xff, 0xfb, 
	0xff, 0x7e, 0x7f, 0xff, 0xff, 0x73, 0xef, 0x7f, 
	0xff, 0xfe, 0x7f, 0xfd, 0xdf, 0xff, 0x2f, 0xf9, 
	0xff, 0xff, 0xdf, 0xaf, 0xbb, 0xf6, 0xfb, 0xeb, 
	0x9f, 0xff, 0xff, 0xff, 0xff, 0xff, 0x77, 0xdf, 
	0x75, 0xdf, 0xff, 0x7f, 0xfd, 0xf7, 0xcd, 0xff, 
	0xff, 0xfc, 0xff, 0xef, 0xdf, 0x7b, 0xff, 0xff, 
	0xff, 0xff, 0xf7, 0xff, 0xbf, 0xff, 0xff, 0xdf, 
	0xff, 0xff, 0xcf, 0x9f, 0xff, 0xdf, 0xff, 0xff, 
	0xb1, 0xc7, 0x5e, 0xff, 0xfb, 0xf5, 0x7f, 0xbe, 
	0xff, 0xef, 0xdf, 0xff, 0x77, 0x9f, 0xff, 0xfb, 
	0xf9, 0xfe, 0xff, 0xfb, 0xff, 0xf7, 0xfe, 0xfb, 
	0xfd, 0x7e, 0xfb, 0xdf, 0xff, 0xff, 0xff, 0xaf, 
	0xdf, 0xff, 0xff, 0xef, 0xfe, 0xff, 0xeb, 0xff, 
	0xff, 0xff, 0x7f, 0xff, 0xfd, 0xe3, 0xff, 0xff, 
	0xff, 0xfc, 0xef, 0xdf, 0xde, 0x7f, 0x7f, 0xfb, 
	0xf7, 0xff, 0xff, 0x7f, 0xc7, 0xff, 0xff, 0x7f, 
	0xf8, 0xff, 0xef, 0xbf, 0xff, 0xff, 0xbf, 0xfd, 
	0xf7, 0x7f, 0xff, 0xbf, 0xff, 0xdf, 0xbf, 0x7f, 
	0xbf, 0xff, 0xfc, 0xfb, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xfe, 0xff, 0xff, 0xee, 0xff, 0xcb, 0xff, 
	0xbf, 0x7c, 0x77, 0xce, 0x77, 0xff, 0xff, 0xff, 
	0xdf, 0xd7, 0xf9, 0xdb, 0xaf, 0xbf, 0x5d, 0xff, 
	0xeb, 0xf7, 0xd3, 0x7f, 0xff, 0xff, 0xe7, 0xff, 
	0x2b, 0xff, 0xbf, 0xff, 0xff, 0xfb, 0xff, 0x9f, 
	0xff, 0xfd, 0xff, 0xff, 0xfd, 0xff, 0xdd, 0xbf, 
	0x7f, 0xf7, 0xff, 0xfd, 0xff, 0xc2, 0x5f, 0xfe, 
	0xf8, 0xbf, 0xda, 0xfa, 0x87, 0xfe, 0xff, 0xfe, 
	0xff, 0xef, 0xff, 0x7e, 0xdb, 0xff, 0xff, 0xff, 
	0xfe, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xf7, 
	0xff, 0xff, 0xfd, 0xff, 0xf7, 0xff, 0x9f, 0xfd, 
	0x7f, 0xf1, 0xff, 0xdb, 0xff, 0xf9, 0xb3, 0x57, 
	0xcb, 0xbf, 0xff, 0xd6, 0xfb, 0x1f, 0x3f, 0xf9, 
	0xcf, 0x97, 0xff, 0xfc, 0xeb, 0xfd, 0xf3, 0xfd, 
	0xff, 0xff, 0xfe, 0xff, 0xbf, 0xed, 0xff, 0xfb, 
	0xff, 0xff, 0x5f, 0x7e, 0xf9, 0xff, 0xfe, 0x9b, 
	0x6f, 0xfd, 0x7c, 0xe7, 0xff, 0xff, 0xff, 0xd3, 
	0xaf, 0xf7, 0xfd, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0x7f, 0x8b, 0xff, 0xaf, 0xff, 0xff, 0xf3, 0xdb, 
	0xbf, 0xa6, 0xce, 0xde, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0x54, 0xd3, 0x4b, 0xfb, 0xff, 0x7a, 0xff, 
	0x2f, 0xff, 0xff, 0xd4, 0xf9, 0x5f, 0xde, 0xff, 
	0xff, 0x7f, 0xb6, 0x6f, 0xb3, 0x9f, 0x7a, 0xee, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0x53, 0x7f, 0xff, 
	0xe5, 0xff, 0xf3, 0xef, 0x9f, 0xff, 0xff, 0x59, 
	0xcf, 0xb5, 0xb9, 0x5f, 0x9f, 0xfd, 0xbf, 0xd5, 
	0xf6, 0x9b, 0x69, 0xbf, 0xfd, 0xf7, 0xdf, 0x7f, 
	0xff, 0xfb, 0xef, 0xb5, 0xf7, 0xff, 0x7b, 0x7b, 
	0xe4, 0xbf, 0xff, 0xff, 0xaf, 0xee, 0x66, 0xbe, 
	0x7d, 0xff, 0xff, 0x54, 0xdb, 0x7b, 0xaf, 0xbd, 
	0x6b, 0xff, 0xff, 0xff, 0xff, 0xff, 0xdf, 0x57, 
	0x5f, 0xff, 0xff, 0x65, 0xb1, 0xed, 0xbe, 0xff, 
	0x3d, 0x3b, 0xaf, 0xf9, 0xff, 0xff, 0xff, 0xff, 
	0xef, 0xa5, 0xff, 0xff, 0x7a, 0xbd, 0xfd, 0xd6, 
	0xfb, 0xdf, 0xd6, 0xdf, 0x6d, 0xff, 0xdf, 0xff, 
	0xff, 0x7f, 0x7d, 0xff, 0xd2, 0xff, 0x66, 0xe2, 
	0xff, 0xff, 0xff, 0xef, 0xbf, 0xd5, 0xf7, 0xdf, 
	0x2b, 0xfd, 0xe1, 0xc7, 0xdf, 0xff, 0x50, 0x79, 
	0x0f, 0xf5, 0xff, 0xea, 0xfe, 0xfb, 0xed, 0xff, 
	0x79, 0xff, 0x3b, 0x72, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xf7, 0xdf, 0x7f, 0x7f, 0xcb, 0xfd, 0xf7, 
	0xdf, 0xef, 0xfe, 0xfb, 0x6f, 0xff, 0xfc, 0xfe, 
	0xdb, 0x6f, 0xbb, 0xed, 0xe7, 0xff, 0x77, 0xe9, 
	0x5d, 0xff, 0xfd, 0xff, 0xf9, 0xff, 0xff, 0xff, 
	0xff, 0x99, 0xe7, 0x9f, 0x7f, 0xbe, 0xf3, 0xcf, 
	0x3f, 0xe3, 0xad, 0xf3, 0xff, 0xfb, 0xff, 0xcf, 
	0x3f, 0xff, 0xde, 0x13, 0xfb, 0xb4, 0xc7, 0xfe, 
	0xeb, 0xef, 0xb5, 0xff, 0x7e, 0xf3, 0x8f, 0x3f, 
	0xfe, 0xf8, 0xc7, 0x9f, 0x7f, 0xbc, 0x7d, 0xc7, 
	0x9b, 0x6f, 0xbe, 0xf9, 0xf6, 0xdf, 0xf7, 0x86, 
	0xee, 0xff, 0x7f, 0xff, 0xfd, 0xff, 0xdd, 0x77, 
	0xdf, 0xfc, 0xff, 0xff, 0x7f, 0xff, 0xff, 0xff, 
	0xff, 0x7f, 0xbe, 0xff, 0xf7, 0xb7, 0x7f, 0x7e, 
	0xfb, 0xff, 0xbf, 0x73, 0xb8, 0xff, 0xbf, 0xff, 
	0xfb, 0xff, 0xbf, 0xff, 0xfe, 0xfe, 0xfe, 0xff, 
	0xbf, 0xff, 0xff, 0xff, 0xf7, 0x7f, 0xff, 0xff, 
	0xff, 0xbf, 0xff, 0xfe, 0xfd, 0xff, 0x7f, 0xef, 
	0xe9, 0xfe, 0xfb, 0xef, 0xff, 0xdb, 0x5e, 0x7f, 
	0xed, 0xbd, 0xf0, 0xc7, 0x1f, 0x35, 0xfd, 0xe3, 
	0xaf, 0x9f, 0x5e, 0xef, 0xe3, 0xe7, 0x9f, 0xcf, 
	0x7e, 0xfd, 0xff, 0xbf, 0x97, 0xf7, 0xff, 0xfe, 
	0xfc, 0x63, 0xaf, 0xad, 0xf6, 0x7a, 0xe2, 0xbf, 
	0xff, 0xbe, 0xd9, 0xdf, 0x7f, 0xdf, 0x7c, 0xdd, 
	0xdf, 0xf7, 0xdf, 0xff, 0xfd, 0xe5, 0x56, 0x7f, 
	0x3f, 0xfe, 0xbf, 0xf7, 0xfe, 0xff, 0xee, 0xf7, 
	0xff, 0xfe, 0x2f, 0xfc, 0xfa, 0xff, 0x9b, 0xbf, 
	0xfe, 0x7b, 0xed, 0x77, 0xff, 0x7e, 0xbf, 0xfd, 
	0xff, 0xdf, 0x7f, 0xf7, 0x3f, 0xf9, 0xff, 0xde, 
	0xec, 0x33, 0xc7, 0x3e, 0xfa, 0xec, 0xe3, 0xcd, 
	0x3f, 0xf7, 0x78, 0xf3, 0x9b, 0x6f, 0xb4, 0xd1, 
	0xe7, 0x9a, 0x7f, 0xfe, 0xd9, 0xc7, 0x0f, 0x7f, 
	0xfe, 0xe7, 0xff, 0xff, 0xff, 0xff, 0x5f, 0xfe, 
	0xef, 0xef, 0xbf, 0xfd, 0xff, 0xdf, 0xbf, 0xff, 
	0xff, 0xef, 0xdf, 0xdf, 0xff, 0xf3, 0xff, 0xff, 
	0xff, 0x7f, 0xb9, 0xfe, 0xff, 0x9f, 0xff, 0xef, 
	0xff, 0xff, 0xdb, 0xaf, 0xbf, 0xfe, 0xfa, 0xf3, 
	0xff, 0xff, 0xbf, 0xff, 0xfe, 0xbf, 0xdf, 0x7f, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xe5, 0xd7, 
	0xff, 0xbf, 0xfe, 0xdf, 0xfb, 0xff, 0xff, 0xef, 
	0xff, 0xfe, 0xfe, 0xbf, 0xff, 0xfe, 0xf9, 0xeb, 
	0x7f, 0xff, 0xfd, 0xf7, 0xff, 0x7f, 0x7f, 0xff, 
	0xfd, 0xff, 0xf7, 0x7f, 0xff, 0x7f, 0xfb, 0xff, 
	0xff, 0xff, 0xbf, 0xf7, 0xfe, 0xfb, 0xef, 0xff, 
	0xff, 0xff, 0xff, 0xfb, 0xff, 0xff, 0xff, 0xff, 
	0xf7, 0xff, 0xff, 0xff, 0xff, 0xdf, 0x7f, 0x6e, 
	0xef, 0xff, 0xe0, 0xff, 0xd3, 0xfb, 0x3f, 0xf5, 
	0xf6, 0xd3, 0xff, 0xff, 0xff, 0xef, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xf7, 0xdf, 0xfe, 0xff, 0xe9, 
	0xbf, 0xff, 0xfc, 0xff, 0xef, 0xff, 0xbb, 0x7f, 
	0xef, 0xfa, 0x63, 0xad, 0xba, 0xea, 0xfa, 0xeb, 
	0xaf, 0xbd, 0xfa, 0xda, 0xab, 0x5e, 0x7f, 0xd5, 
	0xf5, 0xd7, 0x5f, 0x73, 0xf5, 0xf5, 0x97, 0x1f, 
	0x7f, 0xfd, 0x6f, 0xfe, 0xff, 0x7f, 0xff, 0x8f, 
	0xef, 0xff, 0xef, 0xfb, 0xcf, 0x7f, 0xff, 0xef, 
	0xbf, 0xff, 0x7f, 0xf7, 0xf1, 0x79, 0x9e, 0xff, 
	0x3f, 0xdf, 0xfc, 0xff, 0xdb, 0xff, 0x77, 0xd0, 
	0xbf, 0xbf, 0xff, 0xeb, 0xee, 0xff, 0x7f, 0xfe, 
	0xff, 0x6f, 0x7f, 0xff, 0xff, 0xff, 0xbf, 0xff, 
	0xd5, 0xf9, 0xf7, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xef, 0xff, 0x7e, 0xa0, 0xff, 0xff, 0xff, 0x7f, 
	0xff, 0xff, 0xff, 0xde, 0xbf, 0xfe, 0xff, 0x7f, 
	0xff, 0xf7, 0xbf, 0xbb, 0xbf, 0xff, 0x7f, 0xed, 
	0xbf, 0xff, 0xff, 0xbf, 0xff, 0xff, 0xff, 0x12, 
	0xff, 0xcd, 0xbf, 0xfd, 0xf7, 0xff, 0x85, 0xfa, 
	0xff, 0xfb, 0xff, 0xf5, 0xff, 0xff, 0xff, 0xff, 
	0x7f, 0xee, 0xfd, 0xff, 0xe6, 0x9b, 0x6f, 0xbd, 
	0xfd, 0xfd, 0xfb, 0x77, 0xfe, 0xff, 0xff, 0xeb, 
	0xff, 0xf7, 0x7f, 0xb7, 0xbf, 0xfb, 0xff, 0xff, 
	0xff, 0xff, 0xfb, 0xf3, 0xd5, 0xdf, 0xfb, 0xff, 
	0xff, 0xfb, 0xff, 0xed, 0xdd, 0xfd, 0xf7, 0xbf, 
	0xfa, 0xbf, 0xf9, 0xef, 0xfd, 0xf5, 0xfe, 0xfb, 
	0xff, 0xff, 0xff, 0xee, 0xff, 0xff, 0xef, 0xb7, 
	0xff, 0xfd, 0xbf, 0xff, 0xff, 0xdd, 0xff, 0x4f, 
	0x7f, 0xef, 0xff, 0xff, 0x27, 0xef, 0xfb, 0x4f, 
	0xff, 0xdf, 0xdc, 0xf3, 0xfd, 0xff, 0xff, 0xf0, 
	0xff, 0x7f, 0xbf, 0xbb, 0xff, 0xff, 0xef, 0xff, 
	0xff, 0xeb, 0xff, 0xdf, 0xfb, 0xef, 0xff, 0xff, 
	0x97, 0xfb, 0xeb, 0xbf, 0xff, 0xef, 0x3b, 0xed, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xfe, 
	0xff, 0xf8, 0xff, 0xf7, 0xff, 0xfb, 0xff, 0x7f, 
	0xbf, 0xef, 0x7f, 0xff, 0x1f, 0xf6, 0xef, 0xff, 
	0xca, 0xfb, 0xfd, 0x5f, 0xff, 0xfc, 0xf3, 0xcf, 
	0x3f, 0xff, 0xff, 0xff, 0xcf, 0x7f, 0xfc, 0x79, 
	0x85, 0xff, 0x6f, 0xfe, 0xe9, 0xff, 0xff, 0xff, 
	0xbf, 0xd8, 0xff, 0xff, 0xbf, 0xee, 0xef, 0xff, 
	0xfb, 0xd8, 0x6f, 0xff, 0xfd, 0xf6, 0xff, 0xff, 
	0x77, 0xff, 0xee, 0xd7, 0xfb, 0xfe, 0x3f, 0xed, 
	0xb7, 0xff, 0xff, 0xff, 0xff, 0xe5, 0xff, 0xef, 
	0x7f, 0xbe, 0xfe, 0x3a, 0xff, 0x7e, 0xdf, 0xfd, 
	0xfd, 0xdf, 0xf6, 0xf7, 0xff, 0xff, 0xbf, 0xfb, 
	0x3b, 0xbd, 0x77, 0xff, 0x79, 0xf6, 0x79, 0xf7, 
	0xff, 0x37, 0xff, 0x3e, 0xdb, 0xf7, 0xc3, 0x5f, 
	0xfb, 0xff, 0xbf, 0xff, 0xff, 0x77, 0xff, 0x7f, 
	0xbf, 0xbf, 0xfe, 0xd9, 0xbf, 0xf6, 0x4f, 0xff, 
	0xff, 0xff, 0xe7, 0xff, 0xff, 0x3f, 0xfa, 0x3f, 
	0xff, 0xff, 0xef, 0xff, 0xfe, 0xff, 0xff, 0xbf, 
	0xff, 0xff, 0xff, 0xaf, 0xfd, 0xfc, 0xff, 0xff, 
	0xdf, 0xdf, 0xbf, 0xff, 0xff, 0xfd, 0xe6, 0xdf, 
	0xfd, 0xbf, 0xe2, 0xff, 0x7e, 0xfd, 0xfc, 0xbe, 
	0xff, 0xcf, 0xff, 0xef, 0xff, 0x9b, 0xef, 0x7f, 
	0xff, 0xcb, 0x87, 0xff, 0x77, 0x7f, 0xf9, 0xad, 
	0xde, 0x7f, 0xff, 0xfb, 0xed, 0x3f, 0x62, 0xff, 
	0xff, 0xe9, 0xbf, 0xfd, 0xfe, 0xdf, 0xff, 0xbf, 
	0xff, 0x57, 0xfd, 0xff, 0xd7, 0x3f, 0xff, 0xff, 
	0xd9, 0xf9, 0xfb, 0xfd, 0x7f, 0xf3, 0xfe, 0xff, 
	0xfd, 0x7f, 0x9f, 0xff, 0xf5, 0xa7, 0xce, 0xfe, 
	0xeb, 0xfc, 0xff, 0xff, 0xde, 0xfb, 0xff, 0xfe, 
	0xff, 0xff, 0x7f, 0x53, 0x7f, 0xff, 0xf5, 0xda, 
	0xff, 0xff, 0xef, 0xff, 0xdf, 0xff, 0x7f, 0xd6, 
	0xf3, 0xcf, 0x79, 0xeb, 0xf9, 0xff, 0xff, 0xff, 
	0xfb, 0xef, 0xff, 0xff, 0xff, 0xff, 0xff, 0x67, 
	0xfd, 0xfd, 0xf6, 0xd3, 0xff, 0xbf, 0xff, 0xff, 
	0xd9, 0xff, 0x35, 0xba, 0xff, 0xb2, 0x7d, 0xfb, 
	0xfd, 0xf7, 0x9f, 0x7d, 0xef, 0xbd, 0xf6, 0xdf, 
	0x7f, 0xff, 0xfb, 0xef, 0xbf, 0xf7, 0x76, 0x37, 
	0xed, 0xbe, 0xfd, 0xec, 0xba, 0xdd, 0xee, 0x6d, 
	0xfe, 0xdb, 0xf6, 0x99, 0x7b, 0x9f, 0xbd, 0xf5, 
	0xbf, 0xfb, 0xea, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0x7f, 0xdf, 0xfd, 0xff, 0x55, 0xf9, 0xff, 0xad, 
	0x3f, 0xfc, 0x3b, 0x07, 0xfb, 0xff, 0xff, 0xff, 
	0xfe, 0x7e, 0xff, 0xff, 0x5f, 0x5e, 0xbb, 0xfd, 
	0xff, 0xdb, 0xdf, 0x7e, 0x5f, 0x7d, 0xff, 0xf5, 
	0x57, 0xff, 0xfd, 0xbd, 0xfe, 0xd7, 0xf7, 0x86, 
	0xee, 0xff, 0xfd, 0xf7, 0x2b, 0xff, 0xbf, 0xf3, 
	0x6f, 0xbb, 0xeb, 0xe5, 0xf3, 0x7f, 0xfc, 0xf2, 
	0x7b, 0x8d, 0x9d, 0xf7, 0x5e, 0xef, 0xef, 0x5d, 
	0x36, 0xff, 0xff, 0xbb, 0x79, 0xff, 0x37, 0xff, 
	0x76, 0xfb, 0x6f, 0xf7, 0xdf, 0x7b, 0xff, 0xbd, 
	0xf7, 0xff, 0xef, 0xde, 0xfb, 0xef, 0xbf, 0xff, 
	0xff, 0x1f, 0xef, 0xbf, 0xf9, 0xff, 0xff, 0xf7, 
	0x69, 0xfd, 0xff, 0x7d, 0xfe, 0x59, 0xff, 0x9f, 
	0x7d, 0xfe, 0xf9, 0x67, 0x9e, 0xff, 0xff, 0xf3, 
	0xcf, 0x3f, 0xff, 0xfc, 0xf7, 0xff, 0x3f, 0xff, 
	0xec, 0xb3, 0xdf, 0xdf, 0x02, 0xfb, 0x37, 0xef, 
	0xfa, 0xb3, 0xef, 0x37, 0xfb, 0x5c, 0x73, 0x8d, 
	0x3f, 0xff, 0xfe, 0xc7, 0x1f, 0x7f, 0xfc, 0xf9, 
	0xf6, 0xdf, 0x7f, 0xbc, 0x99, 0x66, 0xda, 0xff, 
	0xe6, 0xe6, 0xcf, 0x7f, 0xff, 0xfd, 0xff, 0xfb, 
	0xff, 0xff, 0xfd, 0xff, 0xff, 0xff, 0xff, 0xfb, 
	0xff, 0xff, 0xff, 0x7f, 0xff, 0xff, 0xff, 0xdf, 
	0xfe, 0xfb, 0xef, 0xff, 0xb3, 0xb8, 0xdf, 0xbf, 
	0xff, 0xfd, 0xfe, 0xff, 0xef, 0xbf, 0xfe, 0xfe, 
	0xff, 0xef, 0xff, 0xff, 0xff, 0xff, 0xff, 0x7f, 
	0xff, 0xff, 0xff, 0xff, 0x7e, 0xfd, 0xdf, 0xff, 
	0xef, 0xa2, 0x7e, 0xcf, 0xbd, 0xfe, 0xf4, 0xfe, 
	0x0f, 0x7d, 0xfe, 0xf8, 0x52, 0x5f, 0xff, 0xff, 
	0xeb, 0xaf, 0x3f, 0x4e, 0xe8, 0xad, 0xfd, 0x3f, 
	0x5e, 0xf8, 0xf5, 0xff, 0xbf, 0x86, 0x77, 0xe7, 
	0xbf, 0xf9, 0xe7, 0xbf, 0x67, 0xfe, 0xfb, 0xe6, 
	0x8b, 0xff, 0xfe, 0xfb, 0xdf, 0x7e, 0xfb, 0x7d, 
	0xf1, 0x57, 0x77, 0xff, 0x7f, 0xb1, 0x4f, 0x56, 
	0x7d, 0xaf, 0xfe, 0xff, 0xde, 0xbf, 0xff, 0xef, 
	0xff, 0xdf, 0x6f, 0xaf, 0xbf, 0xfd, 0xff, 0x7f, 
	0xff, 0xfe, 0xfb, 0xeb, 0xf5, 0xdf, 0x7b, 0xff, 
	0xdf, 0xb7, 0x7e, 0xff, 0xff, 0x3f, 0xfa, 0xff, 
	0xff, 0xdc, 0xf3, 0x8f, 0x3f, 0xf7, 0xfc, 0xe3, 
	0x8f, 0x27, 0xff, 0xfc, 0xf3, 0x9b, 0x6f, 0xbe, 
	0x91, 0x67, 0x9f, 0x7d, 0xee, 0xb1, 0xe7, 0x1b, 
	0x7d, 0xfe, 0xe7, 0xff, 0xdf, 0xff, 0xff, 0xff, 
	0xff, 0xdf, 0x7f, 0xff, 0xfc, 0xe6, 0xdf, 0x7f, 
	0xfe, 0xff, 0xef, 0xff, 0xff, 0xfe, 0xfd, 0xcd, 
	0xbf, 0x5f, 0xfc, 0xff, 0xf7, 0xff, 0xbb, 0xff, 
	0xff, 0xff, 0xf7, 0xdf, 0x7b, 0xff, 0xfd, 0xf9, 
	0xe6, 0xef, 0x7b, 0xff, 0xff, 0xfe, 0xbf, 0xff, 
	0xff, 0xfb, 0xfd, 0x77, 0xff, 0x7e, 0xfd, 0xff, 
	0xdf, 0xff, 0x7f, 0xfe, 0xff, 0xfc, 0xff, 0xff, 
	0xff, 0xff, 0xfd, 0xff, 0xef, 0xff, 0xfa, 0xfb, 
	0xff, 0x7f, 0xff, 0xfd, 0xf7, 0xdf, 0xdf, 0x7f, 
	0xff, 0xff, 0xf7, 0xff, 0x7f, 0xff, 0xbf, 0xfb, 
	0xff, 0xf7, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xef, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xdf, 0xfb, 0xfd, 0xff, 0xf7, 0xfe, 
	0xfb, 0xed, 0xff, 0xe9, 0xff, 0xca, 0x6f, 0xbf, 
	0xfd, 0xf4, 0xdb, 0xff, 0xbf, 0xbb, 0xf4, 0xff, 
	0xff, 0xbe, 0x7f, 0xdf, 0xfe, 0x5f, 0x7e, 0xf3, 
	0xff, 0xff, 0xff, 0xff, 0xfb, 0xe7, 0xff, 0x93, 
	0xfe, 0xff, 0xba, 0xab, 0xaf, 0xb7, 0xea, 0xfa, 
	0xeb, 0xae, 0xbd, 0xfe, 0xf2, 0xcb, 0x5f, 0x7d, 
	0xec, 0x55, 0xd7, 0x5d, 0x7f, 0xfd, 0xf5, 0x57, 
	0x5f, 0x7f, 0xf4, 0x8b, 0xf6, 0x7f, 0xfd, 0xbe, 
	0x11, 0xff, 0x9f, 0xef, 0xff, 0xff, 0x1f, 0xfe, 
	0x6c, 0xbf, 0x9f, 0x7f, 0xfb, 0xef, 0x76, 0xf7, 
	0xff, 0xbd, 0xd9, 0xff, 0xf7, 0xf7, 0xed, 0x3f, 
	0xf3, 0xff, 0xfe, 0xbe, 0xef, 0xf7, 0xe7, 0xff, 
	0xff, 0xfd, 0x7e, 0xee, 0xba, 0xff, 0xbe, 0xb7, 
	0xff, 0xff, 0xff, 0xfd, 0xf7, 0xff, 0x7f, 0xff, 
	0xed, 0xb7, 0xff, 0x7e, 0xec, 0xff, 0xff, 0xff, 
	0xfd, 0x7d, 0xcf, 0xbf, 0xff, 0xff, 0xf7, 0xdf, 
	0x3d, 0xff, 0xff, 0xbd, 0xfb, 0xff, 0xbf, 0xfb, 
	0xef, 0xef, 0xd7, 0xff, 0xff, 0xff, 0xff, 0x7f, 
	0x8b, 0xff, 0x0f, 0xff, 0xde, 0xfc, 0xf3, 0xcd, 
	0xfb, 0xff, 0xf7, 0xef, 0xfd, 0xff, 0xf7, 0xef, 
	0xff, 0xff, 0x7f, 0xbd, 0xf9, 0xff, 0xfb, 0xef, 
	0xbf, 0xb7, 0xd7, 0xff, 0x2f, 0xfe, 0xbf, 0xff, 
	0xbe, 0xfd, 0xff, 0xff, 0xbf, 0xef, 0xff, 0xfe, 
	0xef, 0xff, 0xdf, 0xfb, 0xf9, 0xff, 0xff, 0xf3, 
	0xff, 0xfd, 0xff, 0xff, 0xff, 0xf5, 0xd7, 0xf6, 
	0x3f, 0xfb, 0xff, 0xbf, 0xff, 0xff, 0x77, 0xae, 
	0xdb, 0xff, 0x3f, 0x3f, 0xfe, 0xfb, 0xdf, 0x3f, 
	0xff, 0xdf, 0x73, 0xdf, 0xfe, 0xff, 0xdf, 0xff, 
	0xdf, 0x3f, 0xff, 0xfc, 0xff, 0xeb, 0xff, 0xff, 
	0xaf, 0xb9, 0xf7, 0xea, 0xf3, 0x7d, 0xbf, 0xde, 
	0xfe, 0xdb, 0xff, 0x7f, 0xef, 0x5e, 0x97, 0xcf, 
	0xff, 0xff, 0xff, 0xff, 0x9f, 0x7f, 0xf9, 0xff, 
	0x7f, 0x8f, 0xfd, 0xff, 0xff, 0x7f, 0xfb, 0x3f, 
	0xbb, 0xff, 0x7f, 0xf7, 0xef, 0x37, 0xfe, 0xd7, 
	0xb7, 0xff, 0xff, 0xff, 0xa9, 0xfe, 0xfe, 0xff, 
	0xff, 0x7d, 0xff, 0xff, 0xfe, 0x6f, 0xfe, 0xff, 
	0xff, 0xff, 0xff, 0xdd, 0x7f, 0xff, 0xbf, 0xff, 
	0xdf, 0x3e, 0xfe, 0xff, 0xc7, 0x9f, 0x9f, 0xbf, 
	0xfe, 0xff, 0xf7, 0x7f, 0xfe, 0x9f, 0x7f, 0x9f, 
	0xff, 0xff, 0xf8, 0xff, 0xff, 0x3f, 0xff, 0xde, 
	0xfe, 0xfb, 0xff, 0xff, 0x7f, 0xff, 0xf7, 0xff, 
	0x7f, 0x7f, 0xfd, 0xff, 0xff, 0xff, 0xff, 0xbf, 
	0xf7, 0xff, 0xfd, 0x7f, 0xfb, 0xff, 0xa6, 0xfe, 
	0xfd, 0x3f, 0xff, 0xef, 0x3b, 0xed, 0xf3, 0xbf, 
	0x76, 0xfb, 0xef, 0xf7, 0xdf, 0xbb, 0xff, 0xff, 
	0x5f, 0xff, 0xdf, 0x2f, 0xef, 0xf9, 0xf7, 0xfd, 
	0xe7, 0xbf, 0x27, 0xff, 0xff, 0xff, 0x6e, 0xdb, 
	0x4f, 0xbf, 0xff, 0x7f, 0xe3, 0xb7, 0xfe, 0xff, 
	0xff, 0x5f, 0xfb, 0xfd, 0x7f, 0xfd, 0xfe, 0xff, 
	0xff, 0xff, 0xbd, 0xa7, 0xff, 0xff, 0xb2, 0xfe, 
	0xff, 0xff, 0xfa, 0xef, 0xcf, 0x3f, 0xfb, 0xff, 
	0xef, 0xff, 0xbb, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0x5f, 0xff, 0xfd, 0xff, 0xf7, 0xdb, 0xe7, 
	0xdf, 0xff, 0x7f, 0xf1, 0xff, 0xfa, 0xef, 0xef, 
	0xf3, 0xfb, 0xef, 0xff, 0xef, 0xbe, 0x5f, 0xef, 
	0xbf, 0xff, 0xad, 0xf7, 0xb5, 0xf7, 0xff, 0xaf, 
	0xf7, 0x9f, 0xff, 0xd3, 0xfd, 0xfd, 0xbf, 0xe4, 
	0xff, 0xfb, 0xef, 0xff, 0xff, 0xf7, 0xdf, 0xff, 
	0xbf, 0xfb, 0x7f, 0xf9, 0xfc, 0xe7, 0x3b, 0xef, 
	0xf7, 0xdf, 0xff, 0xff, 0xdf, 0x7f, 0xff, 0xe7, 
	0xfd, 0xb5, 0x7f, 0xaf, 0xff, 0xff, 0xe7, 0x9f, 
	0xfe, 0xff, 0xff, 0xff, 0x9d, 0xff, 0xeb, 0xff, 
	0xfe, 0xfb, 0xfd, 0xff, 0xff, 0x7e, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0x9f, 0x97, 0xf4, 0xf9, 0x6f, 
	0xde, 0xff, 0xff, 0x3f, 0xff, 0xf9, 0xff, 0xff, 
	0x3d, 0xff, 0xad, 0xbf, 0xff, 0xff, 0xdf, 0xff, 
	0xcf, 0x97, 0xfd, 0xff, 0xff, 0xff, 0xbf, 0xfd, 
	0xf4, 0xdb, 0xff, 0xf5, 0xb8, 0xff, 0x9a, 0x69, 
	0xe6, 0xd9, 0xf7, 0xda, 0x9f, 0xef, 0xad, 0x5e, 
	0x9e, 0x7d, 0x7f, 0xab, 0xef, 0xf5, 0xfb, 0xfe, 
	0xff, 0xef, 0xbf, 0xb5, 0x5f, 0xdf, 0xed, 0xef, 
	0x6a, 0xfe, 0x7b, 0xf5, 0xd5, 0x67, 0xef, 0xeb, 
	0xff, 0xbf, 0xfb, 0xae, 0xa9, 0xff, 0x7f, 0xf7, 
	0xfe, 0xf7, 0xe7, 0xfb, 0xff, 0xff, 0xff, 0xff, 
	0x3e, 0xd5, 0xd4, 0x3f, 0xbf, 0xf9, 0xff, 0xff, 
	0xf7, 0xdf, 0xfb, 0xaf, 0xfc, 0xdf, 0x2a, 0xeb, 
	0xef, 0xd6, 0xdb, 0xf5, 0x78, 0x5f, 0x5f, 0xad, 
	0xbf, 0x7f, 0xfb, 0x65, 0x7f, 0x75, 0xd6, 0xfb, 
	0x36, 0xee, 0xff, 0xff, 0xff, 0xef, 0xef, 0xbc, 
	0xf6, 0x77, 0x6a, 0xad, 0xbf, 0xd6, 0x1f, 0x56, 
	0xea, 0x7b, 0xfd, 0xb5, 0xfc, 0xff, 0xc3, 0xaf, 
	0xec, 0xd7, 0x5b, 0xdd, 0xfb, 0x72, 0xef, 0xff, 
	0xff, 0xf6, 0xf3, 0xfd, 0xf7, 0xff, 0x7b, 0xef, 
	0xff, 0x37, 0xde, 0xef, 0xfe, 0xfb, 0x3f, 0xff, 
	0xf7, 0xb7, 0x9b, 0xef, 0xbf, 0xed, 0xdf, 0xdb, 
	0x76, 0xe6, 0xfd, 0xfd, 0xfd, 0xff, 0xdf, 0xe7, 
	0xbf, 0xfd, 0xf7, 0xf9, 0x7f, 0x9f, 0x7f, 0xfe, 
	0x33, 0xcf, 0x7f, 0xeb, 0xff, 0xbf, 0xcf, 0x3c, 
	0xff, 0xcf, 0xf7, 0xcf, 0xde, 0x2f, 0xf9, 0xb5, 
	0xdb, 0xfe, 0xfb, 0xce, 0xb5, 0xdb, 0xbc, 0xe3, 
	0xed, 0x3b, 0xfe, 0x78, 0xe6, 0x1f, 0x6f, 0x9f, 
	0xfd, 0xf6, 0x1d, 0x7f, 0xac, 0xf9, 0xf7, 0x9a, 
	0xf7, 0x44, 0xee, 0xff, 0x3f, 0xff, 0xff, 0xff, 
	0xff, 0x3f, 0xff, 0xfd, 0xf6, 0xff, 0xff, 0xff, 
	0xfc, 0xff, 0x9f, 0x6f, 0xff, 0xfd, 0xef, 0xff, 
	0xff, 0xfe, 0xff, 0xef, 0xff, 0x73, 0xbb, 0xff, 
	0xbf, 0xff, 0xff, 0xff, 0xfb, 0xff, 0xfe, 0xfd, 
	0xf7, 0xff, 0xff, 0xff, 0xfb, 0xfd, 0xdf, 0xff, 
	0xff, 0xff, 0xdf, 0xff, 0xff, 0xfd, 0xff, 0xed, 
	0xf7, 0xef, 0xe7, 0x7e, 0xfb, 0xed, 0xff, 0xff, 
	0xc3, 0x0f, 0x3f, 0x9f, 0xf0, 0xfb, 0x0f, 0x7f, 
	0xfc, 0xa1, 0xaf, 0xd7, 0x7f, 0xfb, 0xfd, 0x85, 
	0x3f, 0xfe, 0x7e, 0xef, 0xc5, 0xbf, 0xbf, 0xf7, 
	0xbf, 0xfe, 0xfb, 0xef, 0x8d, 0x7f, 0xba, 0xff, 
	0xe3, 0xcb, 0x3f, 0xfe, 0xfb, 0xcd, 0x7f, 0x5f, 
	0x75, 0xb5, 0xdf, 0xbb, 0xff, 0xff, 0xff, 0xec, 
	0x17, 0x6f, 0xff, 0xfe, 0xff, 0xef, 0xbe, 0xfb, 
	0xbe, 0xff, 0xff, 0xfb, 0x2b, 0xaf, 0xbf, 0xd2, 
	0xef, 0x5f, 0xfe, 0xfb, 0xfd, 0xf7, 0xff, 0xfe, 
	0xef, 0xf7, 0xfe, 0xf6, 0x5f, 0xfd, 0x7f, 0xfa, 
	0xff, 0xff, 0xec, 0xa3, 0xcf, 0x3c, 0xf7, 0xec, 
	0xb2, 0x8d, 0x3e, 0xda, 0xfc, 0x73, 0x9b, 0x6f, 
	0xf4, 0xd1, 0xe7, 0x9b, 0x7b, 0xfe, 0xf9, 0xc7, 
	0x1f, 0x7f, 0xfe, 0xeb, 0xff, 0xfb, 0xff, 0xff, 
	0xff, 0xf7, 0xdf, 0xbf, 0xff, 0xfc, 0xf3, 0xcf, 
	0x7f, 0xbf, 0xfd, 0xff, 0xf7, 0xff, 0xff, 0xff, 
	0xcf, 0xbf, 0x7f, 0x7f, 0xf3, 0xe7, 0xff, 0x87, 
	0xff, 0xbf, 0xfd, 0xf6, 0xdf, 0x7f, 0xff, 0xbd, 
	0xbf, 0xf7, 0xcb, 0x3f, 0xff, 0xff, 0xfd, 0xff, 
	0xdf, 0x7f, 0xf7, 0xfd, 0xff, 0xff, 0x7e, 0xff, 
	0xfd, 0x97, 0xff, 0x3f, 0xfe, 0xbf, 0xff, 0xff, 
	0xef, 0x7f, 0xff, 0xfd, 0xbb, 0xef, 0xbf, 0xff, 
	0xfb, 0xe7, 0x5f, 0xff, 0xfd, 0xff, 0xff, 0x7f, 
	0x5f, 0xff, 0xff, 0xf7, 0xf7, 0x7f, 0xff, 0x7f, 
	0xf9, 0xff, 0xfb, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xef, 0xfe, 0xbd, 0xfd, 0xff, 0xff, 0x7f, 0xff, 
	0xff, 0xdd, 0xf7, 0xff, 0xff, 0xff, 0xff, 0xdf, 
	0xdf, 0x7f, 0xef, 0x3f, 0xe6, 0xff, 0xdf, 0xef, 
	0x3f, 0xf9, 0xe4, 0xff, 0xfb, 0xbe, 0xff, 0xf3, 
	0xcb, 0xff, 0xff, 0xff, 0xff, 0xf7, 0xdf, 0xfe, 
	0xff, 0xff, 0xff, 0xff, 0xfe, 0xf7, 0xa9, 0xff, 
	0x8b, 0x7f, 0xff, 0xba, 0xe3, 0xad, 0x3b, 0xfe, 
	0xea, 0xab, 0x2d, 0xbf, 0xe8, 0xfa, 0xeb, 0x5f, 
	0x7d, 0xdc, 0x35, 0xc3, 0x5e, 0x7b, 0xfc, 0xf5, 
	0xd7, 0x5f, 0x77, 0xfd, 0xbf, 0xfe, 0xbf, 0xff, 
	0xef, 0xbf, 0xbf, 0xff, 0xec, 0xf7, 0xef, 0xff, 
	0xfe, 0xef, 0xdf, 0x7f, 0x7e, 0xfa, 0xd9, 0xfd, 
	0xff, 0x7f, 0xfc, 0xd9, 0xfd, 0xff, 0xfd, 0xfd, 
	0x3f, 0xd1, 0xfe, 0xff, 0xff, 0xfb, 0xe9, 0xef, 
	0xfe, 0xf9, 0xb7, 0xdf, 0x7b, 0xff, 0xfe, 0xfb, 
	0xff, 0x7e, 0x7d, 0xfd, 0xf7, 0xef, 0xfe, 0xff, 
	0xfd, 0xf7, 0xff, 0xe7, 0x7e, 0xe9, 0xff, 0xff, 
	0xff, 0xce, 0x3f, 0xfd, 0x77, 0xff, 0xfd, 0xff, 
	0xdf, 0x7f, 0xff, 0xff, 0xef, 0xf7, 0xd9, 0xef, 
	0xff, 0x7f, 0x7b, 0xd6, 0xff, 0xff, 0xf7, 0xfe, 
	0xf9, 0x32, 0xff, 0xb7, 0xff, 0xfe, 0x5c, 0x63, 
	0x3c, 0xff, 0xdf, 0xdc, 0xef, 0xad, 0xff, 0xff, 
	0xff, 0xff, 0x59, 0x6f, 0x2c, 0xf7, 0xf7, 0xef, 
	0xef, 0xfd, 0xff, 0xdf, 0xff, 0x1f, 0xf6, 0xbf, 
	0xff, 0xff, 0xff, 0xdf, 0xbf, 0xeb, 0xff, 0xff, 
	0xee, 0xef, 0xfd, 0xbf, 0xfb, 0xef, 0x7f, 0xf3, 
	0xfb, 0xef, 0x7f, 0xfb, 0xff, 0xfd, 0xfd, 0xd7, 
	0xff, 0xff, 0xf9, 0xff, 0xff, 0xdb, 0x3f, 0xfd, 
	0xef, 0xff, 0xff, 0xbe, 0x7f, 0xee, 0xfe, 0xff, 
	0xff, 0xfe, 0xef, 0xff, 0xff, 0xff, 0xff, 0xf9, 
	0xff, 0xff, 0xff, 0xfb, 0xff, 0xbf, 0xea, 0xff, 
	0x7f, 0xff, 0x3f, 0xef, 0xf7, 0x6b, 0xff, 0x3f, 
	0xfe, 0xf2, 0xfd, 0xff, 0xfc, 0xff, 0x7f, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xfe, 0xf7, 
	0xff, 0xff, 0x82, 0xbb, 0xfd, 0xff, 0xff, 0xfb, 
	0xff, 0xf7, 0xff, 0x7f, 0x6f, 0xff, 0xff, 0xfd, 
	0x77, 0xbe, 0x7a, 0xfe, 0xff, 0xbf, 0xfe, 0xbd, 
	0xfe, 0xfd, 0xaf, 0xff, 0xff, 0xfd, 0xef, 0xfe, 
	0xff, 0xff, 0xff, 0xfd, 0xf7, 0xff, 0xff, 0xaf, 
	0xf7, 0xcf, 0x0f, 0xee, 0xbf, 0x47, 0x9f, 0x7f, 
	0xfe, 0xbf, 0xff, 0xff, 0xfb, 0xef, 0x7f, 0xe7, 
	0x9d, 0xfe, 0x3f, 0xf9, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xdf, 0x6f, 0xff, 0x7d, 0xad, 0xfd, 0xe7, 
	0xff, 0xfd, 0x7e, 0xfd, 0xf7, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xdf, 0x7e, 0xfb, 0xff, 0x6b, 
	0xde, 0xaa, 0xff, 0xbf, 0xfe, 0xfa, 0xeb, 0xf7, 
	0xbf, 0xda, 0xff, 0xdf, 0x77, 0xff, 0xbb, 0xff, 
	0xee, 0xfb, 0xee, 0xbf, 0xef, 0xfe, 0xff, 0xdf, 
	0xfb, 0xef, 0xbf, 0x0a, 0xfb, 0xbe, 0xff, 0xfe, 
	0xf3, 0xdf, 0x7b, 0xce, 0xf7, 0xf3, 0xff, 0xff, 
	0xf7, 0xfd, 0xbf, 0xff, 0xee, 0x7f, 0xf9, 0xff, 
	0xff, 0xff, 0xff, 0xfd, 0xb7, 0xff, 0xff, 0x67, 
	0xfe, 0xff, 0xfa, 0xfb, 0xef, 0xff, 0xff, 0xff, 
	0xff, 0xf7, 0xff, 0xff, 0xfb, 0xff, 0xfd, 0xff, 
	0xfd, 0xd7, 0xfe, 0xff, 0xff, 0xff, 0xdf, 0xf7, 
	0xdf, 0xff, 0xff, 0x3f, 0xe1, 0x9f, 0xef, 0xff, 
	0xef, 0xd4, 0xfb, 0xff, 0x7f, 0xfc, 0xbe, 0x5b, 
	0xff, 0x9f, 0xfc, 0xfd, 0xf7, 0xbf, 0x7f, 0xdf, 
	0xf9, 0x97, 0xdf, 0xff, 0x7f, 0xfd, 0xff, 0xbf, 
	0xe9, 0x7f, 0xbe, 0xff, 0xbf, 0x9b, 0xef, 0xff, 
	0xff, 0xf3, 0xfd, 0x6e, 0xfe, 0xf9, 0xff, 0x3b, 
	0xdf, 0xff, 0xff, 0x7f, 0xf7, 0xdf, 0xff, 0xff, 
	0xff, 0xf5, 0xff, 0x7f, 0x8b, 0xfe, 0x7f, 0xff, 
	0xfd, 0xf6, 0xdf, 0xf9, 0xff, 0xcf, 0xfe, 0xff, 
	0xff, 0xff, 0xfb, 0xaf, 0xff, 0xff, 0xaf, 0x25, 
	0x77, 0xde, 0xff, 0xfd, 0xff, 0xe7, 0xff, 0xff, 
	0x8f, 0xde, 0xff, 0xff, 0x3d, 0xfb, 0xff, 0xf3, 
	0xfe, 0xff, 0xfb, 0xfc, 0xff, 0xff, 0xff, 0xef, 
	0xff, 0xff, 0xff, 0xfd, 0xf7, 0xdf, 0xff, 0xff, 
	0xff, 0x9f, 0xff, 0xe7, 0xb5, 0xbb, 0x6f, 0xc7, 
	0xbf, 0xef, 0xd6, 0xf1, 0xf3, 0x7f, 0xaf, 0xbd, 
	0xf7, 0xdf, 0xff, 0xff, 0xf8, 0xef, 0xbf, 0xff, 
	0xde, 0x3b, 0xfd, 0xbf, 0xff, 0x7e, 0xfa, 0xef, 
	0xef, 0xe7, 0x7e, 0xfd, 0xff, 0xbf, 0x7f, 0xff, 
	0xdf, 0xf6, 0xbf, 0xfa, 0xee, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0x3d, 0x4f, 0xf4, 0xff, 
	0xff, 0xff, 0xf5, 0xff, 0x3f, 0x8b, 0xfb, 0xff, 
	0xf5, 0xdf, 0x5a, 0x7e, 0xb5, 0xf7, 0xdb, 0xee, 
	0x6b, 0xf5, 0xfe, 0xff, 0xdf, 0xff, 0xff, 0xff, 
	0x1d, 0xf7, 0xdf, 0xff, 0xed, 0xe7, 0xf7, 0xff, 
	0xff, 0xa6, 0xee, 0xff, 0x91, 0x3f, 0x7e, 0x79, 
	0xfc, 0xbc, 0x7f, 0x78, 0xaf, 0xf4, 0xff, 0x7f, 
	0xfd, 0xfe, 0xf9, 0xf7, 0x9d, 0xf7, 0xde, 0xfb, 
	0x2f, 0xfe, 0x1f, 0xfe, 0xfd, 0x73, 0x72, 0xff, 
	0x9f, 0xfc, 0xfe, 0xe1, 0xc7, 0xb7, 0xdd, 0x7f, 
	0xff, 0x8d, 0xf7, 0xff, 0xef, 0xff, 0xfb, 0x6f, 
	0xbe, 0xed, 0xff, 0xdf, 0xee, 0xff, 0xed, 0xb3, 
	0xfb, 0x57, 0xe5, 0x7d, 0x9f, 0xf9, 0xdf, 0xfb, 
	0xe7, 0xff, 0x7f, 0xfe, 0xf9, 0xe7, 0x9f, 0xff, 
	0xbf, 0xf3, 0xcf, 0x3f, 0xf3, 0xcc, 0xff, 0xdf, 
	0x3c, 0xff, 0xcf, 0xb3, 0xcf, 0xdf, 0x23, 0xfb, 
	0x3b, 0xd6, 0xfe, 0x7b, 0x8d, 0x3f, 0xff, 0x78, 
	0xc3, 0x4d, 0x3f, 0xde, 0x5e, 0xe7, 0x9f, 0x7f, 
	0xfe, 0xf9, 0xf6, 0xda, 0x7f, 0xfc, 0xfd, 0xe6, 
	0x9d, 0xff, 0xf6, 0xee, 0xdf, 0x77, 0xff, 0x7f, 
	0xf7, 0xfd, 0xff, 0xbf, 0xff, 0xf3, 0xff, 0x6f, 
	0xff, 0xfd, 0xff, 0xff, 0xff, 0xfe, 0xf9, 0xef, 
	0xff, 0xff, 0xff, 0xfd, 0xfe, 0xff, 0x33, 0xbb, 
	0xef, 0xbf, 0xff, 0xff, 0xfb, 0xff, 0xff, 0xff, 
	0xff, 0xf7, 0xfb, 0x7f, 0xff, 0xfb, 0xfd, 0xf7, 
	0x5f, 0x7f, 0xfd, 0xdf, 0xff, 0xff, 0xff, 0xff, 
	0xfd, 0xf7, 0xef, 0xec, 0x7e, 0x2f, 0x3f, 0xb7, 
	0xf7, 0xcb, 0xcf, 0x7f, 0xbe, 0xf4, 0xc3, 0x9f, 
	0xbf, 0xff, 0xe1, 0x8f, 0x17, 0x5f, 0xf8, 0xed, 
	0xb5, 0x3f, 0x7e, 0xff, 0xf5, 0xef, 0xbf, 0xab, 
	0xf7, 0x7f, 0xfe, 0xff, 0xae, 0xbf, 0xef, 0xff, 
	0x7b, 0xe3, 0x9b, 0xff, 0xbe, 0xfd, 0xcd, 0x7f, 
	0x5f, 0xbc, 0xfb, 0xdd, 0x56, 0xf9, 0xff, 0xf5, 
	0xdf, 0x7f, 0x7f, 0x3f, 0xfe, 0xff, 0xef, 0x7f, 
	0xef, 0xfd, 0xff, 0xfe, 0xef, 0x6b, 0xaf, 0xf4, 
	0xfd, 0x7b, 0xdf, 0xf9, 0xff, 0xfb, 0xdf, 0xdf, 
	0xff, 0xff, 0xf7, 0x3f, 0xbf, 0xff, 0xff, 0xff, 
	0xf8, 0xff, 0xf7, 0xfc, 0xb3, 0xcf, 0x3d, 0xff, 
	0xfc, 0xb3, 0x8d, 0x36, 0xff, 0xf8, 0xd3, 0x96, 
	0x7b, 0xfe, 0xf9, 0xc7, 0x9f, 0x7f, 0xfe, 0xf9, 
	0xe6, 0x9b, 0x7f, 0xfe, 0xee, 0xff, 0xff, 0xff, 
	0xbf, 0xfe, 0xff, 0x9f, 0x7f, 0xff, 0xfe, 0xf3, 
	0xff, 0x3f, 0xbe, 0xeb, 0xef, 0x17, 0xff, 0xfe, 
	0xff, 0xf7, 0xbf, 0x7f, 0xfa, 0xe3, 0xef, 0xff, 
	0xa3, 0xff, 0xff, 0xfe, 0xff, 0xea, 0xbf, 0xff, 
	0xff, 0xff, 0xff, 0xda, 0xff, 0xbf, 0xff, 0xff, 
	0xbd, 0x5f, 0xff, 0xff, 0xbf, 0xf7, 0xfe, 0xff, 
	0x7d, 0xff, 0xbf, 0xff, 0xef, 0xfe, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xfe, 0xf7, 0xef, 0xff, 
	0xff, 0xff, 0xfb, 0xbf, 0xfd, 0xfd, 0xff, 0xe5, 
	0xff, 0x7f, 0xff, 0xfb, 0xef, 0xff, 0xff, 0xff, 
	0x7f, 0xf9, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xf7, 0x7f, 0xff, 0xf9, 0xf7, 0xdf, 0xdb, 0xff, 
	0xf7, 0xff, 0xbf, 0xff, 0xff, 0x7f, 0xff, 0xff, 
	0xff, 0xfd, 0xff, 0xef, 0xff, 0xea, 0xff, 0xff, 
	0xff, 0xef, 0xf9, 0xef, 0x9a, 0xff, 0xae, 0xfc, 
	0xf2, 0xbe, 0xfb, 0xfe, 0xff, 0xff, 0xfd, 0xf7, 
	0x7f, 0xf3, 0xdf, 0xfe, 0xff, 0x7f, 0xfb, 0xff, 
	0xff, 0x97, 0xff, 0xff, 0xfa, 0xeb, 0xaf, 0xbf, 
	0xde, 0xd8, 0xab, 0xaf, 0x9b, 0x7e, 0xea, 0x6b, 
	0x5f, 0x7f, 0xfd, 0xd5, 0xc7, 0x5d, 0x7f, 0xfd, 
	0xf5, 0xd7, 0x5f, 0x3f, 0xfd, 0xaf, 0xf6, 0xff, 
	0xfc, 0xf3, 0xcb, 0x3b, 0xbf, 0xfb, 0xe3, 0xc7, 
	0x7f, 0xff, 0xef, 0xdf, 0x9f, 0x6f, 0xf6, 0xaf, 
	0xe7, 0xb6, 0xff, 0xf6, 0xdf, 0x87, 0x07, 0x7d, 
	0xf6, 0x3f, 0xf1, 0xff, 0xff, 0xfe, 0xf7, 0xdf, 
	0xff, 0xff, 0xfb, 0xe6, 0xbf, 0xdf, 0xff, 0xff, 
	0x6f, 0x64, 0xff, 0xff, 0xff, 0xbf, 0xdf, 0xfe, 
	0xfe, 0xf3, 0xc7, 0xbf, 0xfc, 0x7e, 0xe2, 0xff, 
	0x6b, 0xaf, 0xbf, 0xfe, 0xba, 0xff, 0xdb, 0x79, 
	0xb7, 0xff, 0x3d, 0xff, 0x4f, 0xef, 0xfb, 0xff, 
	0x5e, 0xeb, 0xef, 0xfe, 0xff, 0xbf, 0xbb, 0xfe, 
	0xfe, 0xff, 0x22, 0xff, 0xaf, 0xbf, 0xf6, 0xfa, 
	0xe3, 0x0d, 0xdf, 0x57, 0xbb, 0x71, 0xfd, 0xf7, 
	0xfd, 0xe7, 0xbf, 0xfb, 0x5e, 0xfe, 0xbd, 0xdf, 
	0xef, 0xef, 0x9f, 0xfd, 0xff, 0x6f, 0x8f, 0xfe, 
	0xf7, 0xfe, 0xf6, 0xdb, 0xdf, 0x7e, 0xff, 0xff, 
	0xfb, 0xfd, 0xef, 0x7f, 0xff, 0xf6, 0xf7, 0xf7, 
	0xff, 0xff, 0xbf, 0xb7, 0xff, 0xff, 0xff, 0xbe, 
	0xf3, 0xff, 0xdd, 0xfb, 0xfb, 0xfa, 0x7f, 0xfd, 
	0xff, 0xaf, 0xff, 0xff, 0x9b, 0xfb, 0xef, 0xbf, 
	0xff, 0x7f, 0xff, 0xff, 0xe3, 0xff, 0xfd, 0xfb, 
	0xdf, 0xff, 0xdd, 0x6e, 0xff, 0xff, 0xb7, 0xa4, 
	0xfd, 0xfd, 0xff, 0xbf, 0xf7, 0xfa, 0x3f, 0xfe, 
	0xb7, 0xff, 0xff, 0xdf, 0xff, 0x7d, 0xca, 0xff, 
	0xf7, 0xff, 0xff, 0xff, 0xdf, 0xff, 0x9f, 0xff, 
	0xbf, 0xff, 0xfe, 0x89, 0xff, 0x7f, 0xff, 0xff, 
	0xcb, 0xff, 0xfb, 0xf7, 0xef, 0x7b, 0xff, 0xff, 
	0xff, 0xff, 0xfd, 0xff, 0xdf, 0xfd, 0x6f, 0xff, 
	0xfe, 0xfe, 0xfb, 0xbd, 0x2f, 0xfb, 0xff, 0x2f, 
	0xfe, 0xfe, 0xfb, 0xff, 0xff, 0x77, 0x3b, 0x7e, 
	0x78, 0xd7, 0xcf, 0x3c, 0xee, 0xef, 0xc7, 0x0f, 
	0xfe, 0xff, 0x7f, 0xe3, 0xff, 0x7f, 0xfc, 0x7f, 
	0xe3, 0xff, 0xff, 0x3f, 0xba, 0xff, 0xff, 0x7f, 
	0xff, 0xff, 0xbf, 0xf7, 0xef, 0x7f, 0xbf, 0xfd, 
	0xf7, 0xff, 0x7e, 0xff, 0xfb, 0xdf, 0xfd, 0x5f, 
	0xb7, 0x5f, 0xf7, 0xff, 0x9f, 0xff, 0xff, 0xff, 
	0xee, 0xff, 0xdf, 0xff, 0xff, 0xff, 0xd7, 0xff, 
	0x27, 0xdd, 0xff, 0xf7, 0xef, 0xf7, 0x7f, 0x95, 
	0x6f, 0xd9, 0x7e, 0xff, 0xed, 0x9f, 0xbf, 0xff, 
	0xeb, 0xfb, 0xcf, 0xff, 0x17, 0xff, 0xff, 0xfd, 
	0x77, 0xdb, 0xff, 0xaf, 0xfe, 0xfe, 0xff, 0xdd, 
	0x8f, 0xff, 0x65, 0xbf, 0x7f, 0xbf, 0xfe, 0xfb, 
	0xef, 0xff, 0xfd, 0xbf, 0xff, 0xbf, 0xbf, 0xff, 
	0xeb, 0xfe, 0xff, 0xff, 0xff, 0xaf, 0xff, 0xff, 
	0xef, 0x7d, 0xbf, 0xff, 0xff, 0xff, 0xff, 0x7f, 
	0xff, 0xff, 0xff, 0xfd, 0xf7, 0xff, 0xbf, 0xff, 
	0xff, 0xff, 0xff, 0xf7, 0x7f, 0xe9, 0x77, 0x2e, 
	0xbf, 0xfe, 0xf9, 0xc9, 0xcf, 0xbf, 0xff, 0xfc, 
	0xc1, 0x0f, 0x3f, 0xfb, 0xed, 0xd3, 0x0f, 0x7e, 
	0xf0, 0xd1, 0x3b, 0xdf, 0x7c, 0xf7, 0xe1, 0x77, 
	0x7f, 0x87, 0x3f, 0xbb, 0x7f, 0xfe, 0xcb, 0x3e, 
	0x9f, 0xff, 0xfe, 0xf3, 0x7f, 0x3d, 0xf5, 0xfe, 
	0x2d, 0xdf, 0x74, 0x7f, 0xcf, 0x3f, 0x7f, 0xfe, 
	0xff, 0xfb, 0xf5, 0xff, 0xff, 0xb2, 0xfd, 0xe5, 
	0xfe, 0xfb, 0x6f, 0xf9, 0xf9, 0xff, 0xfb, 0xf7, 
	0xff, 0xe5, 0xf3, 0xff, 0xfd, 0xbf, 0xd3, 0xff, 
	0x3f, 0xff, 0xff, 0xf3, 0xff, 0x3f, 0xed, 0xff, 
	0xff, 0x7d, 0xfe, 0xbf, 0xfd, 0x7f, 0xc9, 0x6f, 
	0xe7, 0xfa, 0x37, 0x7b, 0xff, 0xd7, 0xfa, 0x6f, 
	0xde, 0x58, 0x7f, 0xf7, 0x77, 0xbe, 0xff, 0xfe, 
	0x3f, 0xf5, 0x9a, 0xff, 0xff, 0x7d, 0x79, 0xff, 
	0xf6, 0xff, 0xa9, 0xbf, 0xcd, 0xf9, 0xff, 0xef, 
	0xff, 0xff, 0xfa, 0xff, 0xdc, 0x6f, 0x3d, 0xfd, 
	0xff, 0xbf, 0xfe, 0xfb, 0xff, 0xf6, 0xdf, 0xfe, 
	0xff, 0x57, 0xea, 0xfe, 0x6f, 0xff, 0xff, 0xf3, 
	0xdb, 0xff, 0x7f, 0xbc, 0xf6, 0xdf, 0x7f, 0xf7, 
	0xfd, 0x4f, 0xbe, 0xff, 0x7e, 0xab, 0xed, 0xbf, 
	0xff, 0x56, 0xfb, 0xbd, 0xbe, 0xbf, 0x97, 0xf9, 
	0xbf, 0xfd, 0xff, 0x5e, 0x6d, 0xff, 0xff, 0xff, 
	0xea, 0xff, 0xef, 0xc6, 0xff, 0xdf, 0x7f, 0xff, 
	0xfb, 0xef, 0xae, 0xff, 0xff, 0x7f, 0xcd, 0xfc, 
	0xff, 0xff, 0xc4, 0xe6, 0xff, 0xfe, 0xdf, 0x6b, 
	0xbc, 0xf7, 0xfe, 0xff, 0xfb, 0xfd, 0xff, 0x5e, 
	0xef, 0x57, 0x5b, 0x7d, 0xfd, 0xbf, 0xfd, 0xff, 
	0xad, 0xbf, 0xf4, 0xd7, 0xff, 0xfd, 0x9b, 0xb9, 
	0xff, 0xdb, 0xff, 0xbf, 0xd5, 0x16, 0xfb, 0x1f, 
	0xe5, 0xc4, 0xfb, 0x4a, 0x7f, 0x5f, 0x0b, 0xed, 
	0xb5, 0xdf, 0xdf, 0xff, 0xff, 0xdf, 0xf7, 0xff, 
	0xfb, 0xbf, 0xef, 0xea, 0xbd, 0x7f, 0xff, 0xff, 
	0x67, 0xdf, 0xff, 0xff, 0xf1, 0xfd, 0xf7, 0xdf, 
	0xf8, 0xbb, 0xdb, 0xef, 0xff, 0xff, 0xf7, 0xdf, 
	0xae, 0xbf, 0xfe, 0xb7, 0xef, 0xef, 0xdd, 0xa3, 
	0xf6, 0xf7, 0xff, 0xff, 0x7f, 0xff, 0xff, 0xe6, 
	0x9d, 0xe7, 0x9f, 0x7f, 0xfe, 0x5f, 0xce, 0x3b, 
	0xff, 0xff, 0xff, 0xce, 0x3f, 0xff, 0xfc, 0xbf, 
	0xdf, 0x3f, 0x7f, 0xc5, 0xec, 0xff, 0xfe, 0xfb, 
	0xef, 0xbe, 0xff, 0x5e, 0xf3, 0xcd, 0x3f, 0xff, 
	0xfc, 0xbb, 0x1d, 0x7f, 0xfc, 0xfd, 0xf6, 0x5b, 
	0x7f, 0xfc, 0xf9, 0xf6, 0xdd, 0x6f, 0xbc, 0x9b, 
	0x9a, 0xff, 0xdf, 0xff, 0xff, 0xfd, 0x77, 0xff, 
	0xfd, 0xfe, 0xff, 0xff, 0xff, 0xff, 0xf3, 0xff, 
	0xff, 0x7f, 0xfb, 0xfd, 0xff, 0xef, 0xbf, 0xfd, 
	0xef, 0xdb, 0xdf, 0xce, 0xe1, 0xfe, 0xff, 0xff, 
	0xff, 0xff, 0xbf, 0xff, 0xfb, 0xff, 0xfb, 0xef, 
	0xbf, 0xff, 0xef, 0xff, 0xff, 0xff, 0xfb, 0xff, 
	0xff, 0xff, 0x7f, 0xff, 0xf7, 0xff, 0x7f, 0xbf, 
	0x93, 0xfa, 0xfd, 0xfd, 0xff, 0xde, 0x7f, 0x37, 
	0xff, 0xfa, 0xd2, 0xdf, 0x2f, 0xfc, 0xf7, 0xa7, 
	0xbf, 0x7e, 0x6b, 0xbf, 0xd7, 0x3f, 0xfb, 0xfc, 
	0xbd, 0xbf, 0x1f, 0xff, 0x2e, 0xd8, 0xff, 0xfb, 
	0xff, 0xbb, 0xfd, 0xf6, 0xeb, 0xef, 0x8b, 0xfd, 
	0xbe, 0xf8, 0xef, 0x17, 0xff, 0xfd, 0xf5, 0xdf, 
	0x3f, 0xff, 0xfd, 0xf7, 0xc5, 0x56, 0xdd, 0xfc, 
	0xbd, 0xeb, 0xff, 0xbf, 0xff, 0xaf, 0xff, 0xff, 
	0xff, 0xdf, 0xef, 0xfd, 0xff, 0x4e, 0xeb, 0xf5, 
	0xeb, 0xef, 0xf7, 0xf7, 0xfa, 0xe2, 0xf7, 0xff, 
	0xdf, 0xfe, 0xff, 0xf5, 0xff, 0xa2, 0xff, 0xff, 
	0xf3, 0x8f, 0x3e, 0xff, 0x7c, 0xf3, 0xcf, 0x36, 
	0xbf, 0x78, 0xa3, 0x4f, 0x6f, 0xbe, 0xd9, 0x67, 
	0x9b, 0x6f, 0xfe, 0xb9, 0x47, 0x1f, 0x7d, 0xfc, 
	0xf9, 0xaf, 0xff, 0xae, 0xff, 0xff, 0xfb, 0xeb, 
	0xff, 0xff, 0xff, 0xe3, 0xdf, 0x3f, 0xbf, 0xfc, 
	0xe6, 0xff, 0xff, 0xff, 0xff, 0xff, 0xbf, 0xfe, 
	0xf8, 0xe9, 0xb7, 0x9f, 0xff, 0x9f, 0xfe, 0xbf, 
	0xfa, 0xff, 0xaf, 0xaf, 0xfe, 0xfb, 0xef, 0xfb, 
	0x7f, 0xef, 0xfc, 0xfa, 0x97, 0xff, 0xff, 0xef, 
	0xf7, 0xff, 0xff, 0xfe, 0xff, 0xf5, 0xdf, 0xdf, 
	0xff, 0x7f, 0xfa, 0xff, 0xfa, 0xff, 0xff, 0xbf, 
	0xfe, 0xff, 0xff, 0xff, 0xff, 0xef, 0xef, 0x9f, 
	0xff, 0xfd, 0xf7, 0xff, 0x7f, 0xff, 0xfd, 0xff, 
	0xff, 0xbf, 0x7f, 0xff, 0xf9, 0xff, 0xe2, 0xff, 
	0xeb, 0xff, 0xff, 0xff, 0xfa, 0xff, 0xff, 0xff, 
	0xf7, 0x9f, 0xff, 0xff, 0x7f, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xdf, 0xfe, 0xfd, 
	0xbd, 0xff, 0x8f, 0xff, 0x6f, 0xfc, 0xff, 0xe6, 
	0xdb, 0x6f, 0xff, 0xfb, 0xf6, 0xfe, 0x4f, 0xfe, 
	0xff, 0xff, 0xff, 0xdf, 0xfe, 0xfb, 0xff, 0xff, 
	0xfb, 0x7f, 0xfa, 0xdd, 0xff, 0xff, 0x2f, 0xfe, 
	0xff, 0x6b, 0xaf, 0xba, 0xf8, 0x62, 0xeb, 0xaf, 
	0xbb, 0xf4, 0x7a, 0xeb, 0xaf, 0x7f, 0xfd, 0x71, 
	0xd7, 0x5f, 0x7f, 0xfd, 0xf5, 0x57, 0x5d, 0x77, 
	0xfd, 0xe5, 0x7f, 0xd9, 0xbf, 0xe0, 0xcf, 0x6f, 
	0x3f, 0xfb, 0xe2, 0xcf, 0x2e, 0xfa, 0xfc, 0x9f, 
	0xff, 0xf6, 0xf6, 0xfd, 0xfd, 0x75, 0xff, 0xfd, 
	0xdf, 0xff, 0x1f, 0xfd, 0xf7, 0xf1, 0xff, 0xcd, 
	0x7f, 0xff, 0xef, 0x7f, 0x7f, 0xff, 0xfb, 0xe6, 
	0x5f, 0xfc, 0xfb, 0xfd, 0xf7, 0x7b, 0xdf, 0xff, 
	0xfe, 0xff, 0xff, 0xef, 0xff, 0xff, 0x77, 0xff, 
	0xff, 0xfd, 0xeb, 0xb5, 0xff, 0xaf, 0xfd, 0xfe, 
	0xdf, 0xfd, 0xdf, 0x75, 0xd7, 0xff, 0x7b, 0xf7, 
	0xfd, 0xdf, 0xbf, 0xea, 0xff, 0xfb, 0xff, 0xff, 
	0xfd, 0xff, 0x7e, 0xaf, 0xff, 0xef, 0xff, 0xcf, 
	0xf4, 0xf7, 0xfc, 0xff, 0xef, 0xbf, 0xaf, 0xff, 
	0xfa, 0xef, 0xfc, 0x3d, 0xba, 0xfa, 0xfb, 0xff, 
	0x4f, 0xf4, 0xe7, 0xdf, 0xff, 0xff, 0xf7, 0xfd, 
	0xd5, 0xff, 0xf7, 0xdf, 0xfa, 0xff, 0xff, 0xff, 
	0xe7, 0xef, 0x7f, 0xfb, 0xef, 0xfb, 0xdf, 0x8f, 
	0xff, 0xeb, 0xfb, 0xe7, 0xff, 0xfb, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xdf, 0x7f, 0xdf, 0xff, 0xff, 
	0xe0, 0xef, 0xe3, 0xbf, 0xf5, 0xff, 0xba, 0xff, 
	0xe2, 0xbb, 0xf6, 0xfb, 0xeb, 0xff, 0xff, 0xfd, 
	0xee, 0x9f, 0xdf, 0x7b, 0x7f, 0xff, 0xff, 0xff, 
	0x7e, 0xdf, 0xff, 0xfd, 0xab, 0xff, 0xad, 0x3f, 
	0xff, 0xff, 0xff, 0xaf, 0xed, 0xe7, 0xfc, 0xf3, 
	0x47, 0xff, 0x99, 0x90, 0xff, 0xbf, 0xfb, 0xff, 
	0xbf, 0xff, 0x5f, 0xfe, 0xff, 0x7f, 0xff, 0xf7, 
	0x2f, 0xfe, 0xfd, 0xff, 0xfe, 0xff, 0x7f, 0xf9, 
	0xd3, 0x5f, 0xef, 0xb4, 0xfd, 0xfe, 0xff, 0xfb, 
	0xff, 0xbf, 0xff, 0x3f, 0xff, 0xfe, 0xff, 0xff, 
	0xdf, 0x79, 0xfe, 0xed, 0x3b, 0xf9, 0xdf, 0xfc, 
	0xff, 0x4f, 0xdf, 0x7f, 0xfc, 0xe1, 0xff, 0xfc, 
	0xbf, 0xaf, 0xe3, 0xff, 0xf7, 0xff, 0xe1, 0xff, 
	0xff, 0xf7, 0xfb, 0xef, 0xc3, 0xdd, 0xdf, 0xf3, 
	0xff, 0xae, 0xff, 0xdf, 0xff, 0xbb, 0xfd, 0xff, 
	0xf3, 0x7f, 0xff, 0xff, 0xff, 0xff, 0xbf, 0xff, 
	0xff, 0xff, 0xbf, 0xfb, 0xff, 0xf7, 0xff, 0xff, 
	0xbf, 0x7d, 0xff, 0xff, 0xff, 0x9a, 0xfb, 0x1b, 
	0xb7, 0x56, 0x53, 0xff, 0xb7, 0xdb, 0x6f, 0xfa, 
	0xeb, 0xbe, 0x9e, 0xfe, 0xf5, 0xbf, 0x7e, 0xff, 
	0xbd, 0x7f, 0xff, 0x7b, 0x3f, 0xbd, 0xef, 0xff, 
	0xff, 0x5a, 0xec, 0xdf, 0xfe, 0xff, 0xc7, 0xaf, 
	0xfe, 0xfe, 0xff, 0xaf, 0xbf, 0xff, 0xfc, 0xf2, 
	0xdf, 0xfb, 0xff, 0xf8, 0xf3, 0xff, 0xff, 0xb7, 
	0xfd, 0xf7, 0xbf, 0xfd, 0xff, 0xbd, 0xf9, 0xff, 
	0xfc, 0xff, 0xbf, 0xfe, 0xff, 0xfa, 0xff, 0xbf, 
	0xfb, 0xfe, 0xea, 0xff, 0x7f, 0xff, 0xff, 0xfb, 
	0xef, 0xff, 0xff, 0xef, 0xb5, 0xee, 0xff, 0xef, 
	0xff, 0xff, 0x8e, 0xff, 0xbf, 0x3f, 0xe6, 0xfb, 
	0xff, 0x3b, 0x7f, 0xbe, 0xff, 0xfb, 0xff, 0x7c, 
	0xfc, 0xfd, 0xef, 0x7f, 0xf9, 0xc9, 0xe3, 0x1d, 
	0x7e, 0xde, 0x64, 0x95, 0xcf, 0xff, 0x8c, 0xff, 
	0x7f, 0xbf, 0x5b, 0xff, 0xfb, 0x6f, 0xff, 0xf3, 
	0x7e, 0x79, 0xff, 0xff, 0xf7, 0xfd, 0xff, 0xff, 
	0xce, 0x37, 0x7f, 0x76, 0xf3, 0x7d, 0xf7, 0xff, 
	0xff, 0xff, 0x5d, 0xfa, 0xa7, 0x5c, 0xff, 0x9f, 
	0xf5, 0xbf, 0xfd, 0xff, 0xef, 0x2f, 0xa7, 0xf6, 
	0xef, 0xb7, 0xff, 0xe7, 0xff, 0xff, 0x7f, 0xdf, 
	0xff, 0x3f, 0xf7, 0x57, 0xff, 0xfd, 0xbf, 0x7a, 
	0xcf, 0x3a, 0xff, 0xff, 0xf3, 0xcf, 0xee, 0xf7, 
	0xe9, 0x27, 0xca, 0xfa, 0xff, 0xdd, 0xfe, 0xff, 
	0xff, 0xff, 0xff, 0x7f, 0x9f, 0x7f, 0xf6, 0xff, 
	0xff, 0xff, 0xd7, 0xe8, 0xfe, 0x3f, 0x65, 0xdf, 
	0xf1, 0xb6, 0xf9, 0x75, 0xbf, 0x5e, 0xb6, 0xf9, 
	0x7f, 0x6f, 0xad, 0xfe, 0xff, 0xff, 0xfb, 0xff, 
	0xbd, 0xdf, 0xca, 0x7e, 0xcd, 0xfe, 0xbf, 0xbf, 
	0xfb, 0xe6, 0x99, 0x1d, 0xff, 0xcb, 0xe6, 0xdf, 
	0xff, 0xaa, 0xb9, 0xef, 0xfe, 0xff, 0xb5, 0xfb, 
	0x7f, 0xad, 0xff, 0xff, 0x77, 0xff, 0x35, 0xbb, 
	0xbe, 0xfa, 0xff, 0xec, 0xee, 0xbf, 0xce, 0xdf, 
	0xff, 0xfc, 0x97, 0xb6, 0x7f, 0xe9, 0xaf, 0xbf, 
	0xff, 0xff, 0xf7, 0xf7, 0xaf, 0x9f, 0xff, 0xfe, 
	0xd9, 0x2d, 0xb5, 0xf4, 0xd7, 0xff, 0xff, 0xdb, 
	0xbb, 0xff, 0x5b, 0xdd, 0x7f, 0xd4, 0xfd, 0xdf, 
	0xff, 0xad, 0xbf, 0xfc, 0xfe, 0xdf, 0xdf, 0xff, 
	0x9f, 0x7f, 0x7e, 0xf0, 0x67, 0xef, 0xbd, 0xd7, 
	0xff, 0xfb, 0xff, 0xef, 0xcf, 0xfd, 0x7f, 0x7f, 
	0xf3, 0x67, 0x37, 0x7f, 0xfc, 0xeb, 0xcf, 0xbf, 
	0xff, 0x7f, 0xff, 0xef, 0xff, 0xff, 0xed, 0xfe, 
	0xff, 0xef, 0xbf, 0xf9, 0xf6, 0xef, 0xff, 0xdf, 
	0xaf, 0xf7, 0xfd, 0xf7, 0xff, 0x77, 0xff, 0xfd, 
	0xfe, 0x1d, 0xff, 0xff, 0xf5, 0xff, 0xfd, 0xfe, 
	0xfc, 0xff, 0xed, 0xf3, 0xdf, 0x3f, 0xff, 0xfc, 
	0xbf, 0xfe, 0xff, 0x7f, 0x0d, 0xe4, 0xdb, 0x3e, 
	0xd3, 0xcf, 0x36, 0xef, 0xfe, 0x73, 0xed, 0x37, 
	0xfb, 0xf6, 0xf3, 0xdd, 0x7f, 0xbf, 0xdd, 0xe7, 
	0xdf, 0x7f, 0xb8, 0xf1, 0xb7, 0xdd, 0x7f, 0xff, 
	0x5b, 0xb8, 0x3f, 0xdf, 0xfc, 0xf7, 0xcf, 0xff, 
	0xdf, 0xff, 0xf7, 0xcf, 0xff, 0xff, 0xff, 0xf3, 
	0xfb, 0xdf, 0xff, 0xff, 0xff, 0xff, 0x7f, 0xfe, 
	0xfb, 0xef, 0xfb, 0xff, 0xcf, 0xea, 0xde, 0xfe, 
	0xfd, 0xf7, 0xdf, 0xff, 0xff, 0xff, 0xfb, 0xbf, 
	0xff, 0xff, 0xff, 0xf7, 0xff, 0xff, 0x7f, 0xff, 
	0xfd, 0xf7, 0xff, 0xfd, 0xfb, 0xdf, 0xff, 0xff, 
	0xbf, 0x8f, 0xfb, 0xff, 0xfd, 0xfc, 0x6b, 0xcf, 
	0xed, 0xbd, 0xf8, 0xfa, 0x8b, 0xef, 0xbf, 0xf9, 
	0xf5, 0xff, 0xff, 0xfb, 0xe1, 0xbf, 0x3f, 0xde, 
	0x78, 0xb1, 0xbd, 0xfe, 0xff, 0xba, 0xde, 0xbf, 
	0xfa, 0xfe, 0x9f, 0xff, 0xbf, 0x6b, 0xf2, 0xcb, 
	0xef, 0xbf, 0x7b, 0xf6, 0xd7, 0xff, 0xff, 0xfb, 
	0xc7, 0xff, 0xff, 0x3f, 0xfb, 0xf5, 0x5f, 0xff, 
	0xff, 0x3d, 0xfb, 0xdf, 0xbb, 0xef, 0x7f, 0xff, 
	0xfe, 0xfb, 0xa9, 0xbf, 0xff, 0xfe, 0xbe, 0xd9, 
	0x7f, 0xf7, 0xbf, 0xfe, 0x5a, 0xfb, 0xff, 0xb7, 
	0xff, 0x97, 0xdf, 0xf7, 0xff, 0xff, 0xe2, 0xff, 
	0xfb, 0xf3, 0xcf, 0x3f, 0xfe, 0xec, 0xa3, 0x8f, 
	0x3e, 0xf7, 0xf8, 0xe3, 0x4f, 0x7d, 0xfe, 0xf9, 
	0x67, 0x9f, 0x7f, 0xfe, 0xb9, 0x67, 0x9f, 0x7d, 
	0xfe, 0xf9, 0x8b, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0x8f, 0xaf, 0xbf, 0xfe, 0xfa, 0xeb, 0xff, 0xbf, 
	0xfd, 0xfd, 0xff, 0xff, 0xfc, 0xed, 0xcf, 0xff, 
	0xff, 0xfc, 0xfd, 0xbd, 0xbf, 0xff, 0x7f, 0xf8, 
	0xbf, 0xfa, 0xfe, 0xbf, 0xff, 0xbf, 0xfa, 0xfa, 
	0xef, 0xef, 0xff, 0xff, 0xf6, 0xf3, 0xff, 0xff, 
	0xfd, 0xf7, 0xff, 0xff, 0x7f, 0xff, 0xfd, 0xdf, 
	0xff, 0xff, 0xbf, 0xf8, 0x7f, 0xfb, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xbf, 0xff, 0xfe, 0xff, 
	0xfb, 0xff, 0xff, 0xff, 0xff, 0xdf, 0xff, 0xff, 
	0xff, 0xff, 0xdf, 0xdf, 0xff, 0xfb, 0xff, 0xe3, 
	0xff, 0xfb, 0xff, 0xff, 0xff, 0xf7, 0xff, 0xff, 
	0xff, 0xfe, 0xff, 0x7f, 0x6f, 0x3f, 0xff, 0xff, 
	0xff, 0x7f, 0xff, 0xff, 0xff, 0xf7, 0xff, 0xff, 
	0xfd, 0xef, 0xff, 0xbf, 0xff, 0x7f, 0xef, 0xff, 
	0xff, 0xd3, 0x6d, 0xff, 0xff, 0xee, 0xd3, 0x6f, 
	0xff, 0xff, 0xed, 0xff, 0x9f, 0xfe, 0xff, 0xff, 
	0x37, 0x5f, 0x7e, 0xfa, 0xed, 0xff, 0xff, 0x9f, 
	0xfe, 0xff, 0xeb, 0xaf, 0xbf, 0xde, 0xfa, 0xeb, 
	0x8f, 0x3b, 0xea, 0xba, 0xeb, 0xaf, 0x7f, 0xfc, 
	0xf5, 0x57, 0x5f, 0x7b, 0xdd, 0xf1, 0xd6, 0x5d, 
	0x7b, 0xfd, 0xf5, 0xff, 0xfb, 0x7f, 0xbe, 0xcd, 
	0xf3, 0xff, 0xff, 0xe4, 0xef, 0xf5, 0xff, 0xfc, 
	0xfe, 0xff, 0x96, 0xff, 0xff, 0xbb, 0xbf, 0xff, 
	0x7f, 0xf9, 0xe7, 0x9b, 0x9e, 0xed, 0xf9, 0xff, 
	0xcf, 0xfb, 0xff, 0xbf, 0xf5, 0xff, 0xbf, 0xfb, 
	0xfe, 0xfc, 0xff, 0x5a, 0x7f, 0xff, 0xfc, 0x7f, 
	0xff, 0x7f, 0xde, 0xf7, 0xdf, 0xfe, 0xfb, 0x3f, 
	0xbf, 0xff, 0xff, 0xfb, 0x85, 0xf7, 0xff, 0xbd, 
	0xff, 0xdf, 0xff, 0xff, 0xdf, 0x7a, 0x7f, 0xbe, 
	0xff, 0xff, 0xff, 0xff, 0xfd, 0xee, 0xbf, 0xff, 
	0xde, 0xd7, 0x5f, 0xff, 0xef, 0xff, 0xd7, 0xff, 
	0x4f, 0xfc, 0xe7, 0xff, 0xff, 0xd9, 0xaf, 0x3f, 
	0xfe, 0xf2, 0xcf, 0xfb, 0xbd, 0xfe, 0xdf, 0xf7, 
	0xff, 0x6f, 0xf4, 0xdf, 0xff, 0xff, 0x7f, 0xff, 
	0xfd, 0xff, 0xff, 0xff, 0xfd, 0xfb, 0xff, 0xaf, 
	0xff, 0xbf, 0x6f, 0xbe, 0xbd, 0xef, 0xfa, 0xff, 
	0xaf, 0xfb, 0xff, 0xea, 0xf6, 0xdf, 0xbf, 0xfd, 
	0xfd, 0xff, 0xff, 0x6f, 0x3f, 0xf7, 0xdf, 0xff, 
	0x77, 0xe6, 0xff, 0xef, 0x9f, 0x7f, 0xee, 0xb9, 
	0xef, 0xa7, 0xbf, 0xfe, 0xff, 0xfb, 0xff, 0xff, 
	0xff, 0xff, 0xd7, 0x7f, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xfb, 0xff, 0xbf, 0xdf, 0x97, 0xff, 0xff, 
	0xbf, 0xf7, 0xf6, 0xcb, 0x8f, 0x3f, 0xde, 0xdd, 
	0x3f, 0x7f, 0xff, 0xff, 0xfe, 0xfe, 0x1c, 0x7e, 
	0xfc, 0xfd, 0xd7, 0xff, 0xff, 0xf7, 0xff, 0xff, 
	0xff, 0x57, 0xee, 0xfd, 0xfd, 0xfb, 0x6d, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xde, 0xff, 0xff, 
	0xff, 0xff, 0xdb, 0xff, 0xff, 0xdf, 0xff, 0xff, 
	0xff, 0xff, 0x7f, 0xfe, 0xef, 0x37, 0xfa, 0xff, 
	0xff, 0xfd, 0xff, 0xfe, 0xfd, 0xff, 0x9e, 0x8f, 
	0xfe, 0xf3, 0xbe, 0xff, 0x94, 0xff, 0xff, 0xff, 
	0xff, 0x9b, 0xf7, 0xff, 0xf9, 0xff, 0x9f, 0xff, 
	0xff, 0xff, 0xe7, 0xfe, 0xff, 0xff, 0xff, 0xfb, 
	0xff, 0xff, 0xfe, 0xef, 0xfd, 0xff, 0xfb, 0xff, 
	0xfd, 0xfd, 0xff, 0x7f, 0xfd, 0xff, 0xfe, 0xff, 
	0xbf, 0xff, 0x7f, 0xfd, 0xff, 0xff, 0xab, 0xfb, 
	0xaf, 0xbf, 0xfe, 0xbb, 0xef, 0xb7, 0xff, 0xf6, 
	0xff, 0xff, 0x7f, 0xff, 0xef, 0xe1, 0xff, 0x7e, 
	0xff, 0xdd, 0x7b, 0xff, 0xbf, 0xff, 0xff, 0xf5, 
	0xbd, 0xfe, 0x2e, 0xec, 0xef, 0xfe, 0xbb, 0x67, 
	0xbf, 0x7f, 0x7b, 0x28, 0xff, 0xf5, 0xff, 0x7d, 
	0xf7, 0xff, 0xfd, 0x7f, 0xec, 0xb1, 0xfb, 0xfe, 
	0xff, 0xff, 0xfe, 0xd7, 0xff, 0xff, 0x3f, 0xfa, 
	0xff, 0xeb, 0xef, 0xbd, 0xbf, 0xfe, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xfd, 0xff, 
	0xff, 0xff, 0xfb, 0xf7, 0xf7, 0xbf, 0xff, 0xff, 
	0xed, 0xff, 0xef, 0x85, 0xff, 0xff, 0x7f, 0xff, 
	0xfe, 0xff, 0xff, 0xff, 0xff, 0x9f, 0x7f, 0xbe, 
	0xff, 0xff, 0xf3, 0x85, 0x7f, 0xea, 0xe1, 0xf1, 
	0xfd, 0x57, 0xff, 0xff, 0xff, 0xff, 0xff, 0x94, 
	0xfd, 0xef, 0xbf, 0xff, 0xfe, 0xfb, 0xff, 0xff, 
	0xf6, 0x7f, 0xfa, 0x7c, 0xff, 0xff, 0xef, 0x57, 
	0xf3, 0xa9, 0x8d, 0xef, 0xf7, 0x5f, 0xce, 0xff, 
	0xff, 0xff, 0xff, 0x6d, 0xfe, 0xd7, 0x5f, 0x7f, 
	0xfd, 0xff, 0xd7, 0xcf, 0xff, 0x7f, 0xff, 0xff, 
	0xcd, 0xda, 0xff, 0x5e, 0xe7, 0x2b, 0xdf, 0xd2, 
	0xdf, 0x7f, 0xfd, 0xef, 0xff, 0xff, 0xff, 0x3f, 
	0x59, 0xdf, 0x7f, 0xff, 0xfd, 0xff, 0x9f, 0x7b, 
	0xff, 0xff, 0xbf, 0xff, 0x3f, 0xeb, 0xff, 0x7f, 
	0xfd, 0xf7, 0xfe, 0xdf, 0x7f, 0xff, 0xf5, 0xff, 
	0xff, 0xff, 0xff, 0xd7, 0xe2, 0xfe, 0xda, 0x62, 
	0xaf, 0x5d, 0xb6, 0xfd, 0x71, 0x97, 0xf6, 0xdf, 
	0x3d, 0xff, 0xff, 0xff, 0xf5, 0x7f, 0xf6, 0x2b, 
	0xd9, 0xb6, 0xd7, 0xfe, 0xfd, 0xef, 0xff, 0xbf, 
	0x97, 0xf9, 0x2f, 0x9b, 0xfc, 0xf6, 0xc9, 0xf6, 
	0xd6, 0x7d, 0xea, 0xbb, 0xf3, 0xff, 0xfb, 0xff, 
	0x57, 0xff, 0xb5, 0xf7, 0xfc, 0xdb, 0x5e, 0xfd, 
	0xff, 0xff, 0xff, 0xff, 0xcc, 0xec, 0xff, 0xff, 
	0x3f, 0xff, 0xfb, 0xff, 0xff, 0x7f, 0xe8, 0xff, 
	0xee, 0xfb, 0xfa, 0xff, 0x5b, 0xff, 0x97, 0xdf, 
	0xd5, 0x77, 0x7f, 0xfd, 0x7f, 0x7e, 0xff, 0xff, 
	0x1b, 0xb9, 0xff, 0xf7, 0xdf, 0x75, 0xff, 0xfd, 
	0x77, 0xdd, 0xaf, 0xbe, 0xff, 0x1f, 0xeb, 0xff, 
	0x47, 0xfd, 0xbf, 0xfe, 0xde, 0xfe, 0xfd, 0x95, 
	0xff, 0xf9, 0xff, 0xff, 0x8f, 0xe0, 0xfd, 0xdf, 
	0x7c, 0xdb, 0xfd, 0x37, 0xdf, 0x7f, 0xff, 0xfd, 
	0xf7, 0x7f, 0xff, 0xbf, 0xfb, 0xff, 0xff, 0xe4, 
	0xff, 0xff, 0xff, 0xbf, 0xff, 0xff, 0xfb, 0xff, 
	0xdf, 0x9b, 0xf7, 0xff, 0xf7, 0xdf, 0xff, 0xff, 
	0xfd, 0xff, 0x9f, 0xe7, 0x9f, 0xff, 0xff, 0xff, 
	0xff, 0xfe, 0xff, 0xff, 0x3f, 0xff, 0xfe, 0xff, 
	0xfc, 0xf7, 0xcf, 0xff, 0x7f, 0xff, 0xe4, 0xdf, 
	0x6c, 0xf3, 0xcc, 0x35, 0xc7, 0x54, 0xd3, 0xcf, 
	0x3f, 0xfe, 0xfe, 0x7b, 0xdd, 0x7d, 0xbb, 0xfd, 
	0xb7, 0x9f, 0x7d, 0xff, 0xe9, 0xf7, 0x1f, 0x7f, 
	0xff, 0x5b, 0x9b, 0x3f, 0xff, 0xfc, 0xf3, 0xff, 
	0x7f, 0xff, 0xfd, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xf7, 0xff, 0xdf, 0xff, 0xfb, 0xef, 0xff, 0xff, 
	0xff, 0xff, 0xfe, 0xff, 0xff, 0xcf, 0xeb, 0xfe, 
	0xfe, 0xfb, 0xef, 0xff, 0xbf, 0xff, 0xfd, 0xff, 
	0xfb, 0xff, 0xff, 0xff, 0xef, 0xff, 0xff, 0xff, 
	0xf7, 0xf7, 0xff, 0xff, 0x7f, 0xff, 0xfd, 0xff, 
	0xff, 0xbf, 0x8b, 0x7a, 0x3f, 0xff, 0xfc, 0xf3, 
	0xcf, 0x2f, 0xfe, 0xdd, 0xf2, 0xdf, 0xaf, 0xbf, 
	0x7e, 0xe7, 0xff, 0xef, 0xbf, 0xff, 0xc7, 0xff, 
	0xff, 0xfa, 0xff, 0xc7, 0xf7, 0xff, 0xfa, 0xde, 
	0xbf, 0xff, 0xfe, 0xfb, 0xef, 0xbf, 0x7f, 0xf7, 
	0x8b, 0x7f, 0x9e, 0xf8, 0xf2, 0xff, 0xff, 0xff, 
	0xf3, 0xcf, 0xdf, 0xff, 0xff, 0xff, 0xdf, 0x1f, 
	0xdf, 0xff, 0xbd, 0xfb, 0x7f, 0xfb, 0xef, 0xf6, 
	0xff, 0xfe, 0xeb, 0xfb, 0xaf, 0xdd, 0xf7, 0xfa, 
	0xe9, 0xfd, 0xff, 0xff, 0xff, 0xf7, 0x7f, 0xf7, 
	0xff, 0x7f, 0xff, 0x7b, 0x7a, 0xff, 0xff, 0xe1, 
	0xff, 0xff, 0xf3, 0x8f, 0x3f, 0xff, 0xfc, 0xf1, 
	0x8f, 0x36, 0x5f, 0xe8, 0xa0, 0xcf, 0x7f, 0xfe, 
	0xf1, 0xe7, 0x87, 0x59, 0xfe, 0xf9, 0xe7, 0x9f, 
	0x6d, 0xfc, 0xf9, 0xbb, 0xff, 0xff, 0xff, 0xff, 
	0xfb, 0xff, 0xaf, 0xdf, 0xf9, 0xf2, 0xeb, 0xa7, 
	0xdf, 0xff, 0xcf, 0xff, 0xff, 0xfb, 0x8b, 0xf4, 
	0xd7, 0xff, 0xff, 0xff, 0xef, 0xf7, 0xff, 0x6f, 
	0xfe, 0xff, 0xff, 0xff, 0xff, 0xff, 0xbf, 0xff, 
	0xff, 0xcb, 0xaf, 0xb7, 0xbe, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xd5, 0x57, 0xff, 0xff, 0xff, 
	0xff, 0xdf, 0xff, 0xff, 0xfb, 0x7f, 0xfb, 0xef, 
	0xff, 0xff, 0xfe, 0x7b, 0xff, 0xaf, 0xbd, 0xf6, 
	0x3a, 0xff, 0xff, 0xff, 0xff, 0xff, 0xbf, 0xdf, 
	0x77, 0xfd, 0xff, 0xff, 0xfe, 0x7c, 0xff, 0xff, 
	0xec, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0x7d, 0xbf, 0xfe, 0x7a, 0xab, 0xfc, 0xff, 0xfb, 
	0xff, 0xff, 0xff, 0xff, 0xdd, 0xf5, 0xff, 0xff, 
	0xff, 0xff, 0xfd, 0xff, 0x87, 0xff, 0xef, 0xbf, 
	0xf9, 0xf2, 0xfb, 0x5f, 0x7f, 0xfd, 0xff, 0xdf, 
	0xff, 0xff, 0xff, 0xde, 0xff, 0xbf, 0xfc, 0xfa, 
	0xfd, 0xb7, 0xff, 0xfb, 0xff, 0xff, 0xff, 0xff, 
	0xee, 0xfa, 0xff, 0xc3, 0xad, 0xb6, 0xee, 0xfa, 
	0xeb, 0xaf, 0xbf, 0xfe, 0xfa, 0xeb, 0xa7, 0x7b, 
	0xfc, 0xf5, 0xd7, 0x5f, 0x77, 0xfd, 0xf5, 0xd7, 
	0x5f, 0x3f, 0xf5, 0xf5, 0xaf, 0xfb, 0xff, 0xfd, 
	0xfb, 0xff, 0xff, 0xff, 0xff, 0xbf, 0x3f, 0x9b, 
	0x74, 0xff, 0x7f, 0x7f, 0xf6, 0xf9, 0x77, 0xbf, 
	0xff, 0x57, 0xf6, 0xcf, 0xff, 0x7d, 0xfe, 0xff, 
	0xfe, 0xcf, 0xff, 0xfd, 0xfd, 0xff, 0x7f, 0xbf, 
	0xff, 0xfe, 0xef, 0xfe, 0xfe, 0xff, 0xfd, 0x6f, 
	0xff, 0xfd, 0x7f, 0xbb, 0xbf, 0xdf, 0xbf, 0xee, 
	0xef, 0x7f, 0xff, 0xff, 0xff, 0x85, 0xfe, 0xff, 
	0xff, 0xff, 0xf7, 0xfe, 0xf7, 0xdf, 0xd7, 0x7e, 
	0x6f, 0xff, 0xff, 0xf7, 0xf7, 0xdb, 0xbe, 0xb7, 
	0xef, 0xff, 0xff, 0xff, 0xff, 0xff, 0xf7, 0xff, 
	0xff, 0xed, 0xfe, 0x3f, 0xfe, 0xfa, 0xe3, 0xbf, 
	0xf7, 0xef, 0xf8, 0xfd, 0xe7, 0xff, 0xff, 0x7c, 
	0xd7, 0xff, 0x3f, 0xdc, 0xf1, 0xd7, 0x7f, 0xff, 
	0xff, 0xff, 0xf7, 0xff, 0xff, 0xff, 0xfb, 0xdf, 
	0xfd, 0xed, 0xd9, 0xaf, 0xbf, 0xff, 0x7b, 0xff, 
	0xff, 0xbe, 0xff, 0xff, 0x5f, 0xff, 0xef, 0xfd, 
	0xf5, 0x7e, 0xff, 0xff, 0xff, 0xff, 0x77, 0xf7, 
	0xff, 0xff, 0xee, 0xef, 0x5f, 0x6f, 0xff, 0xfe, 
	0xbf, 0xef, 0xff, 0xef, 0xff, 0xdf, 0xff, 0xbf, 
	0xef, 0xef, 0xff, 0xff, 0xfd, 0x7a, 0xdc, 0xfd, 
	0xff, 0x3a, 0xff, 0x9f, 0xff, 0xff, 0xa3, 0xf7, 
	0xaf, 0xbf, 0xfe, 0xdf, 0xff, 0xf7, 0xff, 0xff, 
	0xdf, 0xff, 0xff, 0x3d, 0xfe, 0xfe, 0xfb, 0xf7, 
	0xfe, 0xff, 0xe5, 0xbe, 0xf7, 0xfd, 0xfb, 0xf5, 
	0xbe, 0xff, 0x8f, 0xfe, 0xf4, 0xdf, 0x7f, 0xdf, 
	0xff, 0xff, 0xf1, 0xff, 0xff, 0xff, 0xdd, 0x7f, 
	0xff, 0xfe, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xdf, 0xff, 0xfd, 0xff, 0xed, 0x2f, 0xf8, 
	0xff, 0xff, 0xff, 0x7f, 0xff, 0xbf, 0xef, 0xdd, 
	0xdd, 0x3f, 0xff, 0xb8, 0xff, 0xf6, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0x7f, 0xfd, 0xfb, 0xff, 0xff, 
	0xfd, 0xfb, 0xdf, 0xe8, 0xff, 0xff, 0xff, 0xfa, 
	0xff, 0xef, 0xff, 0xff, 0x7f, 0xfb, 0xe6, 0xdf, 
	0xff, 0xf7, 0xf7, 0xfe, 0x7f, 0xff, 0xfd, 0xff, 
	0xff, 0xef, 0xff, 0xff, 0xff, 0xff, 0xff, 0x9f, 
	0xfb, 0xbb, 0xff, 0x77, 0x77, 0xff, 0x37, 0xfa, 
	0xfb, 0x7f, 0xfd, 0xfd, 0xfb, 0xe5, 0x3f, 0x7f, 
	0xfc, 0xe7, 0xef, 0xee, 0xff, 0xff, 0x3f, 0xf5, 
	0xff, 0xff, 0xff, 0x1f, 0xfc, 0xff, 0x7e, 0xff, 
	0xff, 0xf6, 0xf7, 0xff, 0xbf, 0xff, 0xff, 0xdf, 
	0x7f, 0xff, 0xd7, 0xff, 0xff, 0xff, 0x7f, 0xdf, 
	0xfc, 0xff, 0xff, 0xa7, 0x7f, 0xff, 0xff, 0xdf, 
	0xba, 0xff, 0xfc, 0xff, 0xff, 0xfb, 0xff, 0xfe, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xbf, 0xfb, 0x7b, 0xff, 0xff, 0xbf, 0xef, 
	0xff, 0xff, 0xf7, 0xbf, 0xc6, 0xff, 0xfb, 0xff, 
	0xd6, 0xf3, 0xda, 0xef, 0xf9, 0xe7, 0xff, 0x3f, 
	0xff, 0xff, 0x7f, 0x2d, 0xcf, 0xfa, 0xdf, 0x7b, 
	0x37, 0xff, 0x7f, 0xff, 0x7f, 0x9d, 0xcf, 0xff, 
	0xa8, 0xfd, 0xff, 0xff, 0xcb, 0xef, 0xff, 0xef, 
	0xe7, 0xcf, 0xff, 0xfb, 0xf5, 0xff, 0xff, 0xff, 
	0xfc, 0xf3, 0xff, 0xbf, 0xfc, 0xfc, 0xff, 0xfe, 
	0xff, 0xf5, 0xff, 0xff, 0xfd, 0xf6, 0xd7, 0x5f, 
	0x77, 0xde, 0xaf, 0xff, 0xfe, 0xff, 0xdf, 0x7f, 
	0xfd, 0x9d, 0xfa, 0xd7, 0x4b, 0xfd, 0xff, 0xfd, 
	0x7c, 0xfe, 0xfe, 0xff, 0xff, 0xfc, 0xfe, 0xfd, 
	0x3f, 0x7b, 0xdf, 0xff, 0xfb, 0x69, 0xff, 0xcf, 
	0xfe, 0xff, 0x7f, 0xbf, 0xfb, 0x2f, 0xeb, 0xff, 
	0x6f, 0xfd, 0xff, 0xff, 0xfb, 0xff, 0xfb, 0xff, 
	0xfe, 0xf9, 0xff, 0xff, 0xd7, 0xe6, 0xfe, 0xca, 
	0xa6, 0x9c, 0x56, 0x9a, 0xed, 0x75, 0xff, 0xff, 
	0x7f, 0x7d, 0xff, 0x5d, 0x49, 0xfe, 0xff, 0xde, 
	0x13, 0xef, 0xbf, 0xff, 0xf6, 0xb6, 0xff, 0xff, 
	0xbf, 0xb7, 0xf9, 0x66, 0x9b, 0x16, 0xda, 0x7d, 
	0xef, 0xff, 0xff, 0xff, 0xbb, 0xf6, 0xbf, 0x27, 
	0xff, 0x7b, 0xff, 0xff, 0xef, 0xdd, 0xff, 0xff, 
	0x7f, 0xfb, 0xde, 0xff, 0xff, 0x8c, 0xea, 0xff, 
	0xff, 0x7a, 0xeb, 0xab, 0xbd, 0xfe, 0xff, 0xff, 
	0xff, 0xf7, 0xf7, 0xda, 0xd7, 0x57, 0xff, 0xff, 
	0xfd, 0x57, 0xeb, 0xad, 0xa7, 0xfc, 0xd5, 0xff, 
	0xff, 0x9b, 0xb8, 0xff, 0xf7, 0x6f, 0xb8, 0xfd, 
	0xff, 0xff, 0xdf, 0xff, 0xb3, 0xbf, 0xef, 0xef, 
	0xff, 0x7c, 0xfd, 0x9f, 0xfe, 0x5f, 0xfb, 0x9f, 
	0x3f, 0x96, 0x5f, 0xff, 0xff, 0xef, 0xc1, 0xbd, 
	0xdd, 0xfc, 0xf9, 0xf7, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0x5f, 0xff, 0xbb, 0xdf, 0xfe, 0xbf, 
	0xf7, 0xe7, 0xfb, 0xaf, 0xbf, 0xff, 0xe6, 0xff, 
	0xff, 0xdf, 0xb3, 0xf6, 0xfd, 0xf7, 0x5f, 0x7f, 
	0xfe, 0xff, 0xfe, 0xf9, 0xef, 0xbf, 0x7d, 0xfe, 
	0x7f, 0xce, 0xfe, 0xff, 0xfc, 0xff, 0xcf, 0x3f, 
	0xff, 0xfc, 0x3f, 0xff, 0xff, 0x7f, 0x9f, 0xec, 
	0xcf, 0xbc, 0xbb, 0xef, 0xbf, 0xdf, 0x5e, 0xe3, 
	0xef, 0xbf, 0xef, 0x7c, 0xfb, 0x9d, 0x77, 0xfb, 
	0xf9, 0xf7, 0x1a, 0x7f, 0xfc, 0xf1, 0xa6, 0xdf, 
	0x7e, 0xff, 0x5b, 0xba, 0x3f, 0xff, 0x7d, 0xff, 
	0xfd, 0xef, 0xff, 0xff, 0xff, 0xfd, 0xf7, 0xdf, 
	0xfc, 0xf3, 0xff, 0xff, 0xff, 0xff, 0xef, 0xff, 
	0xef, 0x7f, 0xff, 0xff, 0xff, 0xff, 0xcf, 0xea, 
	0xfe, 0xfe, 0xfe, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xfb, 0xef, 0xbf, 0xfb, 0xef, 0xff, 0xff, 
	0x7f, 0xff, 0xdf, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xbf, 0xbf, 0xfb, 0x3f, 0xbf, 0xf7, 
	0xde, 0xff, 0xff, 0xf7, 0xfa, 0xdf, 0x7f, 0x7f, 
	0xbe, 0xff, 0xd7, 0xff, 0xff, 0xf8, 0xbf, 0x87, 
	0x3f, 0xff, 0xf8, 0xfb, 0xff, 0xff, 0xff, 0x9e, 
	0xd8, 0x9f, 0xdf, 0x6b, 0xaf, 0xff, 0xbe, 0xea, 
	0x67, 0xbf, 0x7f, 0xfe, 0xfb, 0xeb, 0x7f, 0xfd, 
	0xff, 0xf7, 0xdf, 0x33, 0xff, 0xfd, 0xff, 0xfd, 
	0x3b, 0xff, 0xff, 0x7d, 0xe8, 0xef, 0xfb, 0xeb, 
	0xbe, 0xfb, 0xff, 0x7f, 0xff, 0xff, 0xbf, 0xfe, 
	0x77, 0xef, 0xff, 0xfd, 0xdf, 0xff, 0xff, 0x7f, 
	0xfe, 0xd7, 0xdf, 0xf7, 0xff, 0xff, 0xff, 0xbf, 
	0xa0, 0xff, 0xff, 0xa3, 0x8f, 0x3f, 0xff, 0xe8, 
	0x73, 0xcf, 0x37, 0xfa, 0xfc, 0xe3, 0xcf, 0x7f, 
	0xfe, 0xb9, 0xe7, 0x1f, 0x7f, 0xfe, 0xf9, 0xc7, 
	0x9f, 0x5f, 0xfe, 0xf9, 0xaa, 0xff, 0xff, 0xff, 
	0xfe, 0xfb, 0xeb, 0xbf, 0xff, 0xff, 0xf7, 0xdf, 
	0xff, 0xff, 0xfe, 0xff, 0xff, 0xff, 0xfa, 0xff, 
	0x87, 0xbf, 0xff, 0xfe, 0xff, 0x75, 0xff, 0xff, 
	0x9f, 0xf8, 0xff, 0xff, 0xee, 0xbb, 0xaf, 0xff, 
	0xfe, 0xff, 0x5f, 0x7f, 0xff, 0xff, 0xfe, 0xfb, 
	0xff, 0xff, 0xff, 0xdf, 0xdf, 0xff, 0xfe, 0xff, 
	0xfd, 0xd7, 0xf7, 0xff, 0x0f, 0xfb, 0xff, 0xfb, 
	0xff, 0xbf, 0xbf, 0xfa, 0xeb, 0xbf, 0xff, 0xff, 
	0xff, 0xdf, 0xaf, 0xff, 0xfd, 0xff, 0xef, 0xd7, 
	0x7f, 0xff, 0xef, 0xef, 0xff, 0x5f, 0xfb, 0xff, 
	0xff, 0xe7, 0xff, 0xff, 0xff, 0xbf, 0xff, 0xfa, 
	0xef, 0xff, 0xff, 0xff, 0xdf, 0xff, 0xef, 0xff, 
	0xff, 0xff, 0xff, 0x5f, 0xff, 0xfd, 0xff, 0xff, 
	0xff, 0x7f, 0xfd, 0xff, 0xbf, 0xa3, 0xff, 0x5f, 
	0xbf, 0xfd, 0xfe, 0x9b, 0x6f, 0xff, 0xff, 0xff, 
	0x9b, 0xff, 0xfd, 0xbd, 0xaf, 0x7e, 0xff, 0x7f, 
	0xfb, 0xfd, 0x7f, 0xff, 0x7f, 0xfb, 0xff, 0x7f, 
	0xfb, 0x8f, 0xfe, 0xff, 0xeb, 0xaf, 0xbb, 0xfe, 
	0xfa, 0xe3, 0xaf, 0xbf, 0xfe, 0xfa, 0xab, 0xa7, 
	0x7f, 0xf5, 0xf5, 0x57, 0x5f, 0x77, 0xed, 0xf5, 
	0xd3, 0x5f, 0x7b, 0xfd, 0xf5, 0x3f, 0xf9, 0xff, 
	0xdf, 0x83, 0x3f, 0x3f, 0xfb, 0xfc, 0xff, 0xff, 
	0x1b, 0x78, 0xe3, 0xfb, 0xdc, 0xfd, 0xff, 0xe7, 
	0x67, 0xff, 0xff, 0xff, 0xe7, 0xe5, 0xff, 0xed, 
	0xff, 0xfe, 0xc9, 0xff, 0xfe, 0xaf, 0xaf, 0xd6, 
	0xb7, 0xff, 0xfb, 0xef, 0xff, 0xfb, 0xfb, 0xf7, 
	0xf7, 0xbf, 0xff, 0xde, 0xff, 0xdf, 0xbf, 0xff, 
	0xeb, 0xff, 0xdf, 0xff, 0xff, 0xff, 0xa9, 0xf6, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xdf, 0xfb, 
	0xff, 0x7b, 0xb7, 0xcc, 0xff, 0xfe, 0xff, 0xdf, 
	0xfb, 0xff, 0xff, 0xff, 0xbf, 0xfe, 0x77, 0xdf, 
	0xff, 0xbf, 0x2f, 0xfc, 0xaf, 0xff, 0x5c, 0xeb, 
	0xbf, 0xff, 0xec, 0xfb, 0xff, 0xbf, 0xff, 0xff, 
	0xd8, 0xc7, 0xff, 0xf7, 0xff, 0xf7, 0xa7, 0xff, 
	0xfe, 0xbb, 0xf1, 0xdd, 0xff, 0xff, 0x7f, 0xfb, 
	0xff, 0xf7, 0xff, 0xbf, 0xee, 0xbf, 0xff, 0xf7, 
	0xff, 0xff, 0xff, 0xff, 0x7f, 0xbf, 0xf7, 0x7f, 
	0xdf, 0xff, 0xff, 0xff, 0xbf, 0xff, 0xef, 0xfe, 
	0xff, 0xcf, 0xff, 0xe6, 0xff, 0xef, 0xff, 0xbf, 
	0xef, 0xfa, 0xdb, 0xff, 0xff, 0xb7, 0x7a, 0xfe, 
	0xef, 0xff, 0xff, 0xff, 0xff, 0xdf, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0x7f, 0xdc, 0xff, 0xff, 0x8f, 
	0xff, 0xff, 0xff, 0xf7, 0xff, 0x6b, 0xff, 0xff, 
	0xfd, 0xfa, 0xff, 0xff, 0xf3, 0x73, 0xdf, 0xff, 
	0xff, 0xff, 0xff, 0xbf, 0xfb, 0xff, 0x7f, 0xfd, 
	0xff, 0xff, 0xfd, 0xcf, 0xfe, 0xf5, 0x5d, 0x5f, 
	0xff, 0x7f, 0xff, 0xbf, 0xff, 0xff, 0x7f, 0xff, 
	0xff, 0xff, 0xff, 0xfb, 0xef, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xf7, 0xff, 0x3b, 
	0xf8, 0xff, 0xff, 0xf3, 0x86, 0x3f, 0xf6, 0xf8, 
	0xbf, 0xff, 0xff, 0xff, 0xbd, 0xff, 0xff, 0x7f, 
	0xfe, 0xf1, 0xc7, 0x5f, 0xff, 0xfe, 0xf9, 0xef, 
	0xff, 0xff, 0xbf, 0xff, 0xe2, 0xfe, 0xff, 0x6f, 
	0xff, 0xf5, 0xfb, 0x76, 0xff, 0xff, 0xff, 0xcf, 
	0xef, 0xff, 0xff, 0xff, 0xef, 0xdf, 0xff, 0xfe, 
	0xff, 0xee, 0xdf, 0x7e, 0xff, 0xff, 0xff, 0xff, 
	0x93, 0xff, 0x7f, 0xff, 0x7c, 0xfb, 0xe8, 0x7f, 
	0xfb, 0x7f, 0xfb, 0xdd, 0xff, 0xff, 0xfd, 0xf5, 
	0xee, 0xfb, 0xff, 0xbf, 0xff, 0xff, 0xbf, 0xbf, 
	0xfd, 0x65, 0xff, 0xff, 0xcf, 0xfc, 0xff, 0xff, 
	0xf3, 0xc7, 0xbb, 0xff, 0xed, 0xbf, 0xef, 0xff, 
	0xff, 0x7b, 0xff, 0xdf, 0xfd, 0xff, 0x7b, 0xff, 
	0xbf, 0xff, 0xff, 0xff, 0xf7, 0xdb, 0xff, 0xff, 
	0x1f, 0xfb, 0xff, 0xff, 0xeb, 0xbf, 0xff, 0xfe, 
	0xff, 0xff, 0xaf, 0xff, 0xff, 0xff, 0xbf, 0xbf, 
	0xed, 0xff, 0xff, 0xff, 0xfb, 0xff, 0xb7, 0x9f, 
	0xd7, 0xbf, 0xff, 0xff, 0xff, 0x83, 0xff, 0xaf, 
	0xff, 0xd4, 0xfb, 0xff, 0x6b, 0xac, 0xff, 0xff, 
	0x1f, 0x8b, 0xff, 0xd5, 0x3f, 0x7f, 0xfe, 0xf9, 
	0xff, 0x3f, 0xff, 0xfa, 0xfe, 0xf9, 0x97, 0xff, 
	0xff, 0x9c, 0xfd, 0xef, 0xfe, 0xcb, 0xf7, 0xff, 
	0xff, 0xb3, 0xfe, 0xdb, 0xff, 0x2c, 0xfd, 0x5f, 
	0xff, 0xff, 0xf3, 0xaf, 0xff, 0xff, 0xff, 0xf3, 
	0xcf, 0xff, 0xfd, 0xfa, 0xff, 0x3d, 0xfe, 0xd7, 
	0xfc, 0xf7, 0xdf, 0x7f, 0xbf, 0x5e, 0xdf, 0xff, 
	0xff, 0xff, 0x9d, 0xfa, 0xbf, 0xff, 0xfe, 0xff, 
	0xff, 0xbf, 0xff, 0xfe, 0x3f, 0xd7, 0x7a, 0xff, 
	0xff, 0xbf, 0x5b, 0xff, 0xfe, 0xff, 0x7f, 0x7f, 
	0xfd, 0x7f, 0xfe, 0xf9, 0xb3, 0xcf, 0x35, 0xeb, 
	0xbf, 0xfe, 0xfb, 0xff, 0xff, 0xff, 0xff, 0xfb, 
	0xff, 0xf6, 0xdf, 0xff, 0xff, 0xd7, 0xeb, 0xfe, 
	0xaf, 0x75, 0xfc, 0xf5, 0xdb, 0x6b, 0x26, 0xfd, 
	0xf7, 0xdf, 0xeb, 0xff, 0xd9, 0xeb, 0xbf, 0xff, 
	0xfe, 0xfb, 0xe7, 0xbf, 0xff, 0xf6, 0xab, 0xaf, 
	0xbe, 0xbf, 0x9f, 0xf9, 0xef, 0xff, 0xff, 0xff, 
	0x6f, 0xff, 0x95, 0xfa, 0xef, 0x7d, 0xa5, 0xbe, 
	0x57, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xd7, 0x5e, 0xfb, 0xff, 0x9c, 0xec, 
	0xbf, 0xde, 0x7f, 0x7d, 0xbc, 0xbf, 0x5e, 0xff, 
	0xff, 0xf1, 0xfd, 0xff, 0xfa, 0xff, 0xff, 0xef, 
	0xf7, 0xdf, 0x7f, 0xfa, 0xff, 0xf7, 0xdd, 0xdf, 
	0xff, 0xe5, 0x5b, 0xba, 0xff, 0x5f, 0x1c, 0x7c, 
	0xf4, 0xfe, 0x5f, 0x9d, 0xff, 0xfe, 0xff, 0xf7, 
	0xef, 0xff, 0xfd, 0x37, 0xfe, 0xff, 0xff, 0xc7, 
	0xf7, 0x3f, 0xf6, 0xde, 0xf9, 0xaf, 0xef, 0xe6, 
	0xfd, 0x7f, 0xff, 0xfd, 0xb7, 0x3f, 0x7f, 0xff, 
	0xff, 0xfd, 0xbf, 0xfd, 0xf6, 0xbf, 0xfb, 0xaf, 
	0xbf, 0xfe, 0xfa, 0xeb, 0xaf, 0xbf, 0xff, 0xe6, 
	0xfb, 0xef, 0xdd, 0x8f, 0xf7, 0xff, 0xf6, 0xfd, 
	0x77, 0xbf, 0xff, 0xff, 0xdf, 0xe7, 0xff, 0x7d, 
	0xfe, 0xdf, 0xce, 0x3e, 0xff, 0xfc, 0xf3, 0xcf, 
	0x3f, 0xff, 0xfc, 0xf3, 0xcf, 0x3e, 0x7f, 0x5f, 
	0xec, 0xff, 0x4e, 0x73, 0xcd, 0xb3, 0xfd, 0xfe, 
	0xfb, 0xce, 0xb5, 0xd3, 0x7a, 0x3b, 0x1d, 0x77, 
	0xfe, 0xf1, 0xc7, 0x1f, 0x7f, 0xfe, 0xf1, 0xc7, 
	0x9f, 0x77, 0xfc, 0x93, 0x9a, 0x7f, 0xdf, 0xfd, 
	0xf7, 0xdb, 0x7f, 0xdf, 0xfd, 0xff, 0xdf, 0x7f, 
	0xbf, 0xfd, 0xf7, 0xff, 0xef, 0xbf, 0xff, 0xfe, 
	0xfb, 0xef, 0xff, 0xfb, 0xef, 0xff, 0xff, 0xcf, 
	0xe3, 0x9e, 0xff, 0xfe, 0xef, 0xbf, 0xff, 0xfe, 
	0xfd, 0xff, 0xbb, 0xbf, 0xff, 0xfd, 0xef, 0xff, 
	0xdf, 0xff, 0xff, 0xff, 0xff, 0xdf, 0xff, 0xf7, 
	0xef, 0xf7, 0xff, 0xbf, 0xa7, 0xfa, 0xef, 0xfd, 
	0xfa, 0xeb, 0x8f, 0xed, 0xdd, 0xfa, 0xe3, 0xfb, 
	0xff, 0xbf, 0xff, 0xf7, 0x3f, 0xff, 0xfe, 0xf3, 
	0xef, 0xbf, 0xff, 0x78, 0xe3, 0x8f, 0xbf, 0xfe, 
	0xba, 0xde, 0xff, 0xca, 0xe7, 0xbf, 0xff, 0xf2, 
	0xfa, 0xe6, 0xcd, 0xff, 0xf6, 0xf9, 0xeb, 0x3d, 
	0xff, 0xfd, 0xf7, 0xdf, 0x7f, 0xff, 0xfd, 0xf7, 
	0xc6, 0x3f, 0xff, 0xfd, 0x7d, 0xfa, 0xff, 0xbb, 
	0xdf, 0xfe, 0xfa, 0xf5, 0xbe, 0x7f, 0xbe, 0xb8, 
	0xf7, 0xbf, 0xeb, 0xff, 0xf7, 0xfb, 0x7f, 0xff, 
	0xfd, 0xff, 0xff, 0xef, 0xb7, 0x7f, 0xfe, 0xff, 
	0xff, 0xe9, 0xff, 0xfb, 0xf3, 0xcf, 0x3f, 0xdf, 
	0xec, 0x73, 0xcf, 0x16, 0xde, 0x7c, 0xa3, 0xcf, 
	0x7f, 0xfe, 0xf9, 0xe7, 0x9e, 0x7f, 0xfe, 0xf9, 
	0xe7, 0x1f, 0x7d, 0xfe, 0xf9, 0x9f, 0xff, 0xaf, 
	0xff, 0xf9, 0xff, 0xff, 0xef, 0xff, 0xfe, 0xf2, 
	0xfd, 0xbf, 0xff, 0xfe, 0xff, 0x3f, 0xff, 0xff, 
	0xfb, 0xef, 0xbf, 0xff, 0xfb, 0xeb, 0xaf, 0xff, 
	0xff, 0x3f, 0xfe, 0xbf, 0xfa, 0xef, 0xbf, 0xff, 
	0xfe, 0xfa, 0xee, 0xeb, 0xed, 0xbf, 0xff, 0xfe, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xbf, 0xff, 0xfe, 
	0xff, 0xf7, 0xff, 0xff, 0xff, 0xbf, 0xf8, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xfe, 0xff, 0x2f, 
	0xff, 0xfb, 0xef, 0xbf, 0xff, 0xfd, 0xef, 0xbf, 
	0xbf, 0xfe, 0xfb, 0xef, 0xff, 0xbf, 0xff, 0xfe, 
	0xff, 0xff, 0xeb, 0xff, 0xff, 0x7f, 0xff, 0xff, 
	0xff, 0xef, 0xff, 0xbf, 0xde, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0x7f, 0xff, 0xdf, 0xff, 0x97, 0xff, 
	0xef, 0xff, 0xff, 0xef, 0xd7, 0x6f, 0xff, 0xfb, 
	0xf6, 0xdb, 0xff, 0xff, 0xf5, 0xff, 0xff, 0xff, 
	0xfd, 0xf7, 0xff, 0xfe, 0xfe, 0xff, 0xf1, 0xed, 
	0xff, 0xff, 0xff, 0xfa, 0xbf, 0xeb, 0x2f, 0xbe, 
	0xf6, 0xea, 0xeb, 0x87, 0xbb, 0xee, 0xfa, 0xab, 
	0xaf, 0x7f, 0xf5, 0xf5, 0xd7, 0x5f, 0x7b, 0xf9, 
	0xf5, 0xd7, 0x5f, 0x77, 0xfd, 0xf5, 0xef, 0xfb, 
	0x7f, 0xfb, 0xdf, 0xff, 0xff, 0xff, 0xcc, 0xff, 
	0x2f, 0xbb, 0xff, 0x23, 0xbb, 0xff, 0xf7, 0xff, 
	0xf7, 0x7f, 0x7f, 0xfe, 0x79, 0xfe, 0x9f, 0x5f, 
	0xfd, 0xb9, 0xff, 0xcc, 0xff, 0xff, 0xe7, 0xdc, 
	0xff, 0xff, 0xff, 0xff, 0x5f, 0xdf, 0xff, 0xfb, 
	0xf7, 0xdf, 0xff, 0xff, 0xff, 0xef, 0x77, 0xdd, 
	0xfb, 0xfe, 0x7f, 0xff, 0xff, 0x7b, 0x7f, 0xa1, 
	0xfe, 0x5f, 0xef, 0x7f, 0xff, 0xbd, 0xff, 0xff, 
	0xff, 0xfb, 0x7f, 0xdf, 0xfd, 0xdd, 0xff, 0xff, 
	0x7f, 0xff, 0xf7, 0xbf, 0xef, 0xff, 0x7f, 0xff, 
	0xf7, 0xdf, 0xff, 0x8f, 0xfe, 0xbf, 0xbe, 0x7a, 
	0xef, 0x8f, 0xff, 0xff, 0xef, 0xeb, 0xcf, 0xb7, 
	0xff, 0xfb, 0xc7, 0xfd, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xee, 0xbf, 0xf9, 0xdf, 0xff, 0xff, 0x3f, 
	0xf9, 0xf7, 0xfb, 0xef, 0xef, 0x7f, 0xff, 0xff, 
	0xff, 0x7f, 0xff, 0xfb, 0xfb, 0xfb, 0xbf, 0xff, 
	0xdf, 0xff, 0xfb, 0xff, 0xff, 0xbf, 0xff, 0xff, 
	0xfd, 0xff, 0xff, 0xff, 0xe4, 0xff, 0xeb, 0xec, 
	0xfb, 0xff, 0xff, 0xe7, 0xbf, 0xbb, 0xef, 0xbf, 
	0xff, 0xff, 0xbf, 0xbf, 0xff, 0xbe, 0x3b, 0xff, 
	0xf9, 0x7f, 0xdf, 0xff, 0x7f, 0xff, 0xfe, 0xff, 
	0x87, 0xff, 0x7f, 0xdf, 0x7d, 0xf7, 0xff, 0xaf, 
	0x3d, 0xee, 0xf7, 0xff, 0xff, 0xff, 0xf3, 0xbf, 
	0xff, 0x5e, 0xff, 0xe7, 0xff, 0xff, 0x0f, 0xee, 
	0xff, 0xff, 0xfe, 0xff, 0x6f, 0xf6, 0xf6, 0xfd, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xbf, 0xf1, 
	0xff, 0xfb, 0xdf, 0x7f, 0xfe, 0xdf, 0xff, 0x7f, 
	0xbf, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xfb, 0xfb, 0xff, 0xf8, 0xfd, 0xff, 0x7f, 0xff, 
	0xfe, 0xfd, 0xf7, 0xfb, 0xff, 0x6d, 0xff, 0x96, 
	0xdf, 0x7e, 0x7f, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xe0, 0xff, 0xf7, 
	0xff, 0xfe, 0xff, 0xf7, 0xfb, 0xff, 0xf8, 0xf7, 
	0xff, 0xf7, 0xff, 0x7f, 0xfd, 0xee, 0xff, 0xff, 
	0xff, 0xdf, 0xff, 0xff, 0xfe, 0xff, 0xf7, 0xff, 
	0xbf, 0xab, 0xff, 0x7f, 0xfb, 0x76, 0xfb, 0xfd, 
	0x3f, 0xdf, 0xff, 0xff, 0xff, 0xfb, 0xff, 0x5d, 
	0xf7, 0xff, 0xff, 0xff, 0xdf, 0xe7, 0xfb, 0x7f, 
	0xff, 0xef, 0x7f, 0xff, 0xff, 0x0e, 0xfc, 0xfb, 
	0xed, 0xbb, 0x7f, 0xaf, 0xff, 0xfe, 0xff, 0xff, 
	0x7f, 0xfd, 0xff, 0xdf, 0x9f, 0xfa, 0xeb, 0xbf, 
	0xff, 0xc7, 0xff, 0x7e, 0xfc, 0xef, 0xff, 0xff, 
	0xfe, 0x1b, 0xfb, 0xff, 0xff, 0xf3, 0xef, 0xff, 
	0xdb, 0xfc, 0xff, 0xff, 0xff, 0xff, 0xff, 0x3f, 
	0x9f, 0xff, 0xb7, 0xdf, 0xff, 0xff, 0xfd, 0xff, 
	0xdf, 0xff, 0xfb, 0xff, 0xff, 0xff, 0x8a, 0xff, 
	0xbf, 0xf9, 0xe6, 0xc7, 0xc6, 0xbf, 0x6f, 0xd7, 
	0x5f, 0x2a, 0xab, 0xef, 0x7e, 0x9f, 0xbf, 0xfc, 
	0xff, 0xfc, 0x37, 0x97, 0x5f, 0xfe, 0xff, 0xff, 
	0xff, 0xff, 0xa0, 0xfd, 0xff, 0xd6, 0xfb, 0x3f, 
	0xdb, 0x7f, 0xbf, 0xce, 0x3f, 0xeb, 0x7c, 0xff, 
	0xfb, 0xfd, 0xfc, 0xff, 0xff, 0xf7, 0xdf, 0xfa, 
	0xdf, 0xfd, 0xff, 0xff, 0xff, 0xff, 0x7d, 0xfe, 
	0xa7, 0xfe, 0xf7, 0xdf, 0x75, 0xff, 0xfd, 0xfa, 
	0xff, 0x2f, 0xbc, 0xcc, 0x7a, 0xf6, 0xff, 0x2e, 
	0x39, 0xf7, 0xdc, 0x59, 0xfd, 0xfd, 0xff, 0xff, 
	0xff, 0xff, 0xbf, 0x5a, 0xdf, 0xfa, 0xff, 0x7f, 
	0x7f, 0xfd, 0x7f, 0xea, 0xff, 0xbf, 0xfa, 0xf7, 
	0xfb, 0xdc, 0xff, 0xb7, 0xe5, 0xfe, 0xdb, 0x7f, 
	0x3f, 0xfd, 0xff, 0xff, 0xff, 0xff, 0xd7, 0xe8, 
	0xfe, 0x6a, 0x7f, 0xfc, 0xf5, 0xfe, 0x3b, 0xe5, 
	0xd7, 0xf7, 0xfa, 0x6b, 0xe6, 0x5d, 0xab, 0xbe, 
	0xd5, 0x56, 0x5f, 0xdf, 0x76, 0xfb, 0xed, 0xff, 
	0xff, 0xff, 0xbf, 0xb7, 0xf9, 0xaf, 0xfd, 0xff, 
	0xff, 0xfb, 0xbf, 0x99, 0x5b, 0xff, 0xab, 0xff, 
	0xd5, 0x17, 0x7d, 0xfb, 0x5f, 0x7b, 0xf5, 0xf6, 
	0xfa, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x2c, 
	0xec, 0xff, 0xff, 0x5f, 0xff, 0xfd, 0xbf, 0xff, 
	0x7a, 0xbf, 0xaf, 0xbf, 0xfe, 0xfa, 0xd7, 0xff, 
	0xff, 0xbf, 0xf5, 0xd5, 0x5f, 0x7f, 0xfd, 0xff, 
	0xff, 0xff, 0xff, 0xdb, 0xbb, 0xff, 0xdf, 0x1f, 
	0x78, 0xc4, 0xff, 0x7b, 0xeb, 0xf9, 0xbf, 0xde, 
	0xdf, 0x7b, 0xff, 0xfc, 0xf7, 0xbf, 0xd0, 0x5f, 
	0x7b, 0xef, 0xb5, 0xff, 0xff, 0xff, 0xff, 0xef, 
	0xe0, 0xfd, 0x7f, 0xff, 0xfd, 0xc7, 0xbf, 0xfd, 
	0xfb, 0xef, 0xff, 0xff, 0x7e, 0xff, 0xbb, 0x9f, 
	0xaf, 0xff, 0xfe, 0xb7, 0xff, 0x7f, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xdf, 0xb3, 0xf5, 0xff, 0xff, 
	0xd9, 0x67, 0xbe, 0xfd, 0xf7, 0xff, 0xff, 0xbf, 
	0xff, 0xff, 0x5f, 0xce, 0x3c, 0xff, 0xff, 0x3f, 
	0xdf, 0x7f, 0xef, 0xfd, 0xff, 0xff, 0xff, 0x7f, 
	0x0f, 0xe8, 0xff, 0xfe, 0xa3, 0x8d, 0xb5, 0xcf, 
	0xbe, 0xfb, 0xef, 0xbf, 0xff, 0x5e, 0x3b, 0x9d, 
	0x7f, 0xbe, 0xfd, 0xf7, 0xdf, 0x6f, 0xaf, 0xfd, 
	0xf7, 0xdf, 0x7f, 0xff, 0x5b, 0x99, 0x7f, 0xdf, 
	0x7d, 0xfe, 0xfd, 0x3f, 0xdf, 0xff, 0xff, 0xff, 
	0x77, 0xdf, 0xff, 0xff, 0xff, 0x6f, 0xbe, 0xff, 
	0xff, 0xb7, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xcf, 0xea, 0xfe, 0xfe, 0xfe, 0xff, 0xff, 0xff, 
	0xfe, 0xff, 0xff, 0xff, 0xef, 0xfe, 0xff, 0xff, 
	0xff, 0xdf, 0xfd, 0xff, 0xff, 0xbf, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xbf, 0xb3, 0xfa, 0xed, 
	0xdd, 0xf2, 0x5b, 0xcf, 0xfd, 0xfd, 0xff, 0xff, 
	0x7f, 0xed, 0xfd, 0xd6, 0xd7, 0x3f, 0xdf, 0x7b, 
	0xfd, 0xb5, 0xdf, 0x5e, 0x7b, 0xff, 0xff, 0xff, 
	0xff, 0xfa, 0xde, 0xff, 0xda, 0xe7, 0xbf, 0xfd, 
	0xba, 0xfa, 0xf7, 0xbf, 0xff, 0xf6, 0xfa, 0xee, 
	0x7f, 0xf9, 0xfd, 0xf5, 0xe5, 0x56, 0xdd, 0x75, 
	0x97, 0xd7, 0xff, 0xff, 0xff, 0x7d, 0xf9, 0xff, 
	0xbe, 0xff, 0xff, 0xff, 0xdf, 0xdb, 0xfd, 0xff, 
	0xff, 0xff, 0xbe, 0xeb, 0xff, 0xfe, 0xbf, 0x3d, 
	0xff, 0xdf, 0xff, 0xbd, 0xf5, 0xf7, 0xff, 0xff, 
	0xff, 0xff, 0xea, 0xff, 0xef, 0x73, 0xcf, 0x3f, 
	0xdf, 0xf8, 0xf3, 0xc7, 0x3f, 0xff, 0xec, 0xa3, 
	0x8f, 0x7f, 0xee, 0xd9, 0xe6, 0x1f, 0x7f, 0xfc, 
	0xf1, 0x67, 0x9f, 0x7f, 0xfe, 0xf9, 0xb7, 0xff, 
	0xef, 0xff, 0xff, 0xff, 0xff, 0xff, 0xbf, 0x5e, 
	0xff, 0x9f, 0xef, 0xff, 0xfe, 0xff, 0xbf, 0xdf, 
	0xff, 0xff, 0xf7, 0xd7, 0x7f, 0x7f, 0xff, 0xff, 
	0xff, 0xff, 0xaf, 0xfe, 0xff, 0xfa, 0xef, 0xbf, 
	0xff, 0xbe, 0xfa, 0xfa, 0xff, 0xff, 0xff, 0xfa, 
	0xfe, 0xff, 0xff, 0xfe, 0xfd, 0xff, 0xdf, 0x5f, 
	0xff, 0xff, 0xb7, 0xff, 0xff, 0xff, 0xff, 0xf8, 
	0xff, 0xfe, 0xff, 0xff, 0xff, 0xff, 0xfb, 0xeb, 
	0xff, 0xff, 0xff, 0xfe, 0xff, 0xff, 0xfd, 0xef, 
	0xdd, 0xff, 0xff, 0xff, 0xff, 0xf7, 0xf7, 0xff, 
	0xff, 0xff, 0xff, 0xe5, 0xff, 0xef, 0xff, 0xff, 
	0xff, 0xff, 0xfb, 0xaf, 0xff, 0xfb, 0xdf, 0xef, 
	0xff, 0xbd, 0xff, 0xff, 0xdf, 0xff, 0xff, 0xef, 
	0xbf, 0xff, 0x7e, 0xff, 0xff, 0xff, 0xff, 0xbb, 
	0xff, 0xff, 0xff, 0xbf, 0xff, 0xfb, 0xff, 0xbe, 
	0xff, 0xff, 0xff, 0x6f, 0xbe, 0xfd, 0xed, 0xff, 
	0xff, 0x7f, 0xff, 0xdf, 0xf7, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xdf, 0xf2, 0xff, 0xeb, 0xaf, 
	0xbd, 0xfe, 0xfa, 0x6b, 0xae, 0x3f, 0xfe, 0xfa, 
	0xeb, 0xaf, 0x77, 0xfd, 0xb5, 0xd7, 0x5d, 0x7f, 
	0xbd, 0xf5, 0xd7, 0x4f, 0x7f, 0xfd, 0xf5, 0x6f, 
	0xfa, 0xff, 0xf3, 0xdf, 0x3d, 0x7a, 0xfd, 0xff, 
	0xff, 0xff, 0xdf, 0xdc, 0xec, 0xb3, 0xdf, 0xbf, 
	0xf9, 0xff, 0x9b, 0xff, 0xfe, 0xff, 0xef, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xc6, 0xdf, 0xeb, 0x6f, 
	0xbe, 0xff, 0xfe, 0xfd, 0xff, 0xe7, 0xff, 0x7d, 
	0xff, 0xf7, 0xff, 0xff, 0xf5, 0xff, 0xf7, 0xdd, 
	0xfe, 0xff, 0xeb, 0xff, 0xdf, 0xff, 0xff, 0xff, 
	0xa5, 0xff, 0xff, 0xff, 0xef, 0xfd, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xef, 0xf7, 0xdf, 0x7f, 0xff, 
	0xfe, 0xff, 0x7f, 0xfd, 0x7f, 0xff, 0xef, 0xef, 
	0xff, 0xff, 0xff, 0xff, 0xaf, 0xfe, 0xff, 0xff, 
	0xfa, 0xeb, 0xcf, 0xfb, 0xff, 0x7e, 0xff, 0xff, 
	0xff, 0xff, 0x78, 0xd7, 0xf7, 0xff, 0xff, 0xe5, 
	0xd7, 0xcf, 0x6f, 0xfd, 0xff, 0xff, 0xff, 0xff, 
	0x7f, 0xfb, 0xff, 0xff, 0xdf, 0xbf, 0xff, 0xff, 
	0xff, 0xef, 0xff, 0xdf, 0xff, 0xff, 0xfb, 0xcf, 
	0xfc, 0xff, 0x7f, 0xdf, 0x37, 0x9f, 0xfd, 0xf7, 
	0xff, 0xff, 0xf7, 0xff, 0xff, 0xe3, 0xff, 0xff, 
	0xef, 0x3f, 0xfe, 0xbf, 0xff, 0xff, 0xfb, 0xdf, 
	0xf7, 0xc7, 0xff, 0xbf, 0xef, 0xf7, 0xff, 0x5f, 
	0xf7, 0xfd, 0xf7, 0xf7, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0x93, 0xff, 0xf7, 0xfb, 0xdf, 0xd6, 0xff, 
	0xff, 0xfd, 0xff, 0xe7, 0xff, 0xaf, 0xff, 0xfd, 
	0xfc, 0xbf, 0xff, 0xff, 0xff, 0xbf, 0xff, 0xff, 
	0xbb, 0xff, 0xff, 0xff, 0xff, 0x5f, 0xfe, 0xfe, 
	0xff, 0xef, 0xef, 0xf7, 0xde, 0xff, 0xff, 0xff, 
	0xf9, 0xff, 0xff, 0xbf, 0xdf, 0xf2, 0xfa, 0xff, 
	0xef, 0xbf, 0xff, 0xfe, 0xff, 0xdf, 0xff, 0xfb, 
	0xff, 0x7b, 0xf9, 0xef, 0xff, 0x1b, 0xff, 0xdf, 
	0xff, 0xff, 0xbd, 0xfe, 0xff, 0xfd, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0x7f, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xe6, 0xff, 
	0xff, 0xef, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xfd, 0xff, 0xff, 0xff, 
	0xff, 0xf7, 0x7f, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0x87, 0xff, 0xff, 0xff, 0x6f, 0xb3, 
	0xff, 0xff, 0xdf, 0xff, 0xff, 0xff, 0xbf, 0xff, 
	0xff, 0xff, 0x93, 0xff, 0x7b, 0x79, 0xe7, 0xbf, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x5f, 0xfc, 
	0xff, 0xff, 0xf7, 0xe7, 0xff, 0xdf, 0xff, 0xdf, 
	0x7f, 0xff, 0xff, 0xfe, 0xff, 0xbe, 0x7f, 0xff, 
	0xff, 0xb7, 0xcf, 0xff, 0xbf, 0xf7, 0xff, 0xdf, 
	0xff, 0xff, 0x1f, 0xf9, 0xff, 0xff, 0xff, 0xbe, 
	0xff, 0xfb, 0xff, 0xff, 0xff, 0xff, 0xfb, 0xfc, 
	0xff, 0xff, 0xff, 0xfe, 0xdf, 0xdf, 0x7f, 0xfb, 
	0xf7, 0x7f, 0xff, 0x7f, 0xff, 0xff, 0xff, 0x8c, 
	0xff, 0xff, 0xff, 0xe6, 0x96, 0x0b, 0x3b, 0xff, 
	0xbf, 0xfa, 0x7f, 0xbe, 0xff, 0xff, 0x97, 0xdf, 
	0xff, 0xdf, 0xfb, 0xff, 0xfd, 0xf7, 0xff, 0xff, 
	0xfd, 0xff, 0xff, 0xbc, 0xfd, 0xff, 0xff, 0x9d, 
	0x3e, 0xfb, 0xe5, 0xff, 0xff, 0xef, 0xff, 0xff, 
	0xff, 0xcf, 0xff, 0xff, 0xdf, 0x7f, 0xef, 0xfd, 
	0xf7, 0xdf, 0xff, 0xff, 0xfd, 0xff, 0xff, 0xed, 
	0xfe, 0xff, 0x9f, 0xf2, 0xcf, 0x65, 0xfe, 0xf9, 
	0xff, 0xcb, 0xff, 0xff, 0xf9, 0x7f, 0x1e, 0x7f, 
	0x2d, 0xaf, 0x74, 0xfe, 0xdf, 0x7f, 0xbf, 0xfe, 
	0xf4, 0xff, 0xff, 0x3f, 0x5a, 0xff, 0x7f, 0xea, 
	0x6c, 0xf7, 0xce, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xef, 0xff, 0x39, 0x67, 0xbf, 0xf5, 0xf7, 0xf9, 
	0x7f, 0xff, 0xfd, 0xf4, 0x79, 0xff, 0xff, 0xd7, 
	0xe4, 0xfe, 0xff, 0xaf, 0xfc, 0x26, 0x8b, 0x39, 
	0xff, 0xbf, 0x62, 0xff, 0x7b, 0xf5, 0x7f, 0xe3, 
	0xf5, 0xf8, 0xdf, 0x3b, 0xad, 0xbe, 0xf7, 0x5f, 
	0xaa, 0xc9, 0xfe, 0xbf, 0xb3, 0xf9, 0xff, 0xbf, 
	0xf6, 0xde, 0x7b, 0xef, 0xfd, 0xff, 0x5f, 0xfd, 
	0xbf, 0xd9, 0xff, 0xbd, 0xd7, 0xcd, 0x35, 0x57, 
	0xb6, 0xfb, 0xde, 0x7d, 0xed, 0xbc, 0xfa, 0xff, 
	0xbc, 0xec, 0xff, 0xff, 0x7f, 0xeb, 0xef, 0xd7, 
	0x5f, 0xff, 0xeb, 0xfd, 0xbf, 0xff, 0xff, 0xd7, 
	0x7e, 0x7f, 0xff, 0xfd, 0xff, 0xf7, 0x7f, 0xbf, 
	0xf4, 0xf5, 0xff, 0xff, 0x9b, 0xb9, 0xff, 0xff, 
	0x7f, 0xf9, 0xd5, 0x56, 0x7f, 0xfd, 0xff, 0xed, 
	0xff, 0xfb, 0xff, 0x5f, 0x63, 0xef, 0xf5, 0x37, 
	0xff, 0xff, 0xff, 0x7d, 0x9e, 0x7f, 0xfe, 0xff, 
	0xef, 0xe1, 0xfd, 0xff, 0xff, 0xed, 0xf7, 0xdf, 
	0xff, 0xff, 0xff, 0xf7, 0xff, 0xff, 0xff, 0xbf, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0x9f, 0xff, 0xff, 
	0xff, 0xff, 0xdf, 0xfe, 0xdf, 0xb7, 0xf5, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xf7, 0xff, 0xce, 0x7f, 0xff, 0xff, 
	0xfb, 0xff, 0xff, 0xfb, 0xff, 0xf3, 0xcf, 0xfe, 
	0x7f, 0x1f, 0xec, 0xff, 0xfe, 0xfb, 0xed, 0xb7, 
	0xd7, 0x7e, 0xfb, 0xef, 0xb5, 0xff, 0x1e, 0x7b, 
	0x1d, 0x7f, 0xbf, 0xfd, 0xf6, 0xda, 0x7f, 0xf7, 
	0xfd, 0xc7, 0x1a, 0x7d, 0xff, 0x9b, 0x98, 0xff, 
	0xff, 0x7d, 0xfe, 0xcd, 0x77, 0xdf, 0xfc, 0xff, 
	0xdd, 0xff, 0xdf, 0xfd, 0xf7, 0xbf, 0x7f, 0x7e, 
	0xfb, 0xef, 0xff, 0xff, 0xff, 0xfb, 0xef, 0xff, 
	0xff, 0xcf, 0xec, 0xfe, 0xff, 0xfb, 0xff, 0xbf, 
	0xff, 0xfe, 0xfe, 0xff, 0xbf, 0xff, 0xff, 0xfe, 
	0xfb, 0xbf, 0xff, 0xfd, 0xfb, 0xdf, 0xff, 0xff, 
	0xff, 0xfd, 0xf7, 0xff, 0xff, 0xbf, 0x9b, 0xfa, 
	0xff, 0xb7, 0xf7, 0x5e, 0x7b, 0xef, 0x9d, 0xff, 
	0xdf, 0xfb, 0xef, 0xfd, 0x7f, 0xa7, 0xdf, 0xde, 
	0x7b, 0xb7, 0xfd, 0xff, 0x4b, 0xbf, 0xfd, 0xf7, 
	0xff, 0xff, 0x8a, 0xde, 0xff, 0xdf, 0xeb, 0xab, 
	0xbf, 0xfe, 0xfa, 0xf3, 0xbf, 0x37, 0xbf, 0x7b, 
	0x77, 0x37, 0x7f, 0xf5, 0xf5, 0xcf, 0x7f, 0xff, 
	0x7d, 0xf9, 0xcd, 0x7f, 0xff, 0xff, 0xbd, 0xf9, 
	0xff, 0xff, 0xfb, 0xed, 0xb7, 0xff, 0xdb, 0xfb, 
	0xff, 0xbb, 0xdf, 0x9a, 0xff, 0x77, 0xff, 0xbd, 
	0xfd, 0x77, 0xda, 0xfb, 0xdf, 0xf5, 0xdf, 0xff, 
	0xf7, 0xff, 0xff, 0xe3, 0xff, 0xff, 0xb3, 0x8f, 
	0x3e, 0xbb, 0xf8, 0xb3, 0xcf, 0x3f, 0xfb, 0xe8, 
	0xf3, 0x0f, 0x7f, 0xf6, 0xd9, 0xe7, 0x1f, 0x6f, 
	0xfe, 0xd1, 0xc3, 0x9f, 0x7f, 0xfe, 0xf9, 0x87, 
	0xff, 0xff, 0xbf, 0xff, 0xfb, 0xfb, 0xaf, 0xbf, 
	0xff, 0xfa, 0xfb, 0xbf, 0xbf, 0xff, 0xff, 0xdf, 
	0xdf, 0xff, 0xff, 0xff, 0xff, 0xff, 0xbc, 0xff, 
	0xff, 0xff, 0xff, 0xcf, 0xfe, 0xff, 0xff, 0xeb, 
	0xbf, 0xbf, 0xae, 0xfa, 0xfb, 0xab, 0xbf, 0xff, 
	0xfa, 0xfb, 0xff, 0xdf, 0xff, 0xed, 0xff, 0xf7, 
	0xff, 0xfb, 0xff, 0xbf, 0xff, 0xfe, 0xff, 0x7f, 
	0xf9, 0xff, 0xff, 0xfb, 0xbf, 0xbf, 0xef, 0xff, 
	0xfb, 0xaf, 0xbf, 0xff, 0xff, 0xfb, 0xff, 0xff, 
	0xfd, 0xfd, 0x7f, 0xff, 0x7d, 0xfd, 0xff, 0xfe, 
	0xff, 0xff, 0xff, 0xff, 0xe6, 0xff, 0xff, 0xbf, 
	0xff, 0xff, 0xbb, 0xff, 0xbf, 0xbf, 0xfe, 0xfb, 
	0xff, 0xbf, 0xff, 0xef, 0xbd, 0xdf, 0xff, 0xff, 
	0xff, 0xf5, 0xbf, 0xfb, 0xff, 0xff, 0xff, 0xff, 
	0x93, 0xff, 0xff, 0xff, 0xff, 0x76, 0xff, 0x4f, 
	0xfe, 0xff, 0xf6, 0xbf, 0xfb, 0xff, 0xff, 0xe8, 
	0xff, 0x5e, 0x5a, 0x6f, 0xe7, 0xb7, 0x5f, 0x7e, 
	0xfb, 0xff, 0xff, 0xff, 0x1f, 0xfa, 0xff, 0xeb, 
	0xaf, 0x3f, 0xfe, 0xba, 0xeb, 0xaf, 0xbf, 0xfa, 
	0xfa, 0xeb, 0xaf, 0x72, 0xf4, 0xf5, 0xd6, 0x5b, 
	0x7d, 0xfd, 0x75, 0xd7, 0x5f, 0x7f, 0xfd, 0xf5, 
	0x2f, 0xfa, 0xff, 0x7f, 0xff, 0xcf, 0xfe, 0xf7, 
	0x6f, 0xfd, 0x07, 0xf4, 0xff, 0xe3, 0x7d, 0x7f, 
	0xfe, 0xff, 0xfe, 0xff, 0xff, 0x7f, 0xf6, 0xff, 
	0x07, 0x7f, 0xfc, 0xff, 0xdf, 0xca, 0xff, 0xfe, 
	0xff, 0xdf, 0xff, 0xff, 0xfe, 0xff, 0x73, 0x7e, 
	0xbf, 0xfa, 0xff, 0xdf, 0xd4, 0x7f, 0xff, 0xf3, 
	0xff, 0xff, 0xff, 0xfe, 0xd7, 0x6f, 0x7f, 0xff, 
	0xff, 0xa5, 0xff, 0xff, 0xff, 0xff, 0xff, 0xbd, 
	0xfd, 0xff, 0xdf, 0xff, 0xff, 0xbf, 0xdf, 0x3b, 
	0xff, 0xff, 0xff, 0xfe, 0xff, 0xfe, 0xfb, 0xff, 
	0xff, 0xee, 0xda, 0xff, 0xff, 0x4f, 0xfe, 0xff, 
	0xff, 0xff, 0xef, 0xff, 0x3f, 0xee, 0xff, 0xef, 
	0xff, 0xfd, 0xfe, 0xf8, 0xb7, 0x6f, 0xfe, 0xfb, 
	0xed, 0xe7, 0x7f, 0x7f, 0xbd, 0xff, 0xf7, 0xff, 
	0xff, 0xb7, 0xfb, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xfd, 0xff, 0xfe, 0xee, 0x7f, 0xff, 0xff, 
	0xef, 0x7e, 0xff, 0xff, 0xbb, 0xff, 0xd7, 0xff, 
	0xf3, 0xff, 0x7f, 0xdf, 0xff, 0xff, 0xef, 0xff, 
	0xff, 0xbf, 0xbf, 0xff, 0xff, 0xff, 0x9e, 0x37, 
	0xfc, 0xf1, 0xfb, 0xef, 0xfb, 0xfd, 0xbf, 0xdf, 
	0x3f, 0xf7, 0xff, 0xdd, 0xff, 0x3f, 0xff, 0xff, 
	0xff, 0xff, 0xba, 0xff, 0xff, 0xff, 0xff, 0xef, 
	0x7f, 0xf7, 0xbf, 0xce, 0xfa, 0xff, 0x7f, 0xff, 
	0xfd, 0xe9, 0xff, 0x9f, 0xfe, 0xff, 0xff, 0xfd, 
	0xff, 0x7f, 0xfd, 0xfd, 0xfb, 0xff, 0x4f, 0xfe, 
	0xff, 0xfb, 0xf7, 0xff, 0xff, 0xd7, 0x7f, 0xff, 
	0x3f, 0xfd, 0xef, 0xff, 0xff, 0xdf, 0xf3, 0xff, 
	0xfd, 0xff, 0xff, 0xff, 0xb7, 0xff, 0xff, 0xf7, 
	0xff, 0xff, 0xff, 0xf9, 0xef, 0xdf, 0xff, 0x8f, 
	0xff, 0xff, 0xff, 0xff, 0xe6, 0x3f, 0xff, 0xff, 
	0xe3, 0xfe, 0xfc, 0xfe, 0x3f, 0x7f, 0xef, 0xbe, 
	0xff, 0xff, 0xff, 0x9f, 0xfb, 0xff, 0xff, 0xe0, 
	0xff, 0xff, 0xff, 0x7f, 0xff, 0xff, 0xff, 0xff, 
	0xbf, 0xff, 0xa6, 0xff, 0xb3, 0xdf, 0xf7, 0xbb, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x7f, 
	0xfe, 0xff, 0xbf, 0x9b, 0xfa, 0xff, 0xfe, 0x7c, 
	0xff, 0xfc, 0x3d, 0xb7, 0xfe, 0xfb, 0xcb, 0xff, 
	0xfb, 0xef, 0x65, 0xfd, 0x5f, 0x7f, 0xf9, 0xf5, 
	0xff, 0xff, 0xff, 0xfd, 0xf6, 0xff, 0xff, 0x4f, 
	0xfc, 0xff, 0x7f, 0xfb, 0xff, 0xb6, 0x7f, 0x6c, 
	0xfb, 0xef, 0xbf, 0xff, 0xff, 0x7f, 0xdf, 0xff, 
	0xfe, 0xfd, 0xb7, 0x9f, 0xff, 0xff, 0xff, 0xf7, 
	0xdf, 0xef, 0xff, 0x1f, 0xf8, 0xff, 0xff, 0xeb, 
	0xff, 0xff, 0xfa, 0xfb, 0xef, 0xcf, 0x7f, 0xff, 
	0xff, 0xff, 0x7f, 0xff, 0xff, 0xf7, 0xef, 0x6f, 
	0xbf, 0xff, 0xff, 0xd7, 0x5f, 0xff, 0xff, 0xff, 
	0xa9, 0xfd, 0x9f, 0xff, 0xff, 0xfa, 0x7b, 0xaa, 
	0xef, 0xff, 0x9b, 0xfb, 0xff, 0xff, 0x7f, 0xf7, 
	0x3f, 0x5e, 0xd9, 0xcd, 0x27, 0xff, 0xff, 0x73, 
	0xff, 0xff, 0xff, 0xff, 0x92, 0xfd, 0xff, 0xbe, 
	0xff, 0xf6, 0xef, 0xa9, 0xbc, 0xff, 0x7f, 0xfa, 
	0xf5, 0xfd, 0xc7, 0xfd, 0xff, 0xdf, 0x7f, 0xbf, 
	0x7e, 0xfe, 0xff, 0x2f, 0xff, 0xff, 0xff, 0xff, 
	0xad, 0xee, 0xff, 0xff, 0x7f, 0xf9, 0xf9, 0xbf, 
	0xfc, 0xfe, 0xbf, 0x39, 0xd7, 0xfc, 0xdf, 0x77, 
	0xff, 0xff, 0xf5, 0xdd, 0x7f, 0xff, 0xff, 0xbf, 
	0xf6, 0xfa, 0xff, 0xff, 0x7d, 0x53, 0xff, 0x7f, 
	0xfe, 0xbd, 0xf7, 0xff, 0xf2, 0xfb, 0xff, 0x27, 
	0xdd, 0xec, 0xf7, 0xdf, 0xff, 0x9f, 0xfd, 0xf7, 
	0xff, 0xff, 0xff, 0xff, 0xf4, 0xff, 0xff, 0xff, 
	0x57, 0x67, 0xfe, 0xff, 0xaf, 0xc9, 0x2e, 0xfb, 
	0x6b, 0x7f, 0xbd, 0x57, 0x57, 0xfc, 0xff, 0xff, 
	0xed, 0xbf, 0xff, 0x6b, 0xab, 0xff, 0xff, 0xff, 
	0x7f, 0xaa, 0xbf, 0xee, 0xbf, 0xbb, 0xfb, 0xff, 
	0xff, 0xf6, 0x7e, 0xad, 0xaf, 0xfe, 0xf7, 0x9b, 
	0xfd, 0x7f, 0xde, 0xff, 0xbd, 0xff, 0xff, 0x3f, 
	0xcf, 0xfe, 0xff, 0xff, 0x7f, 0xef, 0x5e, 0xfb, 
	0xff, 0x7c, 0xee, 0xff, 0xff, 0xff, 0xfb, 0xff, 
	0xbf, 0xca, 0xff, 0x7b, 0xaf, 0xf5, 0xbe, 0xff, 
	0xdf, 0xff, 0x6f, 0x6d, 0xf7, 0xff, 0xff, 0xff, 
	0xff, 0xbd, 0xfd, 0xff, 0xff, 0x5b, 0xb3, 0xff, 
	0xff, 0xff, 0xbf, 0xfd, 0xff, 0xda, 0xdf, 0xbf, 
	0xbf, 0x1a, 0xfe, 0xff, 0xdf, 0xfb, 0x17, 0xb5, 
	0xf7, 0xfe, 0xfb, 0xff, 0xff, 0x9f, 0xdf, 0xff, 
	0xfb, 0x6f, 0xe1, 0xfd, 0xff, 0xfe, 0xf3, 0xff, 
	0xbf, 0x7d, 0xf6, 0xff, 0xff, 0x3f, 0xff, 0x76, 
	0xbf, 0xdf, 0xae, 0xff, 0xed, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xb7, 0xdb, 0x7e, 0xde, 0xa7, 0xf7, 
	0xff, 0xdf, 0xff, 0x6f, 0xff, 0xf9, 0xf7, 0xfb, 
	0x7f, 0x9f, 0x79, 0xe6, 0x9d, 0xfe, 0x3c, 0xff, 
	0xed, 0xff, 0xdf, 0x7e, 0xff, 0xff, 0xb3, 0xff, 
	0x3e, 0x73, 0x3b, 0xec, 0xff, 0x5e, 0xfb, 0xef, 
	0xbe, 0xff, 0xee, 0xeb, 0xef, 0x3b, 0xdf, 0xfa, 
	0xfb, 0xdf, 0x7f, 0xfe, 0x9d, 0xf6, 0xda, 0x63, 
	0xff, 0xfd, 0xc7, 0xdd, 0x6d, 0xbc, 0x5b, 0x98, 
	0x7f, 0xff, 0xfd, 0xff, 0xdf, 0x7f, 0xff, 0xff, 
	0xff, 0xdd, 0x3f, 0xff, 0xfd, 0xff, 0xff, 0xef, 
	0xff, 0xfb, 0xef, 0xbf, 0xff, 0xff, 0xfb, 0xef, 
	0xdf, 0xdf, 0xce, 0xe6, 0xbe, 0xff, 0xfe, 0xff, 
	0xef, 0xff, 0xfe, 0xff, 0xff, 0xdf, 0xff, 0xfe, 
	0xfe, 0xff, 0xff, 0xdf, 0xff, 0xf7, 0xf7, 0xb7, 
	0xdf, 0xff, 0xfb, 0xdf, 0xff, 0xff, 0xbd, 0xaf, 
	0xfb, 0xbf, 0xbf, 0xff, 0x5f, 0xeb, 0xaf, 0xff, 
	0xfe, 0xde, 0x1f, 0xff, 0xfe, 0xfd, 0xb5, 0x3f, 
	0x5f, 0x7b, 0xed, 0xb5, 0xdf, 0xff, 0xef, 0xf5, 
	0xc7, 0xdf, 0xff, 0xbe, 0xde, 0xbd, 0xfd, 0xef, 
	0xbf, 0x2c, 0x9e, 0x78, 0x63, 0xbb, 0xff, 0xfa, 
	0xf9, 0xf7, 0x57, 0xff, 0x7d, 0xd5, 0xdd, 0x77, 
	0xdf, 0xfc, 0xbf, 0xdf, 0xfb, 0xfb, 0xfc, 0x7d, 
	0xf9, 0xff, 0x7f, 0x7b, 0xac, 0xbb, 0xfe, 0xfa, 
	0x6f, 0xaf, 0xfd, 0xdd, 0xfb, 0xff, 0xdf, 0xff, 
	0xf7, 0xf7, 0xd7, 0x7b, 0xbf, 0xfd, 0xff, 0xff, 
	0xff, 0xfe, 0xff, 0xff, 0xe8, 0xff, 0xff, 0xb3, 
	0x8d, 0x3e, 0xfa, 0xe8, 0xb3, 0x89, 0x3e, 0xff, 
	0x5c, 0xf3, 0x0f, 0x7f, 0xfe, 0xd9, 0xc7, 0x9f, 
	0x7f, 0xbc, 0xf9, 0xe7, 0x9e, 0x7f, 0xfe, 0xf9, 
	0x8b, 0xff, 0xbf, 0xff, 0xff, 0xff, 0xeb, 0x2f, 
	0xfe, 0xff, 0xfb, 0xef, 0xff, 0xff, 0xff, 0xf7, 
	0xbf, 0xdf, 0xff, 0xfd, 0xc5, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0x9f, 0xfe, 0xbf, 0xff, 
	0xfb, 0xbf, 0xaf, 0xbe, 0xfe, 0xef, 0xae, 0xff, 
	0xff, 0xff, 0xff, 0xdf, 0xff, 0x7f, 0xff, 0xff, 
	0xf7, 0xff, 0xff, 0xff, 0xbf, 0xff, 0xff, 0xfb, 
	0x3f, 0xf9, 0xff, 0xff, 0xff, 0xaf, 0xff, 0xfe, 
	0xfb, 0xff, 0xfe, 0xff, 0xff, 0xff, 0xfb, 0xff, 
	0xff, 0xeb, 0xff, 0xdf, 0x7f, 0xff, 0xf7, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xe1, 0xff, 0xff, 
	0xff, 0xbd, 0xfe, 0xfa, 0xab, 0xff, 0xff, 0xff, 
	0xdf, 0x7f, 0xbf, 0xff, 0xef, 0xff, 0xff, 0xfe, 
	0xfb, 0xef, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0x8f, 0xfd, 0xff, 0xff, 0xfc, 0xff, 0xcf, 
	0xff, 0xfe, 0xfd, 0xff, 0xff, 0xff, 0xff, 0xf7, 
	0xed, 0xff, 0xff, 0x7f, 0xfb, 0xff, 0x7f, 0xff, 
	0xff, 0xf7, 0xe7, 0xfe, 0xff, 0xff, 0xfa, 0xff, 
	0xa9, 0xaf, 0xbe, 0xfe, 0xf8, 0xeb, 0xaf, 0xbe, 
	0xfe, 0xfa, 0xeb, 0xaf, 0x7b, 0xf9, 0xd5, 0xd7, 
	0x1d, 0x7b, 0xfd, 0xb5, 0xd7, 0x5f, 0x7b, 0xfd, 
	0xf5, 0x7f, 0xf9, 0xfb, 0xf7, 0xff, 0xff, 0xff, 
	0xf7, 0xd7, 0xdb, 0xff, 0xfa, 0xef, 0xe3, 0xcf, 
	0x97, 0xfd, 0xf9, 0xbf, 0xf7, 0xff, 0xfb, 0xf7, 
	0xff, 0xff, 0xfe, 0xff, 0xd1, 0xff, 0xc5, 0xff, 
	0xfb, 0xff, 0xff, 0xdf, 0xff, 0xfa, 0xef, 0xd8, 
	0x7f, 0xff, 0xbb, 0xf6, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x5f, 
	0xfd, 0xff, 0xad, 0xff, 0xfb, 0xff, 0xbf, 0xff, 
	0xff, 0xfb, 0xff, 0xff, 0xef, 0xff, 0xdf, 0xbb, 
	0xd6, 0xdf, 0xd3, 0xef, 0x9f, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xef, 0xff, 0x7f, 0xfc, 
	0xff, 0x7b, 0xfb, 0xff, 0xe7, 0xbf, 0x7e, 0xfa, 
	0xae, 0xff, 0x9f, 0xfe, 0xbb, 0xdf, 0xff, 0x7f, 
	0xfc, 0xf5, 0xff, 0x7f, 0xff, 0xff, 0xf7, 0xdf, 
	0xdb, 0xff, 0x9f, 0xfb, 0xff, 0xff, 0x7b, 0xff, 
	0xff, 0xfd, 0xba, 0xef, 0xe7, 0xfb, 0xfb, 0xf6, 
	0xff, 0xce, 0xef, 0xff, 0xfb, 0xaf, 0xff, 0xff, 
	0xff, 0x7f, 0xf7, 0xdf, 0xf7, 0xfb, 0xff, 0xe5, 
	0xff, 0xef, 0x9f, 0xfd, 0xff, 0xd3, 0xff, 0xeb, 
	0xfb, 0xff, 0xf9, 0xff, 0xee, 0xff, 0xdf, 0xff, 
	0xff, 0x7f, 0xff, 0xff, 0xfd, 0xff, 0x79, 0xfe, 
	0xdf, 0xf7, 0xff, 0xbb, 0xff, 0x4f, 0xff, 0xff, 
	0xff, 0xf3, 0xff, 0xdf, 0xef, 0xef, 0xff, 0xff, 
	0xbf, 0xff, 0xff, 0xbf, 0xfd, 0xff, 0xfe, 0xff, 
	0xff, 0xff, 0x7f, 0xfa, 0xfd, 0xff, 0xff, 0x1f, 
	0xfe, 0xff, 0xfe, 0xdf, 0xff, 0xbf, 0xff, 0xdf, 
	0xff, 0xff, 0xfe, 0xff, 0xff, 0xde, 0xfb, 0xfb, 
	0xff, 0xff, 0xff, 0xfd, 0xef, 0xff, 0xff, 0xd7, 
	0xf7, 0xfe, 0xfe, 0xff, 0xfb, 0xfb, 0xef, 0xfd, 
	0xff, 0x3f, 0xdb, 0xfd, 0xbd, 0x5d, 0xef, 0x7f, 
	0xbd, 0xff, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xff, 
	0x7f, 0xff, 0xfb, 0xff, 0xef, 0xfd, 0xff, 0xff, 
	0xe0, 0xff, 0xff, 0xff, 0xef, 0xff, 0xfc, 0xf7, 
	0xff, 0xff, 0x9e, 0xff, 0xb7, 0xf3, 0xfe, 0x7f, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xfd, 0xef, 0xf7, 
	0xff, 0x7f, 0xdf, 0xff, 0x9b, 0xff, 0xff, 0xdf, 
	0xfd, 0xf7, 0xe9, 0x77, 0xff, 0xfe, 0xff, 0x4b, 
	0xad, 0xff, 0x3d, 0xff, 0xbf, 0xff, 0xff, 0xff, 
	0xef, 0xff, 0xff, 0x7f, 0xfd, 0x6f, 0xbf, 0xf7, 
	0x5f, 0xfc, 0xdf, 0xf7, 0xff, 0xef, 0xbf, 0xff, 
	0xff, 0x91, 0xbf, 0x3f, 0xff, 0xf2, 0xfa, 0xff, 
	0xff, 0xff, 0xff, 0xef, 0xff, 0xff, 0xbf, 0xff, 
	0xf7, 0xd3, 0xff, 0xf7, 0xff, 0xf8, 0xff, 0xff, 
	0xbf, 0xbf, 0xde, 0xfe, 0xff, 0xed, 0xff, 0x7e, 
	0xff, 0xfb, 0xff, 0xff, 0xff, 0xf7, 0xdf, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xdf, 0xff, 0xff, 0xff, 
	0xff, 0x81, 0xff, 0xff, 0xff, 0xff, 0x5d, 0xea, 
	0xbf, 0xaf, 0xff, 0xfb, 0xfb, 0xff, 0xbf, 0xff, 
	0x3f, 0xff, 0xff, 0xff, 0xce, 0x3f, 0xc5, 0x7f, 
	0xff, 0xff, 0xef, 0xff, 0xff, 0xb4, 0xfd, 0xff, 
	0xff, 0xf7, 0x3f, 0xeb, 0x7f, 0xb1, 0x9e, 0xcf, 
	0xff, 0xff, 0xbf, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0x9f, 0xff, 0xd2, 0xf3, 0xfd, 0xff, 0xff, 0xff, 
	0xff, 0xed, 0xfe, 0xa7, 0xff, 0xff, 0xdf, 0xff, 
	0xfc, 0xfc, 0xfa, 0xdf, 0xb5, 0xbf, 0xcd, 0x7f, 
	0xd5, 0xff, 0xfe, 0x9f, 0xbd, 0xbf, 0x5f, 0xfd, 
	0xbf, 0xf4, 0xd4, 0xeb, 0xff, 0x7f, 0x5a, 0xcf, 
	0xff, 0xff, 0x5f, 0xf7, 0x9f, 0xfb, 0xeb, 0xff, 
	0xf7, 0xff, 0xec, 0xff, 0x5d, 0xff, 0xff, 0xff, 
	0xf6, 0xff, 0x7f, 0xfd, 0xff, 0xd4, 0x59, 0x7f, 
	0xff, 0xd7, 0xe7, 0xfe, 0xff, 0xff, 0xd7, 0xbd, 
	0xff, 0x3f, 0xaf, 0xff, 0xb9, 0x9b, 0xfd, 0xaf, 
	0xc5, 0xff, 0x9f, 0xff, 0x57, 0xff, 0xe7, 0xf5, 
	0xff, 0xff, 0xb3, 0xcf, 0x34, 0xbb, 0xb7, 0xf9, 
	0xff, 0xbe, 0x6b, 0xf7, 0xbf, 0xff, 0xbe, 0xfa, 
	0xff, 0x7f, 0x75, 0xfe, 0x77, 0xdf, 0xff, 0xff, 
	0x7f, 0xf5, 0xff, 0x57, 0xff, 0x7f, 0xef, 0xbc, 
	0xd2, 0xee, 0xcc, 0xec, 0xff, 0xf7, 0xff, 0xff, 
	0xff, 0xd7, 0xfa, 0xfa, 0xff, 0xff, 0xbf, 0xff, 
	0xdf, 0xd6, 0xff, 0xff, 0xbf, 0x7c, 0x7e, 0x5f, 
	0xff, 0xff, 0xf5, 0xd5, 0x5f, 0xdf, 0x9b, 0xb8, 
	0xff, 0xf7, 0xfb, 0x7f, 0xf7, 0xdf, 0x5b, 0xeb, 
	0x7f, 0xf7, 0xfe, 0xfe, 0xff, 0x5f, 0xff, 0x97, 
	0xff, 0xf7, 0xf9, 0x6f, 0xfd, 0xff, 0xdf, 0x7f, 
	0xff, 0xfd, 0xef, 0xec, 0xfd, 0xff, 0xff, 0xff, 
	0xbd, 0xff, 0x7e, 0xfb, 0xff, 0xbd, 0xbf, 0xfd, 
	0xf6, 0xb3, 0xff, 0xaf, 0xff, 0xff, 0xff, 0x6f, 
	0xff, 0xff, 0xed, 0xe7, 0xff, 0x7f, 0xde, 0xab, 
	0xf7, 0x7f, 0xfe, 0xff, 0xf7, 0xbf, 0xff, 0xff, 
	0xdf, 0x67, 0xff, 0x7d, 0xf6, 0x9f, 0xff, 0x3f, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xcd, 0x3f, 
	0xff, 0xfe, 0x73, 0xbf, 0xec, 0xff, 0x72, 0x7b, 
	0xcd, 0xb7, 0xd7, 0xfe, 0x9b, 0xcd, 0xb6, 0xd3, 
	0x3a, 0xfb, 0xdf, 0x7f, 0xfe, 0xbd, 0xf6, 0xdf, 
	0x7f, 0xfb, 0xfd, 0xf7, 0xdf, 0x7d, 0xff, 0x1b, 
	0x9b, 0x7f, 0xbf, 0xfd, 0xff, 0xdb, 0x77, 0xff, 
	0xff, 0xf3, 0xd9, 0xff, 0xbf, 0xff, 0xff, 0xff, 
	0xef, 0xff, 0xfb, 0xff, 0xff, 0xff, 0xbe, 0xff, 
	0xff, 0xbf, 0xff, 0xcf, 0xe1, 0xbe, 0xff, 0xfd, 
	0xff, 0xdf, 0xaf, 0xff, 0xff, 0xef, 0xdb, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xdf, 0xff, 0xfb, 0xff, 
	0xff, 0x7f, 0x7f, 0xff, 0xff, 0xdf, 0xff, 0xbf, 
	0xb3, 0xfa, 0xfd, 0xbf, 0xff, 0x73, 0x7f, 0xbf, 
	0xff, 0xdf, 0xc3, 0xfb, 0xbf, 0xb6, 0xff, 0xff, 
	0xbe, 0x5f, 0x6f, 0xfd, 0xf5, 0xff, 0x7b, 0x6b, 
	0xbf, 0xfd, 0xf6, 0xff, 0x0a, 0xde, 0xff, 0xd9, 
	0xef, 0xef, 0x6f, 0xbf, 0xf9, 0xf7, 0xdb, 0xff, 
	0xf6, 0xf9, 0xef, 0xff, 0xff, 0x6d, 0xf9, 0xd5, 
	0x13, 0xff, 0xff, 0xf3, 0xdd, 0x77, 0x7f, 0xfd, 
	0xfd, 0xf8, 0xff, 0xff, 0x7f, 0xb7, 0xff, 0xff, 
	0xfb, 0xf9, 0x75, 0xbf, 0xe7, 0xf7, 0x7b, 0xff, 
	0xff, 0xff, 0xf4, 0xd6, 0x5d, 0xef, 0xff, 0xff, 
	0xd7, 0x5f, 0xef, 0xfd, 0xff, 0xe7, 0xff, 0x7f, 
	0xb3, 0xcf, 0x3e, 0xfe, 0xf8, 0xe3, 0xcf, 0x2f, 
	0xdb, 0x7c, 0xb3, 0x8d, 0x7f, 0xfe, 0xd1, 0x47, 
	0x1b, 0x7d, 0xfe, 0xf9, 0x47, 0x1f, 0x7d, 0xfc, 
	0xf9, 0xb7, 0xff, 0xbf, 0xff, 0xfd, 0xfa, 0x9f, 
	0x7f, 0xff, 0xff, 0xbb, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0x5f, 0xfc, 0xfd, 0xb5, 0xbf, 0xff, 
	0xff, 0xfd, 0xf7, 0xd7, 0x7f, 0xdf, 0xfe, 0x9f, 
	0xff, 0xfb, 0xef, 0xef, 0xff, 0xff, 0xff, 0xfb, 
	0xbb, 0xff, 0xff, 0xfb, 0xff, 0xff, 0x7f, 0xfd, 
	0xfd, 0xd7, 0xff, 0xfe, 0xef, 0xb7, 0xdf, 0x5e, 
	0xfb, 0xff, 0xfb, 0xff, 0xfb, 0xdf, 0xaf, 0xff, 
	0xff, 0xf7, 0xfb, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xef, 0xf5, 0xdf, 0x5f, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xe6, 0xff, 
	0x7b, 0x3f, 0xfd, 0xfe, 0xfe, 0xbf, 0x6f, 0xff, 
	0xff, 0xdf, 0x7f, 0xff, 0xfd, 0xff, 0xff, 0x97, 
	0xdf, 0x7d, 0xfd, 0xef, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0x97, 0xff, 0x6f, 0xbf, 0xfd, 0xff, 
	0x3f, 0x4f, 0xff, 0xbf, 0xff, 0xdf, 0x6f, 0xbf, 
	0xfd, 0xff, 0xff, 0xff, 0x7f, 0xfb, 0xff, 0xff, 
	0xff, 0x7f, 0xf3, 0xed, 0xf7, 0xff, 0xdf, 0xfa, 
	0xef, 0xeb, 0xaf, 0xbe, 0xfe, 0xba, 0xeb, 0xaf, 
	0xbf, 0xfc, 0xd8, 0xeb, 0xae, 0x7e, 0xfd, 0x95, 
	0xd7, 0x57, 0x7f, 0xfd, 0xf5, 0xd7, 0x5f, 0x7f, 
	0xdd, 0xf1, 0x2f, 0xf8, 0xff, 0xe0, 0xc3, 0x3e, 
	0xff, 0xff, 0xfb, 0xbf, 0x3f, 0xbf, 0xfb, 0xe0, 
	0xbb, 0xff, 0xff, 0xfd, 0x7f, 0xff, 0xfd, 0xff, 
	0xff, 0xd7, 0x67, 0x9f, 0xfd, 0xf7, 0xdf, 0xc0, 
	0xff, 0xfb, 0xbe, 0x5f, 0xbe, 0xff, 0xef, 0xfe, 
	0x7f, 0x7b, 0xf7, 0xba, 0xff, 0xdf, 0xff, 0xff, 
	0xff, 0xff, 0xf7, 0xdf, 0x7e, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xa5, 0xff, 0xdf, 0xbd, 0x7f, 
	0xf7, 0xbf, 0xff, 0xef, 0xff, 0xeb, 0xff, 0xdf, 
	0xff, 0x7f, 0xff, 0xea, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xbf, 0xff, 0xff, 0xff, 0xfe, 0xff, 0x5f, 
	0xf6, 0xbf, 0xee, 0xfa, 0xe3, 0xff, 0x1e, 0xfe, 
	0x7f, 0xbe, 0xbf, 0xff, 0xfe, 0xfb, 0xff, 0xff, 
	0xcd, 0xff, 0xf7, 0xff, 0xff, 0x7f, 0xff, 0xf7, 
	0xdf, 0x5f, 0xff, 0x3f, 0xfb, 0xff, 0xeb, 0xd9, 
	0xdf, 0xff, 0xff, 0xfd, 0x7f, 0xf7, 0xff, 0xdf, 
	0xff, 0xfd, 0xea, 0xff, 0xff, 0xff, 0xf7, 0xff, 
	0xdf, 0xff, 0xef, 0xff, 0xff, 0xff, 0xbd, 0xff, 
	0xe2, 0xff, 0xfb, 0xbf, 0x7f, 0xfc, 0xbf, 0xdf, 
	0xbf, 0xff, 0xfc, 0xba, 0xfb, 0xef, 0xbf, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xdf, 
	0x7f, 0xff, 0xf5, 0xfd, 0x97, 0xff, 0xff, 0xf9, 
	0xff, 0xfe, 0xff, 0xff, 0x3f, 0xf6, 0xef, 0xdf, 
	0xff, 0xfd, 0xf7, 0xee, 0xff, 0xff, 0xfe, 0xff, 
	0xff, 0xff, 0xdf, 0x7f, 0xfd, 0xff, 0xff, 0xff, 
	0x0f, 0xfe, 0xff, 0xf7, 0xf7, 0xfb, 0xff, 0xdf, 
	0xdf, 0xff, 0xdf, 0x7f, 0xfd, 0xef, 0xdf, 0xfb, 
	0xfb, 0xff, 0x7f, 0xff, 0xff, 0xff, 0xff, 0xfd, 
	0xff, 0xff, 0xff, 0xfc, 0xbf, 0xca, 0xff, 0xef, 
	0xdf, 0xf7, 0x3f, 0xdb, 0xef, 0xff, 0xcf, 0xfb, 
	0xff, 0xfd, 0xff, 0xfb, 0xff, 0xff, 0x7f, 0xc7, 
	0xff, 0xdf, 0xff, 0xff, 0xff, 0xff, 0xdf, 0xff, 
	0xff, 0x6b, 0xff, 0xff, 0xff, 0xff, 0xff, 0x76, 
	0xff, 0xff, 0xbf, 0xd9, 0xff, 0x9f, 0xfd, 0xff, 
	0xff, 0xff, 0xff, 0x7f, 0xff, 0xff, 0xff, 0x7d, 
	0xff, 0xff, 0xff, 0xff, 0xfe, 0x92, 0xff, 0x7f, 
	0xff, 0xfd, 0xfa, 0xec, 0xff, 0xff, 0xfd, 0xf3, 
	0xdf, 0xaf, 0xff, 0x7f, 0xff, 0xbf, 0x7f, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xfe, 0xff, 0xff, 0xbf, 
	0xff, 0x6f, 0xfc, 0xff, 0xf7, 0xff, 0xef, 0x1f, 
	0xff, 0xf3, 0xdf, 0xe3, 0xf7, 0xbf, 0xfa, 0xff, 
	0xff, 0xff, 0xef, 0xfd, 0xff, 0xff, 0xff, 0xbf, 
	0xfd, 0xff, 0xff, 0xff, 0xff, 0x1f, 0xfb, 0xff, 
	0xff, 0xff, 0xbf, 0xff, 0xfe, 0xff, 0xff, 0xff, 
	0xff, 0xdf, 0xff, 0xff, 0xff, 0xbf, 0xff, 0xf5, 
	0xff, 0xff, 0xff, 0xff, 0xf7, 0xff, 0xff, 0xf7, 
	0xff, 0xff, 0xaa, 0xff, 0x2b, 0xff, 0xf1, 0xfe, 
	0xff, 0xeb, 0xbf, 0xb1, 0xf7, 0xdf, 0xeb, 0x79, 
	0xfc, 0xff, 0xb7, 0xfc, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x9a, 0xfd, 
	0xff, 0xfe, 0xcf, 0xdf, 0xff, 0xaf, 0xbf, 0x5e, 
	0xff, 0xff, 0xaf, 0xff, 0xff, 0xff, 0xff, 0xf2, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xdd, 0xfe, 0xbf, 0x9e, 0xff, 0xdf, 
	0xff, 0xbd, 0xce, 0xfa, 0xdf, 0xff, 0xbf, 0xfc, 
	0xef, 0xff, 0x7f, 0xfe, 0xfb, 0xff, 0xff, 0xff, 
	0x4f, 0xb9, 0xe4, 0xff, 0x4b, 0xaf, 0x3f, 0x52, 
	0xff, 0x7f, 0xfa, 0x5c, 0xff, 0xff, 0x3a, 0xeb, 
	0xdf, 0xb3, 0xff, 0xea, 0xff, 0xff, 0xff, 0xf7, 
	0xdf, 0xff, 0xff, 0xff, 0x9f, 0xf5, 0xd6, 0xff, 
	0x6f, 0x9d, 0xd7, 0xc5, 0xbe, 0x69, 0xf6, 0xdd, 
	0x5f, 0xd7, 0xeb, 0xbf, 0xff, 0xf3, 0x9f, 0xe9, 
	0xff, 0xff, 0xff, 0xb7, 0xfb, 0xff, 0xff, 0xff, 
	0xff, 0x7f, 0x46, 0xf9, 0x7f, 0xb5, 0xbf, 0xaf, 
	0xf9, 0xf6, 0xfd, 0x27, 0xaf, 0xfd, 0xaf, 0xbe, 
	0x5f, 0xef, 0x7f, 0xa7, 0xfe, 0xff, 0xff, 0x7f, 
	0xe5, 0xff, 0xff, 0xff, 0xff, 0xdf, 0x13, 0xed, 
	0xff, 0xd7, 0xfe, 0x2c, 0xec, 0xbf, 0xfe, 0xff, 
	0xff, 0xff, 0x96, 0xfe, 0xfa, 0xff, 0xfd, 0xaf, 
	0x5e, 0xff, 0xff, 0x7f, 0xed, 0xff, 0xff, 0xff, 
	0xff, 0x7f, 0xfd, 0xff, 0xfa, 0xff, 0xff, 0x9b, 
	0xbb, 0xff, 0xdf, 0xff, 0x7f, 0xff, 0xdd, 0xfa, 
	0xeb, 0x7f, 0xee, 0xff, 0xfa, 0xff, 0xff, 0xff, 
	0x3f, 0xff, 0xff, 0xff, 0xff, 0xff, 0x77, 0xfe, 
	0xff, 0xff, 0xff, 0xef, 0xce, 0xfd, 0xff, 0xff, 
	0xfb, 0xff, 0xd7, 0xfe, 0xff, 0xff, 0xff, 0x3f, 
	0xff, 0xff, 0xbf, 0xff, 0x3f, 0xfd, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xfc, 0xfb, 0xff, 0xff, 0xdf, 
	0x11, 0xf7, 0xff, 0xff, 0x7f, 0xe7, 0x9f, 0xff, 
	0xff, 0x9b, 0xf7, 0xff, 0xf9, 0xf6, 0xff, 0xff, 
	0x7f, 0xef, 0xfd, 0xff, 0xff, 0xff, 0xff, 0xbf, 
	0xff, 0xff, 0xff, 0x7f, 0x8f, 0xe8, 0xff, 0xfe, 
	0x7b, 0xad, 0x3f, 0xff, 0x7e, 0xfb, 0xef, 0xb5, 
	0xdf, 0xee, 0xfb, 0xdf, 0x7f, 0xaf, 0xfd, 0xf7, 
	0xdb, 0x7f, 0xff, 0xfd, 0xf6, 0xdf, 0x6f, 0xbf, 
	0xdb, 0x99, 0xff, 0xff, 0xff, 0xff, 0xff, 0xf7, 
	0xff, 0xfc, 0xff, 0xdf, 0x3f, 0xdf, 0xff, 0xff, 
	0xff, 0xef, 0xbe, 0xff, 0xe7, 0xff, 0xff, 0xbe, 
	0xf9, 0xfe, 0x9f, 0x7f, 0xce, 0xe5, 0xfe, 0xff, 
	0xff, 0xff, 0xff, 0xef, 0xff, 0xfb, 0xff, 0xbf, 
	0xff, 0xbe, 0xff, 0xff, 0xff, 0xdf, 0x7d, 0xff, 
	0xdf, 0xff, 0xff, 0xfd, 0xf7, 0xff, 0x7f, 0xff, 
	0xbd, 0x9f, 0xfb, 0xfd, 0xbd, 0xff, 0x7f, 0xcf, 
	0xff, 0xff, 0x7e, 0xeb, 0xfb, 0xb7, 0xfd, 0xff, 
	0xff, 0xdf, 0xff, 0xff, 0xff, 0xfd, 0xff, 0xfb, 
	0x6f, 0xff, 0xf4, 0xf6, 0xff, 0xbe, 0xde, 0xff, 
	0xdb, 0xef, 0x8f, 0x35, 0xfe, 0xff, 0xab, 0x8d, 
	0xff, 0xbe, 0xf9, 0xff, 0xff, 0x7f, 0xfe, 0xb3, 
	0xff, 0x5f, 0xff, 0x7f, 0xf5, 0xd7, 0x97, 0x7f, 
	0xfd, 0xfd, 0xf9, 0xff, 0x9f, 0x7f, 0xfd, 0xff, 
	0xee, 0xff, 0x7b, 0x3f, 0xfd, 0xc7, 0xfb, 0xff, 
	0xff, 0xff, 0xbf, 0xf6, 0xff, 0xdb, 0xf7, 0xff, 
	0xf7, 0xf6, 0x5f, 0x6f, 0xff, 0xff, 0xee, 0xff, 
	0x7f, 0xe3, 0xcd, 0x36, 0x7b, 0xfc, 0xe3, 0xcf, 
	0x3e, 0xde, 0x7c, 0xf3, 0x8f, 0x7f, 0xfe, 0xd9, 
	0xe7, 0x9f, 0x7d, 0xfe, 0xf1, 0x67, 0x1f, 0x7d, 
	0xf6, 0xf9, 0x8f, 0xff, 0xaf, 0xff, 0xff, 0xff, 
	0x9f, 0xf5, 0xff, 0xfe, 0xff, 0xfb, 0x3f, 0xff, 
	0xff, 0xff, 0x1f, 0xff, 0xff, 0xfb, 0xfd, 0xff, 
	0xff, 0x7f, 0xff, 0xff, 0xf7, 0xff, 0xcf, 0xfe, 
	0xbf, 0xfe, 0xfe, 0xef, 0xff, 0xff, 0xff, 0xfb, 
	0xff, 0xef, 0xbf, 0xfd, 0xff, 0xff, 0x7f, 0xff, 
	0xff, 0xef, 0xdf, 0xff, 0x7f, 0xed, 0xb7, 0xfd, 
	0x7f, 0xfb, 0x3f, 0xfb, 0xff, 0xfa, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xeb, 
	0xff, 0xff, 0xff, 0xff, 0xf7, 0xff, 0xdf, 0xff, 
	0xff, 0xf7, 0xf7, 0xff, 0x7f, 0xff, 0xff, 0xe8, 
	0xfd, 0x6b, 0xef, 0xfd, 0xf6, 0xef, 0xff, 0xaf, 
	0xfd, 0xf5, 0xdf, 0xff, 0xff, 0xbf, 0xff, 0xff, 
	0x57, 0xff, 0xfe, 0xf5, 0xff, 0xf7, 0x7f, 0xf7, 
	0xff, 0xf7, 0xff, 0x83, 0xf7, 0x6b, 0xff, 0xfd, 
	0xf6, 0xff, 0xff, 0x2f, 0xfd, 0xe4, 0xfb, 0xff, 
	0xf7, 0xff, 0xff, 0xb7, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xdf, 0xfe, 0x7f, 0xff, 0xff, 0xf7, 0x8f, 
	0xee, 0xff, 0xab, 0x8f, 0xbd, 0xfa, 0xea, 0x6b, 
	0xae, 0xb5, 0xe6, 0xf8, 0xeb, 0xaf, 0x7f, 0xed, 
	0xf5, 0x87, 0x5f, 0x7f, 0xfd, 0x75, 0xd7, 0x5f, 
	0x3f, 0xfd, 0xf5, 0x37, 0xfb, 0xff, 0xe0, 0xfa, 
	0x0f, 0xfc, 0xfc, 0xff, 0xff, 0xff, 0xef, 0xfb, 
	0xfb, 0xff, 0xff, 0x7f, 0xf0, 0xe7, 0xff, 0xff, 
	0xff, 0xff, 0xfd, 0xff, 0xfb, 0xff, 0xff, 0xff, 
	0xc0, 0xff, 0xeb, 0xfe, 0xbf, 0xff, 0xee, 0xfd, 
	0xff, 0xff, 0xeb, 0xff, 0xff, 0xf7, 0xff, 0xff, 
	0xdb, 0xee, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0x89, 0xff, 0xff, 0xfb, 
	0xff, 0xfd, 0xef, 0xff, 0xff, 0xbd, 0xef, 0xff, 
	0xdb, 0xdf, 0x7f, 0xdf, 0xef, 0xb7, 0xff, 0x7b, 
	0xfe, 0xff, 0xff, 0xff, 0xff, 0x7f, 0xfd, 0xf7, 
	0xaf, 0xfe, 0xff, 0xde, 0xf3, 0xff, 0x3b, 0xff, 
	0xfd, 0xbc, 0xef, 0x2f, 0xff, 0xff, 0xef, 0x7f, 
	0xff, 0x7e, 0xfe, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xc7, 0xff, 0xeb, 0x5f, 0xfa, 0xff, 0xff, 
	0xfb, 0xf7, 0xbb, 0xbf, 0xff, 0xff, 0xef, 0xfb, 
	0xfe, 0xbf, 0xfe, 0xeb, 0xd5, 0xff, 0xfe, 0xff, 
	0xfd, 0xdf, 0xff, 0xff, 0xff, 0xdf, 0xbb, 0xef, 
	0xfe, 0xe9, 0xff, 0xfb, 0xff, 0xfa, 0xfe, 0xff, 
	0xff, 0xff, 0xff, 0xfe, 0xfe, 0xef, 0xff, 0xff, 
	0xfd, 0xff, 0xfb, 0x77, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0x83, 0xff, 0xaf, 
	0xff, 0xfd, 0xfe, 0xdf, 0x7f, 0xff, 0xfd, 0xee, 
	0xff, 0xcf, 0xbf, 0xe7, 0xfe, 0xbf, 0xff, 0x7f, 
	0xfd, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xfd, 0x8f, 0xfe, 0xff, 0xff, 0xff, 0xcf, 0xfe, 
	0xff, 0xff, 0xff, 0xdb, 0xfb, 0xbf, 0xdf, 0xfe, 
	0xbf, 0xff, 0xff, 0xff, 0xef, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xf8, 0xff, 
	0xff, 0xff, 0xfe, 0xfd, 0xfb, 0xff, 0xff, 0xff, 
	0xdf, 0xff, 0xff, 0xff, 0xff, 0xdf, 0xff, 0xff, 
	0xff, 0xfe, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xe9, 0x7f, 0xff, 0xfd, 0xff, 0xfb, 
	0xff, 0xff, 0xff, 0xef, 0xff, 0xbf, 0xff, 0xff, 
	0xef, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xfb, 0xff, 0xef, 0xff, 0xff, 0x8f, 0xfb, 
	0xff, 0xff, 0xdb, 0x6a, 0xbf, 0xfc, 0xf7, 0xff, 
	0xfb, 0xfe, 0xff, 0xbf, 0xfe, 0xff, 0xde, 0xfb, 
	0xef, 0xf5, 0xbf, 0x7f, 0xff, 0xfd, 0xff, 0xff, 
	0x7f, 0xff, 0xcf, 0xec, 0xff, 0xff, 0x5f, 0xff, 
	0xfd, 0xff, 0xef, 0xf7, 0x47, 0xff, 0xfd, 0xf7, 
	0xfb, 0x7f, 0x3f, 0xf7, 0xfb, 0xff, 0xff, 0xff, 
	0xfb, 0xfb, 0xff, 0xff, 0xff, 0xff, 0x7f, 0xfb, 
	0xff, 0xaf, 0xff, 0xff, 0xff, 0xfb, 0xbf, 0xff, 
	0xbe, 0xf5, 0xff, 0xff, 0xf7, 0xff, 0xf7, 0xed, 
	0xfe, 0xf7, 0xfd, 0xbf, 0xff, 0xff, 0xff, 0xfd, 
	0xff, 0xff, 0xfe, 0x69, 0xdf, 0x4d, 0x3f, 0xef, 
	0xfc, 0xf3, 0xcf, 0x3f, 0xef, 0xfd, 0xf3, 0xcf, 
	0x3f, 0xdf, 0xf9, 0xef, 0x9f, 0xff, 0xfe, 0xf9, 
	0xe7, 0x9f, 0x7f, 0xfe, 0xf9, 0xe7, 0x7f, 0x8d, 
	0xbf, 0xff, 0xef, 0x7d, 0xf7, 0xce, 0x7f, 0xff, 
	0xfd, 0xff, 0xdf, 0xff, 0xef, 0xfc, 0xff, 0xbf, 
	0xff, 0xfe, 0xfb, 0xe7, 0x9b, 0xff, 0xfe, 0xf9, 
	0xef, 0xbf, 0xff, 0xbf, 0xfe, 0x1f, 0xfe, 0x68, 
	0xc7, 0xbf, 0xbf, 0xfe, 0xfb, 0x6d, 0xb7, 0xff, 
	0x4a, 0x33, 0xd7, 0x4f, 0xff, 0xfd, 0xfe, 0xdf, 
	0x1f, 0xff, 0xfc, 0xf6, 0xdf, 0x6f, 0xff, 0xf7, 
	0xf9, 0xff, 0xff, 0x7f, 0xfe, 0xff, 0xef, 0xff, 
	0x7f, 0xff, 0xff, 0xef, 0xff, 0xdf, 0xef, 0xfb, 
	0xff, 0xff, 0xff, 0xff, 0xfe, 0xfd, 0xff, 0xdf, 
	0xff, 0xff, 0xef, 0xdf, 0xef, 0xff, 0xeb, 0xaf, 
	0xdf, 0xfe, 0xf3, 0xeb, 0xbf, 0xdf, 0xdf, 0xfb, 
	0x7f, 0xb7, 0xf7, 0xfc, 0xf3, 0xdf, 0x7f, 0xff, 
	0xfd, 0xf5, 0xdf, 0x6f, 0xff, 0xfd, 0xf7, 0xff, 
	0xbf, 0xff, 0xdf, 0x7f, 0xff, 0x6d, 0xf3, 0xdf, 
	0x3f, 0xff, 0xfc, 0xf3, 0xcf, 0x3d, 0xf7, 0xfb, 
	0xef, 0x9f, 0x7f, 0xfe, 0xf9, 0xef, 0xbf, 0x7f, 
	0xfe, 0xf9, 0xe7, 0xff, 0xff, 0xfe, 0xbf, 0xfd, 
	0xfe, 0x5b, 0xeb, 0xbb, 0xff, 0xfe, 0xfb, 0xef, 
	0xbf, 0xff, 0xf6, 0xb7, 0xdf, 0x7f, 0xff, 0xfd, 
	0xf7, 0xdf, 0x7f, 0xff, 0x7d, 0xf7, 0xdf, 0xff, 
	0xff, 0xf9, 0xff, 0xf9, 0xa7, 0xdf, 0x67, 0xfa, 
	0xe9, 0xa7, 0xdf, 0x67, 0xfa, 0xf9, 0xf6, 0x39, 
	0xff, 0xf6, 0xd3, 0x6f, 0x3f, 0xfd, 0xf4, 0xd3, 
	0x4f, 0x3f, 0xfd, 0xf4, 0xff, 0xe9, 0xff, 0xf7, 
	0xdf, 0x7f, 0xf5, 0xfd, 0xf7, 0xdf, 0x7f, 0xf5, 
	0xdd, 0xf7, 0x5f, 0xeb, 0xfe, 0xab, 0xef, 0xbd, 
	0xff, 0xce, 0xfb, 0xee, 0xbf, 0xff, 0xfe, 0xfb, 
	0xff, 0xa3, 0xff, 0xff, 0xdf, 0xfb, 0xfb, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xfd, 0xff, 0xdc, 0xfe, 
	0xbf, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0x0f, 0xfe, 0xff, 
	0xff, 0xff, 0xff, 0xdf, 0x7f, 0xff, 0xfd, 0xd7, 
	0xdf, 0x7f, 0xff, 0xff, 0x6e, 0xbf, 0xbd, 0xff, 
	0xdb, 0xef, 0xff, 0xff, 0xfe, 0xfb, 0xef, 0xbf, 
	0xff, 0x3f, 0xf9, 0xff, 0xfd, 0xf5, 0xdf, 0xff, 
	0xff, 0xff, 0xff, 0xf7, 0xdf, 0x7f, 0xff, 0xf1, 
	0xff, 0xff, 0xff, 0xfb, 0x7f, 0xef, 0xff, 0xfe, 
	0xfe, 0xff, 0xef, 0xff, 0xff, 0xff, 0xe3, 0xff, 
	0x9f, 0x7f, 0x7d, 0xf9, 0xf7, 0xdd, 0x7f, 0xff, 
	0xfd, 0xd7, 0xdf, 0x7f, 0xff, 0xfb, 0xef, 0xbf, 
	0xff, 0xfe, 0xeb, 0xef, 0xbf, 0xff, 0xfe, 0xfb, 
	0xef, 0xff, 0xab, 0xfb, 0x4f, 0x7b, 0xed, 0x37, 
	0xdf, 0x7e, 0xfb, 0x6d, 0x37, 0xdf, 0x7e, 0x93, 
	0x4d, 0x6f, 0xbb, 0x9d, 0xf6, 0xdb, 0x6f, 0xa7, 
	0xfd, 0xf6, 0xdb, 0x6f, 0xbb, 0xfd, 0x2f, 0xfe, 
	0xff, 0xb7, 0xdf, 0x5e, 0xf1, 0xed, 0xf7, 0xdf, 
	0x5f, 0xfd, 0xed, 0xd7, 0x57, 0xff, 0xfe, 0xfb, 
	0xef, 0xbf, 0xff, 0xfe, 0xfb, 0xef, 0xbf, 0xff, 
	0xfe, 0xfb, 0x3f, 0xfa, 0xff, 0xf7, 0xdf, 0x77, 
	0xff, 0x7d, 0xf7, 0xdf, 0x7f, 0xff, 0xfd, 0xf7, 
	0xdf, 0xff, 0xfe, 0xfb, 0xef, 0xbf, 0xff, 0xfe, 
	0xfb, 0xef, 0xbf, 0xff, 0xfe, 0xfb, 0xef, 0xe6, 
	0xff, 0xbf, 0xff, 0x6e, 0xbb, 0xef, 0xbe, 0xff, 
	0x7e, 0xfb, 0xef, 0xbf, 0xff, 0xfe, 0xf7, 0xdf, 
	0x7f, 0xff, 0xfd, 0xf7, 0xdf, 0x7f, 0xff, 0xfd, 
	0xf7, 0xdf, 0xff, 0xaf, 0xff, 0xff, 0xea, 0xeb, 
	0xef, 0xbf, 0xfe, 0xfa, 0xeb, 0xef, 0xbf, 0xfa, 
	0xfe, 0xfb, 0x5f, 0x7f, 0xfd, 0xf5, 0xf7, 0x5f, 
	0x7f, 0xfd, 0xf5, 0xf7, 0x5f, 0x7f, 0xfd, 0xef, 
	0xfe, 0xff, 0xfb, 0xef, 0xbf, 0xfb, 0xfe, 0xfb, 
	0xef, 0xbf, 0xff, 0xfe, 0x3b, 0xef, 0x7f, 0xff, 
	0xfd, 0xf7, 0xdf, 0x7f, 0xff, 0xfd, 0xf7, 0xdf, 
	0x7d, 0xff, 0xfd, 0x7f, 0xfa, 0xff, 0xcf, 0x3f, 
	0xef, 0xfc, 0xf3, 0xcf, 0x3f, 0xef, 0xfc, 0xf3, 
	0xcf, 0x3f, 0xff, 0xf9, 0xe7, 0x9f, 0x7f, 0xfe, 
	0xf9, 0xe7, 0x9f, 0x7f, 0xee, 0xf9, 0xe7, 0xf7, 
	0xe1, 0xff, 0x5f, 0x7f, 0x7d, 0xb5, 0xd7, 0x5f, 
	0x7f, 0x7d, 0x75, 0xd7, 0x5f, 0x4f, 0xdd, 0xeb, 
	0xaf, 0xbf, 0xfe, 0xfa, 0xeb, 0xaf, 0xbf, 0xfe, 
	0xfa, 0xeb, 0xaf, 0xff, 0xab, 0xff, 0xff, 0xf5, 
	0xd7, 0x1b, 0x7f, 0xfd, 0xf5, 0xd7, 0x1b, 0x7f, 
	0xfd, 0xf1, 0xc6, 0xbf, 0xfe, 0xfa, 0xeb, 0xae, 
	0xbf, 0xfe, 0xfa, 0xeb, 0x2f, 0xbf, 0xfe, 0xfa, 
	0x8f, 0xee, 0xff, 0xf6, 0xfb, 0x6f, 0xbf, 0xff, 
	0xfe, 0xfb, 0x2f, 0xbf, 0xff, 0xf6, 0xfb, 0x5f, 
	0x7f, 0xff, 0xfd, 0xf3, 0xdf, 0x7f, 0xff, 0xfd, 
	0xf7, 0xdb, 0x7f, 0xff, 0xff, 0xf9, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xfe, 
	0xff, 0xff, 0xff, 0xff, 0xfd, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xef, 0xff, 0x5b, 0xef, 0xbf, 0xf5, 0xfe, 
	0xfb, 0xef, 0xbf, 0xf7, 0xfe, 0x7b, 0x6f, 0x7b, 
	0xfb, 0xfd, 0xf7, 0x5f, 0x7f, 0xff, 0xfd, 0xf7, 
	0xdf, 0x7f, 0xff, 0xfd, 0xff, 0x87, 0xff, 0x7f, 
	0x57, 0x7f, 0x73, 0xf5, 0xd7, 0x5f, 0x7f, 0x7b, 
	0xf5, 0xd7, 0xd7, 0x4c, 0xeb, 0xda, 0xaf, 0xbf, 
	0xbd, 0xfb, 0xeb, 0xaf, 0xbf, 0xfe, 0xfb, 0xeb, 
	0xcf, 0x1a, 0xfe, 0xff
};
#define PGA_ADAT_CNT 178144

#endif /* !defined(PDPLUS_PGA_ADAT_H) */
